#ifdef HAVE_CONFIG_H                                                                                                                           
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "ocr-dbus-client.h"
#include "ocr-dbus-server.h"
#include "ocr-window.h"
#include "ocr-commands.h"

void
free_memory (OcrDBusClient   *dbus_client,
	     OcrDBusServer   *dbus_server,
	     GOptionContext  *context,
	     gchar          **image_files,
	     GError          *error, 
	     gchar           *prog_desc)
{
  if (dbus_client != NULL)
    g_object_unref (dbus_client);

  if (dbus_server != NULL)
    g_object_unref (dbus_server);

  if (context != NULL)
    g_option_context_free (context);

  if (image_files != NULL)
    g_strfreev (image_files);

  if (error != NULL)
    g_clear_error (&error);

  if (prog_desc != NULL)
    g_free (prog_desc);
}

int
main (int argc, char *argv[])
{
  OcrDBusClient   *dbus_client = NULL;
  OcrDBusServer   *dbus_server = NULL;
  GOptionContext  *context     = NULL;
  gboolean         version     = FALSE;
  gchar          **image_files = NULL;
  GError          *error       = NULL;
  GtkWidget       *window      = NULL;
  gboolean         ret_val;
  gint             i;
  gchar           *prog_desc   = NULL;

  prog_desc = g_strdup (_(" - GUI for OCR programs"));

#ifdef ENABLE_NLS
  bindtextdomain (GETTEXT_PACKAGE, OCR_LOCALE_DIR);                                                                                  
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
#endif

  g_type_init();

  GOptionEntry entries[] =
  {
    { "version", 'v', 0, G_OPTION_ARG_NONE, &(version), _("Print program version"), NULL },
    { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &(image_files), _("Image files to open"), "FILES" },
    { NULL }
  };

  context = g_option_context_new (prog_desc);
  g_option_context_add_main_entries (context, entries, NULL);
  g_option_context_add_group (context,
			      gtk_get_option_group (FALSE));
  if (!g_option_context_parse (context, &argc, &argv, &error))
  {
    g_print ("Error parsing options: %s\nType %s --help to see a full list of available command line options.\n",
	     error->message, argv[0]);

    free_memory (dbus_client,
		 dbus_server,
		 context,
		 image_files,
		 error, 
		 prog_desc);

    return 1;
  }

  if (version)
  {
    g_print (_("%s%s.\nVersion: %s.\n"), PACKAGE, prog_desc, VERSION);

    free_memory (dbus_client,
		 dbus_server,
		 context,
		 image_files,
		 error, 
		 prog_desc);

    return 0;
  }

  dbus_client = g_object_new (OCR_TYPE_DBUS_CLIENT, NULL);
  if (ocr_dbus_client_test (dbus_client) && image_files)
  {
    ret_val = TRUE;
    for (i = 0; image_files[i] != NULL && ret_val; i++)
      ret_val = ocr_dbus_client_load_image_file (dbus_client,
						 image_files[i]);

    free_memory (dbus_client,
		 dbus_server,
		 context,
		 image_files,
		 error, 
		 prog_desc);

    return ret_val;
  }
  else
    dbus_server = g_object_new (OCR_TYPE_DBUS_SERVER, NULL);

  gtk_init_with_args (&argc, &argv, prog_desc, entries, GETTEXT_PACKAGE, &error);

  if (error != NULL)
  {
    free_memory (dbus_client,
		 dbus_server,
		 context,
		 image_files,
		 NULL, 
		 prog_desc);

    g_error ("gtk_init_with_args: %s.\n", 
	     error->message);

    g_clear_error (&error);
    return 1;
  }

  window = GTK_WIDGET (g_object_new (OCR_TYPE_WINDOW,
				     "type", GTK_WINDOW_TOPLEVEL,
				     NULL));

  ocr_dbus_server_set_window (dbus_server,
			      OCR_WINDOW (window));

  gtk_widget_show_all (window);

  if (image_files)
  {
    for (i = 0; image_files[i]; i++)
    {
      ocr_load_image (OCR_WINDOW (window),
		      image_files[i]);
    }
  }

  gtk_main ();

  free_memory (dbus_client,
	       dbus_server,
	       context,
	       image_files,
	       error, 
	       prog_desc);

  return 0;
}
