#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "ocr-window.h"
#include "ocr-preferences.h"
#include "ocr-panel.h"
#include "ocr-imageview.h"
#include "ocr-textbuffer.h"
#include "ocr-textview.h"
#include "ocr-file-chooser-dialog.h"
#include "ocr-engine.h"
#include "ocr-engine-tesseract.h"
#include "ocr-engine-gocr.h"
#include "ocr-utils.h"

void
ocr_cmd_panel_run_ocr (GtkAction *action,
		       OcrPanel  *panel)
{
  OcrWindow                        *parent_window;
  OcrPreferences                   *preferences;
  OcrPreferencesPreferredOcrEngine  preferred_engine;
  OcrEngine                        *engine;
  OcrTextBuffer                    *textbuffer;
  GFile                            *file;
  gboolean                          retval;

  g_return_if_fail (OCR_IS_PANEL (panel));

  parent_window = OCR_WINDOW (ocr_panel_get_main_window (panel));

  ocr_window_set_cursor (parent_window,
			 OCR_CURSOR_OCCUPIED);

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (parent_window));
  preferred_engine = ocr_preferences_get_preferred_ocr_engine (preferences);
  switch (preferred_engine)
  {
    case OCR_PREFERENCES_PREFERRED_OCR_ENGINE_TESSERACT:
      engine = g_object_new (OCR_TYPE_ENGINE_TESSERACT,
			     "main-window", parent_window,
			     NULL);
      break;
    case OCR_PREFERENCES_PREFERRED_OCR_ENGINE_GOCR:
      engine = g_object_new (OCR_TYPE_ENGINE_GOCR,
			     "main-window", parent_window,
			     NULL);
      break;
    default:
      engine = g_object_new (OCR_TYPE_ENGINE_TESSERACT,
			     "main-window", parent_window,
			     NULL);
      break;
  }

  retval = ocr_engine_launch (engine,
			      G_FILE (ocr_panel_get_image_file (panel)));

  textbuffer = OCR_TEXTBUFFER (ocr_panel_get_textbuffer (panel));
  if (retval &&
      g_file_query_exists (engine->output_file,
			    NULL))
  {
    OCR_PANEL_SET_STATE (panel,
			 OCR_PANEL_STATE_OCR_PERFORMED);
    ocr_textbuffer_load_content (textbuffer,
				 engine->output_file);
  }
  else
  {
    gchar *title;
    gchar *message;

    if (engine->standard_error != NULL)
    {
      title = g_strdup (_("<b>Error launching the OCR engine</b>\nOCR engine output is shown below."));
      message = g_strconcat ("\n",
			     engine->standard_error,
			     NULL);
    }
    else
    {
      title   = g_strdup (_("<b>Error launching the OCR engine</b>"));
      message = NULL;
    }

    ocr_textview_load_error_message (OCR_TEXTVIEW (ocr_panel_get_textview (panel)),
				     GTK_STOCK_DIALOG_ERROR,
				     NULL,
				     title,
				     message);
    if (title != NULL)
      g_free (title);
    if (message != NULL)
      g_free (message);
  }

  g_object_unref (engine);

  ocr_window_set_cursor (parent_window,
			 OCR_CURSOR_NORMAL);

  /*
  GAppInfo *info;
  info = g_app_info_create_from_commandline ("tesseract imm out -l lang",
					     "tesseract",
					     G_APP_INFO_CREATE_NONE,
					     &error);
  if (error != NULL)
  {
    g_print ("Error %s.\n", error->message);
    g_clear_error (error);
  }*/
}

void
ocr_cmd_panel_zoom (GtkAction *action,
		    OcrPanel  *panel)
{
  GtkSpinButton *spinbutton;
  OcrImageView  *imageview;
  gdouble        percentage;
  const gchar   *action_name;

  g_return_if_fail (OCR_IS_PANEL (panel));
 
  spinbutton = GTK_SPIN_BUTTON (ocr_panel_get_image_zoom (panel));

  g_assert (GTK_IS_SPIN_BUTTON (spinbutton));

  percentage = gtk_spin_button_get_value (spinbutton);

  /* The string belongs to GTK+ and should not be freed! */
  action_name = gtk_action_get_name (action);

  if (g_strcmp0 (action_name, "ZoomIn") == 0)
    percentage += IMAGE_ZOOM_IN_OUT;
  if (g_strcmp0 (action_name, "ZoomOut") == 0)
    percentage -= IMAGE_ZOOM_IN_OUT;
  if (g_strcmp0 (action_name, "Zoom100") == 0)
    percentage = 100;
  if (g_strcmp0 (action_name, "ZoomFit") == 0)
  {
    imageview = OCR_IMAGEVIEW (ocr_panel_get_imageview (panel));
    percentage = ocr_imageview_get_fit_image_percentage (imageview);
  }

  percentage = CLAMP (percentage, IMAGE_ZOOM_MIN, IMAGE_ZOOM_MAX);

  gtk_spin_button_set_value (spinbutton, percentage);
}

gchar *
ocr_modify_content (OcrPanel *panel,
		    gchar    *content)
{
  OcrWindow       *parent_window;
  OcrPreferences  *preferences;
  gboolean         substitute_with_space;
  gchar          **tokens;
  gchar           *temp;
  gint             i;

  g_return_val_if_fail (OCR_IS_PANEL (panel), NULL);

  if (content == NULL)
    return content;

  parent_window = OCR_WINDOW (ocr_panel_get_main_window (panel));
  preferences = OCR_PREFERENCES (ocr_window_get_preferences (parent_window));

  if (!ocr_preferences_get_remove_nl (preferences))
    return content;

  substitute_with_space = ocr_preferences_get_substitute_with_space (preferences);

  temp = NULL;
  if (substitute_with_space)
  {
    temp = g_strdup (content);
    g_free (content);
    g_strdelimit (temp, "\n", ' ');
    content = g_strdup (temp);
    g_free (temp);
  }
  else
  {
    tokens = g_strsplit_set (content, "\n", -1);

    g_free (content);
    temp = NULL;
    content = g_strdup (tokens[0]);
    for (i = 1; tokens[i] != NULL; i++)
    {
      temp = g_strdup (content);
      g_free (content);
      content = g_strconcat (temp, tokens[i], NULL);
      g_free (temp);
    }

    g_strfreev(tokens);
  }
  return content;
}

void
ocr_save_text (OcrPanel *panel,
	       gchar    *uri)
{
  GtkTextBuffer *textbuffer;
  gchar         *filename;
  gchar         *content;
  gboolean       retval;
  GError        *error;

  g_return_if_fail (OCR_IS_PANEL (panel));
  g_return_if_fail (uri != NULL);

  error = NULL;
  filename = g_filename_from_uri (uri, NULL, &error);

  if (error != NULL)
  {
    g_critical (error->message,
		NULL);
    g_clear_error (&error);
    return;
  }

  textbuffer = GTK_TEXT_BUFFER (ocr_panel_get_textbuffer (panel));

  content = ocr_textbuffer_get_content (OCR_TEXTBUFFER (textbuffer)); 

  if (content != NULL)
  {
    content = ocr_modify_content (panel,
				  content);

    retval = ocr_utils_write_file (filename,
				   content);

    if (retval)
      gtk_text_buffer_set_modified (textbuffer,
				    FALSE);

    g_free (content);
  }

  g_free (filename);
}

static void
ocr_save_text_dialog_destroyed (OcrPanel             *panel, 
				OcrFileChooserDialog *dialog)
{
  g_return_if_fail (OCR_IS_PANEL (panel));
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));

  g_object_set_data (G_OBJECT (panel),
		     OCR_PANEL_SAVE_TEXT_DIALOG_KEY,
		     NULL);
}

static void
ocr_save_text_dialog_response_cb (OcrFileChooserDialog *dialog,
				  gint                  response_id,
				  OcrPanel	       *panel)
{
  gchar *uri;
	
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));
  g_return_if_fail (OCR_IS_PANEL (panel));

  if (response_id != GTK_RESPONSE_OK)
  {
    gtk_widget_destroy (GTK_WIDGET (dialog));

    return;
  }

  uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));

  g_assert (uri != NULL);

  gtk_widget_destroy (GTK_WIDGET (dialog));

  ocr_save_text (panel, uri);
  
  g_free (uri);
}

void
ocr_cmd_panel_save_text (GtkAction *action,
			 OcrPanel  *panel)
{
  GtkWidget     *save_dialog;
  gpointer       data;

  g_return_if_fail (OCR_IS_PANEL (panel));

  data = g_object_get_data (G_OBJECT (panel), OCR_PANEL_SAVE_TEXT_DIALOG_KEY);

  if (!OCR_PANEL_OCR_PERFORMED (panel))
  {
    gchar *file_display_name = ocr_panel_get_image_file_display_name (panel);
    gchar *message = g_strdup_printf (_("Text has not been yet extracted from image «%s»."),
                                      file_display_name);

    ocr_utils_message (_("OCR not performed"),
		       message,
		       GTK_MESSAGE_ERROR);

    g_free (file_display_name);
    g_free (message);

    return;
  }

  if (data != NULL)
  {
    g_assert (OCR_IS_FILE_CHOOSER_DIALOG (data));
    
    gtk_window_present (GTK_WINDOW (data));
   
    return;
  }

  save_dialog = ocr_file_chooser_dialog_new (GTK_FILE_CHOOSER_ACTION_SAVE,
					     GTK_WINDOW (ocr_panel_get_main_window (panel)),
					     OCR_FILE_CHOOSER_DIALOG_TEXT);

  g_object_set_data (G_OBJECT (panel),
		     OCR_PANEL_SAVE_TEXT_DIALOG_KEY,
		     save_dialog);

  g_object_weak_ref (G_OBJECT (save_dialog),
		     (GWeakNotify) ocr_save_text_dialog_destroyed,
		     panel);

  g_signal_connect (save_dialog,
		    "response",
		    G_CALLBACK (ocr_save_text_dialog_response_cb),
		    panel);

  gtk_widget_show (save_dialog);
}

void
ocr_cmd_panel_text_spellcheck (GtkAction *action,
			       OcrPanel  *panel)
{
  g_return_if_fail (OCR_IS_PANEL (panel));

  ocr_textview_spellcheck (OCR_TEXTVIEW (ocr_panel_get_textview (panel)));
}

static void
ocr_select_font_dialog_response_cb (GtkFontSelectionDialog *dialog,
				    gint                    response_id,
				    OcrPanel	           *panel)
{
  gchar *font_name;

  g_return_if_fail (GTK_IS_FONT_SELECTION_DIALOG (dialog));
  g_return_if_fail (OCR_IS_PANEL (panel));

  if (response_id != GTK_RESPONSE_OK)
  {
    gtk_widget_destroy (GTK_WIDGET (dialog));

    return;
  }

  font_name = gtk_font_selection_dialog_get_font_name (dialog);

  gtk_widget_destroy (GTK_WIDGET (dialog));

  ocr_panel_set_font (panel,
		      font_name);

  g_free (font_name);
}

static void
ocr_select_font_dialog_destroyed (OcrPanel               *panel, 
				  GtkFontSelectionDialog *dialog)
{
  g_return_if_fail (OCR_IS_PANEL (panel));
  g_return_if_fail (GTK_IS_FONT_SELECTION_DIALOG (dialog));

  g_object_set_data (G_OBJECT (panel),
		     OCR_PANEL_FONT_SELECTION_DIALOG_KEY,
		     NULL);
}

void
ocr_cmd_panel_select_font (GtkAction *action, 
			   OcrPanel  *panel)
{
  GtkWidget *font_dialog;
  gpointer   data;
  GtkStyle  *style;
  gchar     *font_name;

  g_return_if_fail (OCR_IS_PANEL (panel));

  data = g_object_get_data (G_OBJECT (panel),
			    OCR_PANEL_FONT_SELECTION_DIALOG_KEY);

  if (data != NULL)
  {
    g_return_if_fail (GTK_IS_FONT_SELECTION_DIALOG (data));
    
    gtk_window_present (GTK_WINDOW (data));
   
    return;
  }

  font_dialog = gtk_font_selection_dialog_new (_("Font selection"));

  g_object_set_data (G_OBJECT (panel),
		     OCR_PANEL_FONT_SELECTION_DIALOG_KEY,
		     font_dialog);

  style = gtk_widget_get_style (ocr_panel_get_textview (panel));
  font_name = pango_font_description_to_string (style->font_desc);
  gtk_font_selection_dialog_set_font_name (GTK_FONT_SELECTION_DIALOG (font_dialog),
					   font_name);

  g_object_weak_ref (G_OBJECT (font_dialog),
		     (GWeakNotify) ocr_select_font_dialog_destroyed,
		     panel);
		     
  g_signal_connect (font_dialog,
		    "response",
		    G_CALLBACK (ocr_select_font_dialog_response_cb),
		    panel);

  gtk_widget_show (font_dialog);

  g_free (font_name);
}

