#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "ocr-window.h"
#include "ocr-file-chooser-dialog.h"
#include "ocr-preferences.h"
#include "ocr-imagelist.h"
#include "ocr-panel.h"
#include "ocr-imageview.h"
#include "ocr-commands-panel.h"
#include "ocr-utils.h"

static void
ocr_add_recent_item (OcrWindow *window,
		     GFile     *file)
{
  GFileInfo        *file_info;
  GtkRecentManager *manager;
  GtkRecentData    *data;
  gchar            *uri;
  
  g_return_if_fail (OCR_IS_WINDOW (window));
  g_return_if_fail (G_IS_FILE (file));

  file_info = g_file_query_info (file,
				 "standard::*",
				 G_FILE_QUERY_INFO_NONE,
				 NULL,
				 NULL);

  data = g_slice_new (GtkRecentData);
  data->display_name = (gchar *) g_file_info_get_display_name (file_info);
  data->description = NULL;
  data->mime_type = (gchar *) g_file_info_get_content_type (file_info);
  data->app_name = (gchar *) g_get_application_name ();
  data->app_exec = g_strjoin (g_get_prgname (), " %u", NULL);
  data->groups = NULL;
  data->is_private = TRUE;
  
  uri = g_file_get_uri (file);
  manager = gtk_recent_manager_get_default ();
  gtk_recent_manager_add_full (manager,
			       uri,
			       data);
  
  g_free (uri);
  g_free (data->app_exec);
  g_slice_free (GtkRecentData,
		data);
}

void
ocr_load_image (OcrWindow *window,
		gchar     *uri_arg)
{
  OcrImageList *imagelist;
  GFile        *file;

  g_return_if_fail (OCR_IS_WINDOW (window));
  g_return_if_fail (uri_arg != NULL);

  file = g_file_new_for_commandline_arg (uri_arg);

  if (g_file_query_exists (file, NULL))
  {
    imagelist = OCR_IMAGE_LIST (ocr_window_get_imagelist (OCR_WINDOW (window)));

    g_assert (OCR_IS_IMAGE_LIST (imagelist));

    ocr_image_list_add_image (imagelist,
			      file);

    ocr_add_recent_item (window,
			 file);
  } 
  else 
  {
    GFileInfo *file_info;
    gchar     *file_display_name;
    gchar     *message;

    file_info = g_file_query_info (file,
				   G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				   G_FILE_QUERY_INFO_NONE,
				   NULL,
				   NULL);

    file_display_name = g_strdup (g_file_info_get_display_name (file_info)); 
   
    message = g_strdup_printf (_("Image «%s» does not exist."),
			       file_display_name);

    ocr_utils_message (_("Error loading image"),
		       message,
		       GTK_MESSAGE_ERROR);

    g_object_unref (file_info);
    g_free (file_display_name);
    g_free (message);
  }

  g_object_unref (file);
}

static void
ocr_open_dialog_image_response_cb (OcrFileChooserDialog *dialog, 
				   gint                  response_id, 
				   OcrWindow	        *window)
{
  gchar *uri;
	
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));
  g_return_if_fail (OCR_IS_WINDOW (window));

  if (response_id != GTK_RESPONSE_OK)
  {
    gtk_widget_destroy (GTK_WIDGET (dialog));

    return;
  }

  uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));

  g_assert (uri != NULL);

  gtk_widget_destroy (GTK_WIDGET (dialog));

  ocr_load_image (window, uri);
  
  g_free (uri);
}

static void
ocr_open_dialog_image_destroyed (OcrWindow            *window, 
				 OcrFileChooserDialog *dialog)
{
  g_return_if_fail (OCR_IS_WINDOW (window));
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));

  g_object_set_data (G_OBJECT (window),
		     OCR_WINDOW_OPEN_DIALOG_IMAGE_KEY,
		     NULL);
}

void
ocr_cmd_open_image (GtkAction *action, 
		    OcrWindow *window)
{
  GtkWidget *open_dialog;
  gpointer   data;

  data = g_object_get_data (G_OBJECT (window), OCR_WINDOW_OPEN_DIALOG_IMAGE_KEY);

  if (data != NULL)
  {
    g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (data));
    
    gtk_window_present (GTK_WINDOW (data));
   
    return;
  }

  open_dialog = ocr_file_chooser_dialog_new (GTK_FILE_CHOOSER_ACTION_OPEN,
					     GTK_WINDOW (window),
					     OCR_FILE_CHOOSER_DIALOG_IMAGE);

  g_object_set_data (G_OBJECT (window),
		     OCR_WINDOW_OPEN_DIALOG_IMAGE_KEY,
		     open_dialog);

  g_object_weak_ref (G_OBJECT (open_dialog),
		     (GWeakNotify) ocr_open_dialog_image_destroyed,
		     window);

  g_signal_connect (open_dialog,
		    "response",
		    G_CALLBACK (ocr_open_dialog_image_response_cb),
		    window);

  gtk_widget_show (open_dialog);
}

static void
ocr_save_multiple_text_dialog_destroyed (OcrImageList         *imagelist, 
					 OcrFileChooserDialog *dialog)
{
  g_return_if_fail (OCR_IS_IMAGE_LIST (imagelist));
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));

  g_object_set_data (G_OBJECT (imagelist),
		     OCR_IMAGE_LIST_SAVE_MULTIPLE_TEXT_DIALOG_KEY,
		     NULL);
}

static void
ocr_save_multiple_text_dialog_response_cb (OcrFileChooserDialog *dialog,
					   gint                  response_id,
					   OcrImageList	        *imagelist)
{
  gchar    *uri;
	
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (dialog));
  g_return_if_fail (OCR_IS_IMAGE_LIST (imagelist));

  if (response_id != GTK_RESPONSE_OK)
  {
    gtk_widget_destroy (GTK_WIDGET (dialog));

    return;
  }

  uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));

  g_assert (uri != NULL);

  gtk_widget_destroy (GTK_WIDGET (dialog));

  ocr_image_list_save_multiple_text (imagelist, uri);
  
  g_free (uri);
}

void
ocr_cmd_save_multiple_texts (GtkAction *action,
			     OcrWindow *window)
{
  GtkWidget    *save_dialog;
  OcrImageList *imagelist;
  gpointer      data;

  g_return_if_fail (OCR_IS_WINDOW (window));

  imagelist = OCR_IMAGE_LIST (ocr_window_get_imagelist (window));

  if (ocr_image_list_check_selected (imagelist))
    return;

  if (ocr_image_list_check_state_for_selected (imagelist))
    return;

  data = g_object_get_data (G_OBJECT (imagelist), OCR_IMAGE_LIST_SAVE_MULTIPLE_TEXT_DIALOG_KEY);

  if (data != NULL)
  {
    g_assert (OCR_IS_FILE_CHOOSER_DIALOG (data));
    
    gtk_window_present (GTK_WINDOW (data));
   
    return;
  }

  save_dialog = ocr_file_chooser_dialog_new (GTK_FILE_CHOOSER_ACTION_SAVE,
					     GTK_WINDOW (window),
					     OCR_FILE_CHOOSER_DIALOG_TEXT);

  g_object_set_data (G_OBJECT (imagelist),
		     OCR_IMAGE_LIST_SAVE_MULTIPLE_TEXT_DIALOG_KEY,
		     save_dialog);

  g_object_weak_ref (G_OBJECT (save_dialog),
		     (GWeakNotify) ocr_save_multiple_text_dialog_destroyed,
		     imagelist);

  g_signal_connect (save_dialog,
		    "response",
		    G_CALLBACK (ocr_save_multiple_text_dialog_response_cb),
		    imagelist);

  gtk_widget_show (save_dialog);
}

void
ocr_cmd_close_image (GtkAction *action,
		     OcrWindow *window)
{
  g_return_if_fail (OCR_IS_WINDOW (window));

  ocr_image_list_close_images (OCR_IMAGE_LIST (ocr_window_get_imagelist (window)));
}

void
ocr_cmd_quit (GtkAction *action,
	      OcrWindow *window)
{
  g_return_if_fail (OCR_IS_WINDOW (window));

  ocr_window_destroy (window);
}

void
ocr_cmd_run_ocr (GtkAction *action,
		 OcrWindow *window)
{
  OcrPanel *panel;

  g_return_if_fail (OCR_IS_WINDOW (window));

  panel = OCR_PANEL (ocr_image_list_get_active_panel (OCR_IMAGE_LIST (ocr_window_get_imagelist (window))));

  g_assert (OCR_IS_PANEL (panel));

  ocr_cmd_panel_run_ocr (action,
			 panel);
}

void
ocr_cmd_zoom (GtkAction *action,
	      OcrWindow *window)
{
  OcrPanel *panel;

  g_return_if_fail (OCR_IS_WINDOW (window));

  panel = OCR_PANEL (ocr_image_list_get_active_panel (OCR_IMAGE_LIST (ocr_window_get_imagelist (window))));

  g_assert (OCR_IS_PANEL (panel));

  ocr_cmd_panel_zoom (action,
		      panel);
}

void
ocr_cmd_save_text (GtkAction *action,
		   OcrWindow *window)
{
  OcrPanel *panel;

  g_return_if_fail (OCR_IS_WINDOW (window));

  panel = OCR_PANEL (ocr_image_list_get_active_panel (OCR_IMAGE_LIST (ocr_window_get_imagelist (window))));

  g_assert (OCR_IS_PANEL (panel));

  ocr_cmd_panel_save_text (action,
			   panel);
}

void
ocr_cmd_text_spellcheck (GtkAction *action, 
			 OcrWindow *window)
{
  OcrPanel *panel;

  g_return_if_fail (OCR_IS_WINDOW (window));

  panel = OCR_PANEL (ocr_image_list_get_active_panel (OCR_IMAGE_LIST (ocr_window_get_imagelist (window))));

  g_assert (OCR_IS_PANEL (panel));

  ocr_cmd_panel_text_spellcheck (action,
				 panel);
}

void
ocr_cmd_select_font (GtkAction *action,
		     OcrWindow *window)
{
  OcrPanel *panel;

  g_return_if_fail (OCR_IS_WINDOW (window));

  panel = OCR_PANEL (ocr_image_list_get_active_panel (OCR_IMAGE_LIST (ocr_window_get_imagelist (window))));

  g_assert (OCR_IS_PANEL (panel));

  ocr_cmd_panel_select_font (action,
			     panel);
}

void
ocr_cmd_preferences (GtkAction *action,
		     OcrWindow *window)
{
  OcrPreferences *preferences;

  g_return_if_fail (OCR_IS_WINDOW (window));

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (window));
  ocr_preferences_create_dialog (preferences);
}

void
ocr_cmd_help (GtkAction *action,
	      OcrWindow *window)
{
  GError *error = NULL;

  g_return_if_fail (OCR_IS_WINDOW (window));

  gtk_show_uri (gtk_widget_get_screen (GTK_WIDGET (window)),
		"file://"OCR_DOC_DIR"/ocrgui_manual.html",
		GDK_CURRENT_TIME,
		&error);

  if (error != NULL)
    g_clear_error (&error);
}

void
ocr_cmd_about (GtkAction *action,
	       OcrWindow *window)
{
  gchar *license_text;
  g_return_if_fail (OCR_IS_WINDOW (window));

  const gchar *authors[] = {"Emanuele Sicchiero <emanuele.sicchiero@gmail.com>", NULL};
  const char *license[] = { 
    "ocrgui is free software; you can redistribute it and/or modify " 
      "it under the terms of the GNU General Public License as published by " 
      "the Free Software Foundation; either version 3 of the License, or " 
      "(at your option) any later version.", 
    "ocrgui is distributed in the hope that it will be useful, " 
      "but WITHOUT ANY WARRANTY; without even the implied warranty of " 
      "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the " 
      "GNU General Public License for more details.", 
    "You should have received a copy of the GNU General Public License " 
      "along with ocrgui; if not, write to the Free Software Foundation, Inc., " 
      "51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA" 
  }; 

  license_text = g_strconcat (license[0], "\n\n", license[1], "\n\n",
			      license[2], "\n\n", NULL);

  gtk_show_about_dialog (GTK_WINDOW (window),
			 "authors", authors,
			 "comments", _("A GUI for OCR programs."),
			 "license", license_text,
			 "wrap-license", TRUE,
			 "logo-icon-name", "ocrgui",
			 "version", VERSION,
			 "website", "http://sourceforge.net/projects/ocrgui",
			 "copyright", "Copyright \xc2\xa9 2009-2011 Emanuele Sicchiero",
			 NULL);

  g_free (license_text);
}
