#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <dbus/dbus-glib.h>
#include <ocr-dbus-server.h>

#include "ocr-commands.h"

gboolean ocr_dbus_server_test (OcrDBusServer  *self,
			       GError        **error);

gboolean ocr_dbus_server_load_image_file (OcrDBusServer  *self,
					  gchar          *file,
					  GError        **error);

#include "ocr-dbus-server-binding.h"

#define OCR_DBUS_SERVER_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_DBUS_SERVER, OcrDBusServerPrivate))

G_DEFINE_TYPE (OcrDBusServer, ocr_dbus_server, G_TYPE_OBJECT);

struct _OcrDBusServerPrivate
{
  OcrWindow *window;
};

static void
ocr_dbus_server_dispose (GObject *gobject)
{
  g_return_if_fail (OCR_IS_DBUS_SERVER (gobject));

  G_OBJECT_CLASS (ocr_dbus_server_parent_class)->dispose (gobject);
}

static void
ocr_dbus_server_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_DBUS_SERVER (gobject));

  G_OBJECT_CLASS (ocr_dbus_server_parent_class)->finalize (gobject);
}

static void
ocr_dbus_server_constructed (GObject *gobject)
{
  OcrDBusServer        *self;
  OcrDBusServerPrivate *priv;

  g_return_if_fail (OCR_IS_DBUS_SERVER (gobject));

  self = OCR_DBUS_SERVER (gobject);
  priv = self->priv;
}

static void
ocr_dbus_server_class_init (OcrDBusServerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GError       *error = NULL;

  g_type_class_add_private (gobject_class, sizeof (OcrDBusServerPrivate));

  gobject_class->constructed = ocr_dbus_server_constructed;
  gobject_class->dispose = ocr_dbus_server_dispose;
  gobject_class->finalize = ocr_dbus_server_finalize;

  /* no need to free the following pointer */
  klass->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (klass->connection == NULL)
  {
    g_warning ("Unable to connect to dbus: %s", error->message);
    g_error_free (error);
    return;
  }

  dbus_g_object_type_install_info (OCR_TYPE_DBUS_SERVER, &dbus_glib_ocr_dbus_server_object_info);
}

static void
ocr_dbus_server_init (OcrDBusServer *self)
{
  OcrDBusServerPrivate *priv;
  DBusGProxy *proxy;
  OcrDBusServerClass *klass = OCR_DBUS_SERVER_GET_CLASS (self);
  guint request_ret;
  GError *error;

  g_return_if_fail (OCR_IS_DBUS_SERVER (self));

  self->priv = OCR_DBUS_SERVER_GET_PRIVATE (self);
  priv = self->priv;

  dbus_g_connection_register_g_object (klass->connection,
				       "/net/sourceforge/projects/ocrgui",
				       G_OBJECT (self));

  proxy = dbus_g_proxy_new_for_name (klass->connection,
				     DBUS_SERVICE_DBUS,
				     DBUS_PATH_DBUS,
				     DBUS_INTERFACE_DBUS);

  if (!org_freedesktop_DBus_request_name (proxy,
					  "net.sourceforge.projects.ocrgui",
					  0, &request_ret,
					  &error))
  {
    g_warning ("Unable to register service: %s", error->message);
    g_error_free (error);
  }
  g_object_unref (proxy);
}

gboolean ocr_dbus_server_test (OcrDBusServer  *self,
			       GError        **error)
{
  return TRUE;
}

gboolean
ocr_dbus_server_load_image_file (OcrDBusServer  *self,
				 gchar          *file,
				 GError        **error)
{
  OcrDBusServerPrivate *priv;

  g_return_val_if_fail (OCR_IS_DBUS_SERVER (self), FALSE);
  g_return_val_if_fail (file != NULL, FALSE);

  priv = self->priv;

  ocr_load_image (priv->window,
		  file);

  return TRUE;
}

void
ocr_dbus_server_set_window (OcrDBusServer *self,
			    OcrWindow     *window)
{
  OcrDBusServerPrivate *priv;

  g_return_if_fail (OCR_IS_DBUS_SERVER (self));
  g_return_if_fail (OCR_IS_WINDOW (window));

  priv = self->priv;

  priv->window = window;
}
