#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <glib/gprintf.h>

#include "ocr-engine-gocr.h"
#include "ocr-window.h"
#include "ocr-preferences.h"

#define OCR_ENGINE_GOCR_EXECUTABLE "gocr"
#define OCR_ENGINE_GOCR_OUTPUT "gocr_output.txt"
#define OCR_ENGINE_GOCR_FILE_TYPE "ppm"

G_DEFINE_TYPE (OcrEngineGocr, ocr_engine_gocr, OCR_TYPE_ENGINE);

static void
ocr_engine_gocr_dispose (GObject *gobject)
{
  g_return_if_fail (OCR_IS_ENGINE_GOCR (gobject));

  G_OBJECT_CLASS (ocr_engine_gocr_parent_class)->dispose (gobject);
}

static void
ocr_engine_gocr_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_ENGINE_GOCR (gobject));

  G_OBJECT_CLASS (ocr_engine_gocr_parent_class)->finalize (gobject);
}

static gboolean 
ocr_engine_gocr_check_programs (OcrEngine *engine)
{
  OcrEngineGocr  *self;
  OcrWindow      *main_window;
  OcrPreferences *preferences;
  gboolean        retval;

  g_return_val_if_fail (OCR_IS_ENGINE_GOCR (engine), FALSE);

  self = OCR_ENGINE_GOCR (engine);

  g_object_get (self,
		"main-window", &main_window,
		NULL);

  g_assert (OCR_IS_WINDOW (main_window));

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (main_window));
  
  g_assert (OCR_IS_PREFERENCES (preferences));

  retval = ocr_preferences_check_program (preferences,
					  "gocr");
  if (!retval)
  {
    ocr_utils_message (_("Error extracting text"),
		       _("Gocr not installed or not found."),
		       GTK_MESSAGE_ERROR);

    return retval;
  }

  retval = ocr_preferences_check_program (preferences,
					  "convert");
  if (!retval)
  {
    ocr_utils_message (_("Error extracting text"),
		       _("Convert (ImageMagick) not installed or not found."),
		       GTK_MESSAGE_ERROR);

    return retval;
  }

  return retval;
}

static gboolean 
ocr_engine_gocr_get_options (OcrEngine *engine)
{
  OcrEngineGocr  *self;
  OcrWindow      *main_window;
  OcrPreferences *preferences;
  gchar          *path;

  g_return_val_if_fail (OCR_IS_ENGINE_GOCR (engine), FALSE);

  self = OCR_ENGINE_GOCR (engine);

  g_object_get (self,
		"main-window", &main_window,
		NULL);

  g_assert (OCR_IS_WINDOW (main_window));

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (main_window));
  
  g_assert (OCR_IS_PREFERENCES (preferences));

  self->gocr_l = (gint) ocr_preferences_get_gocr_gray_level (preferences);
  self->gocr_l_autodetect = ocr_preferences_get_gocr_gray_level_autodetect (preferences);
  self->gocr_d = (gint) ocr_preferences_get_gocr_dust_size (preferences);
  self->gocr_d_autodetect = ocr_preferences_get_gocr_dust_size_autodetect (preferences);
  self->gocr_s = (gint) ocr_preferences_get_gocr_spacewidth_between_words (preferences);
  self->gocr_s_autodetect = ocr_preferences_get_gocr_spacewidth_between_words_autodetect (preferences);
  self->gocr_a = (gint) ocr_preferences_get_gocr_value_for_certainty_of_recognition (preferences);
  self->gocr_a_default = ocr_preferences_get_gocr_value_for_certainty_of_recognition_default (preferences);

  /* set temporary file type and output file path */
  engine->temp_file_type = g_strdup (OCR_ENGINE_GOCR_FILE_TYPE);
  if (engine->temp_folder == NULL)
    return FALSE;
  else
  {
    path = g_strconcat (engine->temp_folder,
			"/",
			OCR_ENGINE_GOCR_OUTPUT,
			NULL);
    engine->output_file = g_file_new_for_path (path);

    g_free (path);
  }

  return TRUE;
}

static gboolean
ocr_engine_gocr_check_valid_image_type (OcrEngine *engine,
					GFile     *file)
{
  GFileInfo *file_info;
  gchar     *type;
  gboolean   retval;

  g_return_val_if_fail (OCR_IS_ENGINE (engine), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  file_info = g_file_query_info (file,
				 G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
				 G_FILE_QUERY_INFO_NONE,
				 NULL,
				 NULL);

  engine->file_mime_type = g_strdup (g_file_info_get_content_type (file_info));

  type = g_strconcat ("image/",
		      OCR_ENGINE_GOCR_FILE_TYPE,
		      NULL);

  if (!g_strcmp0 (engine->file_mime_type, type))
    retval = TRUE;
  else
  {
    engine->convert_options = NULL;

    retval = FALSE;
  }

  g_object_unref (file_info);
  g_free (type);

  return retval;
}

gboolean
ocr_engine_gocr_create_command (OcrEngine *engine,
				GFile     *file)
{
  gchar *file_path;
  gint   i;

  g_return_val_if_fail (OCR_IS_ENGINE (engine), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  if (engine->command != NULL)
  {
    g_strfreev (engine->command);
    engine->command = NULL;
  }

  file_path = g_file_get_path (file);

  i=6;
  if (!(OCR_ENGINE_GOCR (engine)->gocr_l_autodetect))
    i+=2;
  if (!(OCR_ENGINE_GOCR (engine)->gocr_d_autodetect))
    i+=2;
  if (!(OCR_ENGINE_GOCR (engine)->gocr_s_autodetect))
    i+=2;
  if (!(OCR_ENGINE_GOCR (engine)->gocr_a_default))
    i+=2;

  engine->command = g_new (gchar*, i);
  engine->command[0] = g_strdup (OCR_ENGINE_GOCR_EXECUTABLE);
  engine->command[1] = g_strdup ("-i");
  engine->command[2] = file_path;
  engine->command[3] = g_strdup ("-o");
  engine->command[4] = g_strdup (OCR_ENGINE_GOCR_OUTPUT);
  i=4;
  if (!(OCR_ENGINE_GOCR (engine)->gocr_l_autodetect))
  {
    engine->command[++i] = g_strdup ("-l");
    engine->command[++i] = g_strdup_printf ("%d",
					    OCR_ENGINE_GOCR (engine)->gocr_l);
  }
  if (!(OCR_ENGINE_GOCR (engine)->gocr_d_autodetect))
  {
    engine->command[++i] = g_strdup ("-d");
    engine->command[++i] = g_strdup_printf ("%d",
					    OCR_ENGINE_GOCR (engine)->gocr_d);
  }
  if (!(OCR_ENGINE_GOCR (engine)->gocr_s_autodetect))
  {
    engine->command[++i] = g_strdup ("-s");
    engine->command[++i] = g_strdup_printf ("%d",
					    OCR_ENGINE_GOCR (engine)->gocr_s);
  }
  if (!(OCR_ENGINE_GOCR (engine)->gocr_a_default))
  {
    engine->command[++i] = g_strdup ("-a");
    engine->command[++i] = g_strdup_printf ("%d",
					    OCR_ENGINE_GOCR (engine)->gocr_a);
  }
  engine->command[++i] = NULL;

  return TRUE;
}

static void
ocr_engine_gocr_class_init (OcrEngineGocrClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  OcrEngineClass *ocr_engine_class = OCR_ENGINE_CLASS (klass);

  gobject_class->dispose = ocr_engine_gocr_dispose;
  gobject_class->finalize = ocr_engine_gocr_finalize;

  ocr_engine_class->check_programs = ocr_engine_gocr_check_programs;
  ocr_engine_class->get_options = ocr_engine_gocr_get_options;
  ocr_engine_class->check_valid_image_type = ocr_engine_gocr_check_valid_image_type;
  ocr_engine_class->create_command = ocr_engine_gocr_create_command;
}

static void
ocr_engine_gocr_init (OcrEngineGocr *self)
{
  g_return_if_fail (OCR_IS_ENGINE_GOCR (self));

  self->gocr_l            = 0;
  self->gocr_l_autodetect = TRUE;
  self->gocr_d            = 0;
  self->gocr_d_autodetect = TRUE;
  self->gocr_s            = 0;
  self->gocr_s_autodetect = TRUE;
  self->gocr_a            = 0;
  self->gocr_a_default    = TRUE;
}

