#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <glib/gprintf.h>

#include "ocr-engine-tesseract.h"
#include "ocr-window.h"
#include "ocr-preferences.h"

#define OCR_ENGINE_TESSERACT_EXECUTABLE "tesseract"
#define OCR_ENGINE_TESSERACT_OUTPUT "tesseract_output"
#define OCR_ENGINE_TESSERACT_FILE_TYPE "bmp"

G_DEFINE_TYPE (OcrEngineTesseract, ocr_engine_tesseract, OCR_TYPE_ENGINE);

static void
ocr_engine_tesseract_dispose (GObject *gobject)
{
  g_return_if_fail (OCR_IS_ENGINE_TESSERACT (gobject));

  G_OBJECT_CLASS (ocr_engine_tesseract_parent_class)->dispose (gobject);
}

static void
ocr_engine_tesseract_finalize (GObject *gobject)
{
  OcrEngineTesseract *self;

  g_return_if_fail (OCR_IS_ENGINE_TESSERACT (gobject));

  self = OCR_ENGINE_TESSERACT (gobject);

  if (self->tesseract_language_code != NULL)
    g_free (self->tesseract_language_code);

  G_OBJECT_CLASS (ocr_engine_tesseract_parent_class)->finalize (gobject);
}

static gboolean 
ocr_engine_tesseract_check_programs (OcrEngine *engine)
{
  OcrEngineTesseract *self;
  OcrWindow          *main_window;
  OcrPreferences     *preferences;
  gboolean            retval;

  g_return_val_if_fail (OCR_IS_ENGINE_TESSERACT (engine), FALSE);

  self = OCR_ENGINE_TESSERACT (engine);

  g_object_get (self,
		"main-window", &main_window,
		NULL);

  g_assert (OCR_IS_WINDOW (main_window));

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (main_window));
  
  g_assert (OCR_IS_PREFERENCES (preferences));

  retval = ocr_preferences_check_program (preferences,
					  "tesseract");
  if (!retval)
  {
    ocr_utils_message (_("Error extracting text"),
		       _("Tesseract not installed or not found."),
		       GTK_MESSAGE_ERROR);

    return retval;
  }

  retval = ocr_preferences_check_program (preferences,
					  "convert");
  if (!retval)
  {
    ocr_utils_message (_("Error extracting text"),
		       _("Convert (ImageMagick) not installed or not found."),
		       GTK_MESSAGE_ERROR);

    return retval;
  }

  return retval;
}

static gboolean 
ocr_engine_tesseract_get_options (OcrEngine *engine)
{
  OcrEngineTesseract *self;
  OcrWindow          *main_window;
  OcrPreferences     *preferences;
  gchar              *path;

  g_return_val_if_fail (OCR_IS_ENGINE_TESSERACT (engine), FALSE);

  self = OCR_ENGINE_TESSERACT (engine);

  if (self->tesseract_language_code != NULL)
    g_free (self->tesseract_language_code);

  g_object_get (self,
		"main-window", &main_window,
		NULL);

  g_assert (OCR_IS_WINDOW (main_window));

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (main_window));
  
  g_assert (OCR_IS_PREFERENCES (preferences));

  self->tesseract_language_code = ocr_preferences_get_tesseract_valid_language_code (preferences);

  if (self->tesseract_language_code == NULL)
    return FALSE;

  engine->temp_file_type = g_strdup (OCR_ENGINE_TESSERACT_FILE_TYPE);
  if (engine->temp_folder == NULL)
    return FALSE;
  else
  {
    path = g_strconcat (engine->temp_folder,
			"/",
			OCR_ENGINE_TESSERACT_OUTPUT,
			".txt",
			NULL);
    engine->output_file = g_file_new_for_path (path);

    g_free (path);
  }

  return TRUE;
}

static gboolean
ocr_engine_tesseract_check_valid_image_type (OcrEngine *engine,
					     GFile     *file)
{
  GFileInfo *file_info;
  gchar     *type;
  gboolean   retval;

  g_return_val_if_fail (OCR_IS_ENGINE (engine), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  file_info = g_file_query_info (file,
				 G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
				 G_FILE_QUERY_INFO_NONE,
				 NULL,
				 NULL);

  engine->file_mime_type = g_strdup (g_file_info_get_content_type (file_info));

  type = g_strconcat ("image/",
		      OCR_ENGINE_TESSERACT_FILE_TYPE,
		      NULL);

  if (!g_strcmp0 (engine->file_mime_type, type))
    retval = TRUE;
  else
  {
    engine->convert_options = g_new (gchar*, 3);
    engine->convert_options[0] = g_strdup ("-depth");
    engine->convert_options[1] = g_strdup ("24");
    engine->convert_options[2] = NULL;

    retval = FALSE;
  }

  g_object_unref (file_info);
  g_free (type);

  return retval;
}

gboolean
ocr_engine_tesseract_create_command (OcrEngine *engine,
				     GFile     *file)
{
  gchar *file_path;

  g_return_val_if_fail (OCR_IS_ENGINE (engine), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  if (engine->command != NULL)
  {
    g_strfreev (engine->command);
    engine->command = NULL;
  }

  file_path = g_file_get_path (file);

  if (!g_strcmp0 (OCR_ENGINE_TESSERACT (engine)->tesseract_language_code,
		  "NOT_SET"))
  {
    engine->command = g_new (gchar*, 4);
    engine->command[0] = g_strdup (OCR_ENGINE_TESSERACT_EXECUTABLE);
    engine->command[1] = file_path;
    engine->command[2] = g_strdup (OCR_ENGINE_TESSERACT_OUTPUT);
    engine->command[3] = NULL;
  }
  else
  {
    engine->command = g_new (gchar*, 6);
    engine->command[0] = g_strdup (OCR_ENGINE_TESSERACT_EXECUTABLE);
    engine->command[1] = file_path;
    engine->command[2] = g_strdup (OCR_ENGINE_TESSERACT_OUTPUT);
    engine->command[3] = g_strdup ("-l");
    engine->command[4] = g_strdup (OCR_ENGINE_TESSERACT (engine)->tesseract_language_code);
    engine->command[5] = NULL;
  }

  return TRUE;
}

static void
ocr_engine_tesseract_class_init (OcrEngineTesseractClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  OcrEngineClass *ocr_engine_class = OCR_ENGINE_CLASS (klass);

  gobject_class->dispose = ocr_engine_tesseract_dispose;
  gobject_class->finalize = ocr_engine_tesseract_finalize;

  ocr_engine_class->check_programs = ocr_engine_tesseract_check_programs;
  ocr_engine_class->get_options = ocr_engine_tesseract_get_options;
  ocr_engine_class->check_valid_image_type = ocr_engine_tesseract_check_valid_image_type;
  ocr_engine_class->create_command = ocr_engine_tesseract_create_command;
}

static void
ocr_engine_tesseract_init (OcrEngineTesseract *self)
{
  g_return_if_fail (OCR_IS_ENGINE_TESSERACT (self));

  self->tesseract_language_code = NULL;
}

