#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-engine.h"
#include "ocr-window.h"
#include "ocr-preferences.h"
#include "ocr-utils.h"

#define OCR_ENGINE_TEMP_FILE "/temp."

#define OCR_ENGINE_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_ENGINE, OcrEnginePrivate))

G_DEFINE_ABSTRACT_TYPE (OcrEngine, ocr_engine, G_TYPE_OBJECT);

enum
{
  PROP_MAIN_WINDOW = 1
};

struct _OcrEnginePrivate
{
  OcrWindow *main_window;
};

static void
ocr_engine_dispose (GObject *gobject)
{
  OcrEngine        *self;
  OcrEnginePrivate *priv;
  static gboolean   first_run = TRUE;

  g_return_if_fail (OCR_IS_ENGINE (gobject));

  self = OCR_ENGINE (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->main_window != NULL)
      g_object_unref (priv->main_window);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_engine_parent_class)->dispose (gobject);
}

static void
ocr_engine_finalize (GObject *gobject)
{
  OcrEngine *self;

  g_return_if_fail (OCR_IS_ENGINE (gobject));

  self = OCR_ENGINE (gobject);

  if (self->temp_folder != NULL)
    g_free (self->temp_folder);

  if (self->file_mime_type != NULL)
    g_free (self->file_mime_type);

  if (self->temp_file != NULL)
    g_object_unref (self->temp_file);

  if (self->temp_file_type != NULL)
    g_free (self->temp_file_type);

  if (self->convert_options != NULL)
    g_strfreev (self->convert_options);

  if (self->output_file != NULL)
    g_object_unref (self->output_file);

  if (self->command != NULL)
    g_strfreev (self->command);

  if (self->standard_output != NULL)
    g_free (self->standard_output);

  if (self->standard_error != NULL)
    g_free (self->standard_error); 

  G_OBJECT_CLASS (ocr_engine_parent_class)->finalize (gobject);
}

gboolean
ocr_engine_check_programs (OcrEngine *self)
{
  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);

  return (OCR_ENGINE_GET_CLASS (self)->check_programs (self));
}

static void
ocr_engine_get_common_options (OcrEngine *self)
{
  OcrEnginePrivate *priv;
  OcrPreferences   *preferences;

  g_return_if_fail (OCR_IS_ENGINE (self));

  priv = self->priv;

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (priv->main_window));

  g_assert (OCR_IS_PREFERENCES (preferences));

  self->temp_folder = ocr_preferences_get_temp_folder (preferences);

  g_assert (self->temp_folder != NULL);
}

gboolean
ocr_engine_get_options (OcrEngine *self)
{
  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);

  return (OCR_ENGINE_GET_CLASS (self)->get_options (self));
}

gboolean
ocr_engine_check_valid_image_type (OcrEngine *self,
				   GFile     *file)
{
  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);

  return (OCR_ENGINE_GET_CLASS (self)->check_valid_image_type (self,
							       file));
}

static gboolean
ocr_engine_convert_image (OcrEngine *self,
			  GFile     *file)
{
  gchar   *file_path;
  gchar   *temp_path;
  gint     i;
  gchar  **argv;
  GError  *error = NULL;

  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  g_assert (self->temp_file_type != NULL);

  file_path = g_file_get_path (file);
  temp_path = g_strconcat (self->temp_folder,
			   OCR_ENGINE_TEMP_FILE,
			   self->temp_file_type,
			   NULL);

  if (self->temp_file != NULL)
    g_object_unref (self->temp_file);

  self->temp_file = g_file_new_for_path (temp_path);

  if (self->convert_options != NULL)
    for (i = 0; self->convert_options[i] != NULL; i++) {}
  else
    i = 0;

  argv = g_new (gchar*, 4+i);
  argv[0] = g_strdup ("convert");
  argv[1] = file_path;

  if (self->convert_options != NULL)
    for (i = 0; self->convert_options[i] != NULL; i++)
      argv[1+i+1] = g_strdup (self->convert_options[i]);
  else
    i = 0;

  argv[1+i+1] = g_strdup (temp_path);
  argv[1+i+2] = NULL;

  g_free (temp_path);

  g_spawn_sync (self->temp_folder,
		argv,
		NULL,
		G_SPAWN_SEARCH_PATH,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		&error);

  g_strfreev (argv);

  if (error != NULL)
  {
    GFileInfo *file_info;
    gchar     *file_display_name;
    gchar     *message;
   
    file_info = g_file_query_info (file,
				   G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				   G_FILE_QUERY_INFO_NONE,
				   NULL,
				   NULL);

    file_display_name = g_strdup (g_file_info_get_display_name (file_info)); 

    message = g_strdup_printf (_("Unable to convert image «%s» in %s format: %s."),
			       file_display_name,
			       self->temp_file_type,
			       error->message);

    ocr_utils_message (_("Error converting image"),
		       message,
		       GTK_MESSAGE_ERROR);


    g_object_unref (file_info);
    g_free (file_display_name);
    g_free (message);
    g_clear_error (&error);
    return FALSE;
  }

  return TRUE;
}

gboolean
ocr_engine_create_command (OcrEngine *self,
			   GFile     *file)
{
  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);

  return (OCR_ENGINE_GET_CLASS (self)->create_command (self,
						       file));
}

gboolean
ocr_engine_delete_output_file (OcrEngine *self)
{
  GFileInfo *file_info;
  gchar     *file_display_name;
  GError    *error = NULL;

  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);
  g_return_val_if_fail (G_IS_FILE (self->output_file), FALSE);

  if (g_file_query_exists (self->output_file,
			   NULL))
  {
    g_file_delete (self->output_file,
		   NULL,
		   &error);

    if (error != NULL)
    {
      file_info = g_file_query_info (self->output_file,
				     G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				     G_FILE_QUERY_INFO_NONE,
				     NULL,
				     NULL);

      file_display_name = g_strdup (g_file_info_get_display_name (file_info)); 
      gchar *title = g_strdup_printf (_("Error deleting «%s»"),
				      file_display_name);

      ocr_utils_message (title,
			 error->message,
			 GTK_MESSAGE_ERROR);

      g_object_unref (file_info);
      g_free (file_display_name);
      g_free (title);
      g_clear_error (&error);
      return FALSE;
    }
  }

  return TRUE;
}

gboolean
ocr_engine_command_execute (OcrEngine *self)
{
  GError *error = NULL;

  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);

  if (self->command == NULL)
    return FALSE;

  if (self->standard_output != NULL)
    g_free (self->standard_output);

  g_spawn_sync (self->temp_folder,
		self->command,
		NULL,
		G_SPAWN_SEARCH_PATH,
		NULL,
		NULL,
		&(self->standard_output),
		&(self->standard_error),
		NULL,
		&error);

  g_strfreev (self->command);
  self->command = NULL;

  if (error != NULL)
  {
    gchar *message = g_strdup_printf (_("Can't execute command:\n%s"),
				      error->message);

    ocr_utils_message (_("Error extracting text"),
		       message,
		       GTK_MESSAGE_ERROR);

    g_free (message);
    g_clear_error (&error);
    return FALSE;
  }

  return TRUE;
}

gboolean
ocr_engine_launch (OcrEngine *self,
		   GFile     *file)
{
  OcrEnginePrivate *priv;
  gboolean          retval;
  GFile            *_file;
  guint             message_id;

  g_return_val_if_fail (OCR_IS_ENGINE (self), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  priv = self->priv;

  _file = file;

  retval = ocr_engine_check_programs (self);

  if (retval == FALSE)
    return FALSE;

  ocr_engine_get_common_options (self);

  retval = ocr_engine_get_options (self);

  if (retval == FALSE)
    return FALSE;
  
  retval = ocr_engine_delete_output_file (self);

  if (retval == FALSE)
    return FALSE;

  retval = ocr_engine_check_valid_image_type (self,
					      file);

  if (!retval)
  {
    retval = ocr_engine_convert_image (self,
				       file);

    g_assert (G_IS_FILE (self->temp_file));

    if (retval)
      _file = self->temp_file;
  }

  if (retval == FALSE)
    return FALSE;

  retval = ocr_engine_create_command (self,
				      _file);

  if (retval == FALSE)
    return FALSE;

  gtk_statusbar_pop (priv->main_window->statusbar,
		     priv->main_window->context_id);
  message_id = gtk_statusbar_push (priv->main_window->statusbar,
				   priv->main_window->context_id,
				   _("Extracting text..."));

  retval = ocr_engine_command_execute (self);

  if (retval &&
      g_file_query_exists (self->output_file,
			   NULL))
  {
    gtk_statusbar_remove (priv->main_window->statusbar,
			  priv->main_window->context_id,
			  message_id);

    message_id = gtk_statusbar_push (priv->main_window->statusbar,
				     priv->main_window->context_id,
				     _("Text extracted successfully"));
  }
  else
  {
    gtk_statusbar_remove (priv->main_window->statusbar,
			  priv->main_window->context_id,
			  message_id);

    message_id = gtk_statusbar_push (priv->main_window->statusbar,
				     priv->main_window->context_id,
				     _("Text extraction failed"));

    return FALSE;
  }

  return TRUE;
}

static void
ocr_engine_set_property (GObject      *gobject,
			 guint         property_id,
			 const GValue *value,
			 GParamSpec   *pspec)
{
  OcrEngine        *self;
  OcrEnginePrivate *priv;

  g_return_if_fail (OCR_IS_ENGINE (gobject));

  self = OCR_ENGINE (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_MAIN_WINDOW:
      g_assert (OCR_IS_WINDOW (g_value_get_object (value)));
      priv->main_window = OCR_WINDOW (g_value_dup_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_engine_get_property (GObject      *gobject,
			 guint         property_id,
			 GValue       *value,
			 GParamSpec   *pspec)
{
  OcrEngine        *self;
  OcrEnginePrivate *priv;

  g_return_if_fail (OCR_IS_ENGINE (gobject));

  self = OCR_ENGINE (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_MAIN_WINDOW:
      g_value_set_object (value, priv->main_window);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_engine_class_init (OcrEngineClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrEnginePrivate));

  gobject_class->set_property = ocr_engine_set_property;
  gobject_class->get_property = ocr_engine_get_property;
  gobject_class->dispose = ocr_engine_dispose;
  gobject_class->finalize = ocr_engine_finalize;

  klass->check_programs         = NULL;
  klass->get_options            = NULL;
  klass->check_valid_image_type = NULL;
  klass->create_command         = NULL;

  g_object_class_install_property (gobject_class, PROP_MAIN_WINDOW,
				   g_param_spec_object ("main-window", "main-window", "main-window",
							OCR_TYPE_WINDOW,
							G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}

static void
ocr_engine_init (OcrEngine *self)
{
  OcrEnginePrivate *priv;

  g_return_if_fail (OCR_IS_ENGINE (self));

  self->priv = OCR_ENGINE_GET_PRIVATE (self);
  priv = self->priv;

  priv->main_window     = NULL;

  self->temp_folder     = NULL;
  self->file_mime_type  = NULL;
  self->temp_file       = NULL;
  self->temp_file_type  = NULL;
  self->convert_options = NULL;
  self->output_file     = NULL;
  self->command         = NULL;
  self->standard_output = NULL;
  self->standard_error  = NULL;
}

