#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>

#include "ocr-file-chooser-dialog.h"
#include "ocr-window.h"
#include "ocr-preferences.h"

#define OCR_FILE_CHOOSER_DIALOG_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_FILE_CHOOSER_DIALOG, OcrFileChooserDialogPrivate))

enum
{
  PROP_PARENT_WINDOW = 1,
  PROP_TYPE
};

struct _OcrFileChooserDialogPrivate
{
  GtkWindow                *parent_window;
  OcrFileChooserDialogType  type;
};

G_DEFINE_TYPE(OcrFileChooserDialog, ocr_file_chooser_dialog, GTK_TYPE_FILE_CHOOSER_DIALOG)

static void
ocr_file_chooser_dialog_dispose (GObject *gobject)
{
  OcrFileChooserDialog        *self;
  OcrFileChooserDialogPrivate *priv;
  static gboolean              first_run = TRUE;

  self = OCR_FILE_CHOOSER_DIALOG (gobject);
  priv = self->priv;

  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (gobject));

  if (first_run)
  {
    if (priv->parent_window != NULL)
      g_object_unref (priv->parent_window);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_file_chooser_dialog_parent_class)->dispose (gobject);
}

static void
ocr_file_chooser_dialog_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (gobject));

  G_OBJECT_CLASS (ocr_file_chooser_dialog_parent_class)->finalize (gobject);
}

static void
ocr_file_chooser_dialog_image_add_filters (OcrFileChooserDialog *self)
{
  GtkFileFilter    *filter;
  GtkFileFilter    *all_images_filter;
  GSList           *formats;
  GSList           *it;
  gchar            *description;
  gchar            *extension;
  gchar            *filter_name;
  gchar           **mime_types;
  gchar           **pattern;
  gchar            *tmp;
  GdkPixbufFormat  *format;
  gint              i;

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter,
			    _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (self),
			       filter);

  all_images_filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter,
			    _("Image files only"));
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (self),
			       all_images_filter);

  formats = gdk_pixbuf_get_formats ();
  for (it = formats; it != NULL; it = it->next)
  {
    filter = gtk_file_filter_new (); 

    format = (GdkPixbufFormat*) it->data;
    description = gdk_pixbuf_format_get_description (format);
    extension = gdk_pixbuf_format_get_name (format);

    /* Filter name: First description then file extension, eg. "The PNG-Format (*.png)". */
    filter_name = g_strdup_printf (_("%s (*.%s)"),
				   description,
				   extension);

    g_free (description);
    g_free (extension);

    gtk_file_filter_set_name (filter,
			      filter_name);

    g_free (filter_name);

    mime_types = gdk_pixbuf_format_get_mime_types ((GdkPixbufFormat *) it->data);
    for (i = 0; mime_types[i] != NULL; i++)
    {
      gtk_file_filter_add_mime_type (filter,
				     mime_types[i]);
      gtk_file_filter_add_mime_type (all_images_filter,
				     mime_types[i]);
    }

    g_strfreev (mime_types);

    pattern = gdk_pixbuf_format_get_extensions ((GdkPixbufFormat *) it->data);
    for (i = 0; pattern[i] != NULL; i++)
    {
      tmp = g_strconcat ("*.",
			 pattern[i],
			 NULL);
      gtk_file_filter_add_pattern (filter,
				   tmp);
      gtk_file_filter_add_pattern (all_images_filter,
				   tmp);

      g_free (tmp);
    }

    g_strfreev (pattern);

    gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (self),
				 filter);
  }
  g_slist_free (formats);
}

GtkWidget *
ocr_file_chooser_dialog_new (GtkFileChooserAction      action,
			     GtkWindow                *parent_window,
			     OcrFileChooserDialogType  type)
{
  OcrFileChooserDialog        *self;
  OcrFileChooserDialogPrivate *priv;
  GtkFileFilter               *filter;
  gchar                       *title = NULL;
  OcrPreferences              *preferences;
  gchar                       *folder;

  g_return_val_if_fail (GTK_IS_WINDOW (parent_window), NULL);

  self = g_object_new (OCR_TYPE_FILE_CHOOSER_DIALOG,
		       "action", action,
		       "select-multiple", FALSE,
		       "local-only", TRUE,
		       "parent-window", parent_window,
		       "type", type,
		       NULL);

  g_assert (OCR_IS_FILE_CHOOSER_DIALOG (self));

  priv = self->priv;

  action = gtk_file_chooser_get_action (GTK_FILE_CHOOSER (self));

  switch (action)
  {
    case GTK_FILE_CHOOSER_ACTION_OPEN:
      gtk_dialog_add_buttons (GTK_DIALOG (self),
			      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			      GTK_STOCK_OPEN, GTK_RESPONSE_OK,
			      NULL);
	    
      switch (priv->type)
      {
	case OCR_FILE_CHOOSER_DIALOG_IMAGE:
	  title = _("Load Image");

	  ocr_file_chooser_dialog_image_add_filters (self);

	  break;
	default:
	  g_assert_not_reached ();
	  break;
      }

      break;
  
    case GTK_FILE_CHOOSER_ACTION_SAVE:
      gtk_dialog_add_buttons (GTK_DIALOG (self),
			      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			      GTK_STOCK_SAVE, GTK_RESPONSE_OK,
			      NULL);
	    
      switch (priv->type)
      {
	case OCR_FILE_CHOOSER_DIALOG_TEXT:
	  title = _("Save Text");

	  preferences = OCR_PREFERENCES (ocr_window_get_preferences (OCR_WINDOW (parent_window)));

	  g_assert (OCR_IS_PREFERENCES (preferences));

	  folder = ocr_preferences_get_save_default_folder (preferences);
	  gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (self),
					       folder);
	  g_free (folder);

	  filter = gtk_file_filter_new ();
	  gtk_file_filter_set_name (filter,
				    _("All files"));
	  gtk_file_filter_add_pattern (filter,
				       "*");
	  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (self),
				       filter);

	  filter = gtk_file_filter_new ();
	  gtk_file_filter_set_name (filter,
				    _("Txt files"));
	  gtk_file_filter_add_mime_type (filter,
					 "text/plain");
	  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (self),
				       filter);

	  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (self),
				       filter);
	  break;
	default:
	  g_assert_not_reached ();
	  break;
      }

      break;

    default:
      g_assert_not_reached ();
      break;
  }

  gtk_window_set_title (GTK_WINDOW (self), title);
  gtk_window_set_modal (GTK_WINDOW (self), TRUE);
  gtk_window_set_transient_for (GTK_WINDOW (self), priv->parent_window);
  gtk_window_set_destroy_with_parent (GTK_WINDOW (self), TRUE);
  gtk_dialog_set_default_response (GTK_DIALOG (self), GTK_RESPONSE_OK);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (self), TRUE);

  return GTK_WIDGET (self);
}

static void
ocr_file_chooser_dialog_set_property (GObject *gobject,
				      guint property_id,
				      const GValue *value,
				      GParamSpec *pspec)
{
  OcrFileChooserDialog        *self;
  OcrFileChooserDialogPrivate *priv;

  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (gobject));

  self = OCR_FILE_CHOOSER_DIALOG (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_PARENT_WINDOW:
      g_return_if_fail (GTK_IS_WINDOW (g_value_get_object (value)));
      priv->parent_window = GTK_WINDOW (g_value_dup_object (value));
      break;
    case PROP_TYPE:
      priv->type = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_file_chooser_dialog_class_init (OcrFileChooserDialogClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrFileChooserDialogPrivate));
  
  gobject_class->set_property = ocr_file_chooser_dialog_set_property;
  gobject_class->dispose = ocr_file_chooser_dialog_dispose;
  gobject_class->finalize = ocr_file_chooser_dialog_finalize;

  g_object_class_install_property
    (gobject_class, PROP_PARENT_WINDOW,
     g_param_spec_object ("parent-window", "parent-window", "parent-window",
			  GTK_TYPE_WINDOW,
			  G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));

  g_object_class_install_property
    (gobject_class, PROP_TYPE,
     g_param_spec_int ("type", "type", "type",
		       1, OCR_FILE_CHOOSER_DIALOG_MAX_TYPE, 1,
		       G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_file_chooser_dialog_init (OcrFileChooserDialog *self)
{
  OcrFileChooserDialogPrivate *priv;

  g_return_if_fail (OCR_IS_FILE_CHOOSER_DIALOG (self));

  self->priv = OCR_FILE_CHOOSER_DIALOG_GET_PRIVATE (self);
  priv = self->priv;

  priv->parent_window = NULL;
  priv->type = 0;
}
