#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-imagelist.h"
#include "ocr-window.h"
#include "ocr-preferences.h"
#include "ocr-panel.h"
#include "ocr-textbuffer.h"
#include "ocr-utils.h"
#include "ocr-commands-panel.h"

#define OCR_IMAGE_LIST_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_IMAGE_LIST, OcrImageListPrivate))

G_DEFINE_TYPE (OcrImageList, ocr_image_list, GTK_TYPE_VBOX);

enum 
{
  COLUMN_FILE,
  COLUMN_PATH,
  COLUMN_FILE_NAME,
  COLUMN_PIXBUF,
  COLUMN_PANEL_POINTER,
  N_COLUMNS
};

enum
{
  PROP_MAIN_WINDOW = 1,
};

struct _OcrImageListPrivate
{
  OcrWindow           *main_window;
  GtkWidget           *toolbar;
  GtkListStore        *liststore;
  GtkIconView         *iconview;
  GtkTreeRowReference *activated_item;
  GList               *reference;
  GList               *file_list;
};

gboolean
ocr_image_list_destroy_images (GtkTreeModel *liststore,
			       GtkTreePath  *path,
			       GtkTreeIter  *iter,
			       gpointer      data)
{
  GFile     *col_file;
  gchar     *col_path;
  gchar     *col_file_name;
  GdkPixbuf *col_pixbuf;
  OcrPanel  *col_panel;

  g_return_val_if_fail (GTK_IS_TREE_MODEL (liststore), TRUE);

  gtk_tree_model_get (GTK_TREE_MODEL (liststore),
		      iter,
		      COLUMN_FILE, &col_file,
		      COLUMN_PATH, &col_path,
		      COLUMN_FILE_NAME, &col_file_name,
		      COLUMN_PIXBUF, &col_pixbuf,
		      COLUMN_PANEL_POINTER, &col_panel,
		      -1);

  g_object_unref (col_file);
  g_free (col_path);
  g_free (col_file_name);
  g_object_unref (col_pixbuf);

  ocr_panel_destroy_selection_font_dialog (col_panel);
  ocr_panel_destroy_save_text_dialog (col_panel);

  gtk_widget_destroy (GTK_WIDGET (col_panel));

  return FALSE;
}

static void
ocr_image_list_free_file_list (OcrImageList *self)
{
  OcrImageListPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  g_list_foreach (priv->file_list,
		  (GFunc) g_free,
		  NULL);

  g_list_free (priv->file_list);
  priv->file_list = NULL;
}

static void
ocr_image_list_dispose (GObject *gobject)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  static gboolean first_run = TRUE;

  g_return_if_fail (OCR_IS_IMAGE_LIST (gobject));

  self = OCR_IMAGE_LIST (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->main_window != NULL)
      g_object_unref (priv->main_window);

    if (priv->liststore != NULL)
    {
      gtk_tree_model_foreach (GTK_TREE_MODEL (priv->liststore),
			      ocr_image_list_destroy_images,
			      NULL);

      g_object_unref (priv->liststore);
    }

    if (priv->activated_item != NULL)
      gtk_tree_row_reference_free (priv->activated_item);

    if (priv->reference != NULL)
    {
      GList *it;

      for (it = priv->reference; it != NULL; it = it->next)
	gtk_tree_row_reference_free ((GtkTreeRowReference*) it->data);

      g_list_free (priv->reference);
    }

    if (priv->file_list != NULL)
      ocr_image_list_free_file_list (self);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_image_list_parent_class)->dispose (gobject);
}

static void
ocr_image_list_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_IMAGE_LIST (gobject));

  G_OBJECT_CLASS (ocr_image_list_parent_class)->finalize (gobject);
}

GtkWidget *
ocr_image_list_get_active_panel (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  OcrPanel            *panel;
  GtkTreePath         *path_activated_item;
  GtkTreeIter          iter;

  g_return_val_if_fail (OCR_IS_IMAGE_LIST (self), NULL);

  priv = self->priv;

  if (priv->activated_item == NULL ||
      !gtk_tree_row_reference_valid (priv->activated_item))
    return NULL;
  
  path_activated_item = gtk_tree_row_reference_get_path (priv->activated_item);

  gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->liststore),
			   &iter,
			   path_activated_item);

  gtk_tree_model_get (GTK_TREE_MODEL (priv->liststore),
		      &iter,
		      COLUMN_PANEL_POINTER, &panel,
		      -1);

  gtk_tree_path_free (path_activated_item);

  return GTK_WIDGET (panel);
}

static void
ocr_image_list_show_panel (OcrImageList *self,
			   GtkTreePath  *path,
			   gboolean      action)
{
  OcrImageListPrivate *priv;
  GtkTreeIter          iter;
  OcrPanel            *panel;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->liststore),
			   &iter,
			   path);

  gtk_tree_model_get (GTK_TREE_MODEL (priv->liststore),
		      &iter,
		      COLUMN_PANEL_POINTER, &panel,
		      -1);

  g_assert (OCR_IS_PANEL (panel));

  switch (action)
  {
    case TRUE:
      gtk_widget_show_all (GTK_WIDGET (panel));
      break;
    case FALSE:
      gtk_widget_hide_all (GTK_WIDGET (panel));
      break;
    default:
      gtk_widget_hide_all (GTK_WIDGET (panel));
      break;
  }
}

static void
ocr_image_list_action_group_set_sensitivity (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  GtkAction           *action;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  if (priv->activated_item == NULL ||
      !gtk_tree_row_reference_valid (priv->activated_item))
  {
    gtk_action_group_set_sensitive (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
										   1)),
				    FALSE);
    gtk_action_group_set_sensitive (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
										   2)),
				    FALSE);
    action = gtk_action_group_get_action (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
                                                                                         3)),
                                          "SaveMultipleText");
    gtk_action_set_sensitive (action,
                              FALSE);
    action = gtk_action_group_get_action (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
                                                                                         3)),
                                          "CloseImage");
    gtk_action_set_sensitive (action,
                              FALSE);
  }
  else
  {
    gtk_action_group_set_sensitive (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
										   1)),
				    TRUE);
    gtk_action_group_set_sensitive (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
										   2)),
				    TRUE);
    action = gtk_action_group_get_action (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
                                                                                         3)),
                                          "SaveMultipleText");
    gtk_action_set_sensitive (action,
                              TRUE);
    action = gtk_action_group_get_action (GTK_ACTION_GROUP (ocr_window_get_action_group (priv->main_window,
                                                                                         3)),
                                          "CloseImage");
    gtk_action_set_sensitive (action,
                              TRUE);
  }
}

static void
ocr_image_list_check_activated_item (OcrImageList *self,
				     GtkTreeIter  *iter)
{
  OcrImageListPrivate *priv;
  GtkTreePath         *path;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  if (priv->activated_item == NULL ||
      !gtk_tree_row_reference_valid (priv->activated_item))
  {
    path = gtk_tree_model_get_path (GTK_TREE_MODEL (priv->liststore),
				    iter);
    gtk_icon_view_select_path (priv->iconview,
			       path);

    gtk_icon_view_item_activated (priv->iconview,
				  path);
    
    ocr_image_list_action_group_set_sensitivity (self);
  }
}

gboolean
ocr_image_list_add_image (OcrImageList *self,
			  GFile        *file)
{
  OcrImageListPrivate *priv;
  gchar               *file_path;
  gchar               *file_display_name;
  GFileInfo           *file_info;
  GdkPixbuf           *pixbuf;
  GdkPixbuf           *scaled_pixbuf;
  GtkTreeIter          iter;
  gint                 dest_width, dest_height;
  gint                 width, height;
  gint                 width_ratio, height_ratio;
  gint                 ratio;
  OcrPanel            *panel;
  GError              *error = NULL;

  g_return_val_if_fail (OCR_IS_IMAGE_LIST (self), FALSE);
  g_return_val_if_fail (G_IS_FILE (file), FALSE);

  priv = self->priv;

  file_path = g_file_get_path (file);
  pixbuf = gdk_pixbuf_new_from_file (file_path,
				     &error);
  g_free (file_path);

  file_info = g_file_query_info (file,
				 G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				 G_FILE_QUERY_INFO_NONE,
				 NULL,
				 NULL);

  file_display_name = g_strdup (g_file_info_get_display_name (file_info));

  if (error != NULL)
  {
    gchar *title = g_strdup_printf (_("Error loading image «%s»"),
				    file_display_name);

    ocr_utils_message (title,
		       error->message,
		       GTK_MESSAGE_ERROR);

    g_free (title);
    g_object_unref (file_info);
    g_free (file_display_name);
    g_clear_error (&error);
    return FALSE;
  }

  g_assert (GDK_IS_PIXBUF (pixbuf));

  /* Change pixbuf size for thumbnail */
  width = gdk_pixbuf_get_width (pixbuf);
  height = gdk_pixbuf_get_height (pixbuf);
  width_ratio = width / ICON_VIEW_ICON_MAX_WIDTH;
  height_ratio = height / ICON_VIEW_ICON_MAX_HEIGHT;
  if (width_ratio > 1 || height_ratio > 1)
  {
    ratio = MAX (width_ratio,
		 height_ratio);
    dest_width = width / ratio;
    dest_height = height / ratio;

    if (dest_width == 0)
      dest_width = 1;
    if (dest_height == 0)
      dest_height = 1;

    scaled_pixbuf = gdk_pixbuf_scale_simple (pixbuf,
					     dest_width,
					     dest_height,
					     GDK_INTERP_BILINEAR);

    g_assert (GDK_IS_PIXBUF (scaled_pixbuf));
  }
  else
    scaled_pixbuf = gdk_pixbuf_copy (pixbuf);

  panel = g_object_new (OCR_TYPE_PANEL,
			"homogeneous", FALSE,
			"spacing", 0,
			"main-window", priv->main_window,
			"image-file", file,
			NULL);

  g_assert (OCR_IS_PANEL (panel));

  gtk_list_store_insert_with_values (priv->liststore, &iter, G_MAXINT32,
				     COLUMN_FILE, file,
				     COLUMN_PATH, file_display_name,
				     COLUMN_FILE_NAME, g_filename_display_basename (file_display_name),
				     COLUMN_PIXBUF, scaled_pixbuf,
				     COLUMN_PANEL_POINTER, panel,
				     -1);

  ocr_window_add_panel (priv->main_window,
			GTK_WIDGET (panel));

  ocr_image_list_check_activated_item (self, &iter);

  g_object_unref (file_info);
  g_object_unref (pixbuf);

  return TRUE;
}

static void
ocr_image_list_item_activated (GtkIconView *iconview,
			       GtkTreePath *arg1,
			       gpointer     user_data)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  GtkTreePath         *path_activated_item;
  GtkTreeRowReference *rowref;

  g_return_if_fail (GTK_IS_ICON_VIEW (iconview));
  g_return_if_fail (OCR_IS_IMAGE_LIST (user_data));

  self = OCR_IMAGE_LIST (user_data);
  priv = self->priv;

  rowref = gtk_tree_row_reference_new (GTK_TREE_MODEL (priv->liststore),
				       arg1);

  if (priv->activated_item == NULL)
    priv->activated_item = gtk_tree_row_reference_copy (rowref);
  else
  {
    if (!gtk_tree_row_reference_valid (priv->activated_item))
    {
      gtk_tree_row_reference_free (priv->activated_item);
      priv->activated_item = gtk_tree_row_reference_copy (rowref);
    }
    else
    {
      path_activated_item = gtk_tree_row_reference_get_path (priv->activated_item);
      if (gtk_tree_path_compare (arg1, path_activated_item) != 0)
      {
	ocr_image_list_show_panel (self,
				   path_activated_item,
				   FALSE);

	gtk_tree_row_reference_free (priv->activated_item);
	priv->activated_item = gtk_tree_row_reference_copy (rowref);

	gtk_tree_path_free (path_activated_item);
      }
      else
      {
	gtk_tree_path_free (path_activated_item);
	gtk_tree_row_reference_free (rowref);
	return;
      }
    }
  }

  ocr_image_list_show_panel (self,
			     arg1,
			     TRUE);

  gtk_tree_row_reference_free (rowref);
}

static void
ocr_image_list_activate_next_row (OcrImageList *self,
				  GtkTreePath  *path)
{
  GtkTreeRowReference *rowref;
  GtkTreePath         *path_copy;
  GtkTreePath         *path_activated_item;
  OcrImageListPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  rowref = gtk_tree_row_reference_new (GTK_TREE_MODEL (priv->liststore),
				       path);

  path_activated_item = gtk_tree_row_reference_get_path (priv->activated_item);

  if (gtk_tree_path_compare (path, path_activated_item) == 0)
  {
    path_copy = gtk_tree_path_copy (path);

    gtk_tree_path_next (path_copy);

    /* path_copy could be invalid so i need to use a row reference to test validity */
    rowref = gtk_tree_row_reference_new (GTK_TREE_MODEL (priv->liststore),
					 path_copy);

    if (gtk_tree_row_reference_valid (rowref))
    {
      gtk_icon_view_select_path (priv->iconview,
				 path_copy);

      gtk_icon_view_item_activated (priv->iconview,
				    path_copy);
    }
    else
    {
      gtk_tree_path_free (path_copy);
      path_copy = gtk_tree_path_copy (path);

      if (gtk_tree_path_prev (path_copy))
      {
	gtk_icon_view_select_path (priv->iconview,
				   path_copy);

	gtk_icon_view_item_activated (priv->iconview,
				      path_copy);
      }
      else
      {
	/* No other images are in the image list, so logo frame is shown */
	gtk_widget_show_all (ocr_window_get_logo (priv->main_window));
      }
    }
  }
	
  gtk_tree_path_free (path_activated_item);
  gtk_tree_path_free (path_copy);
  gtk_tree_row_reference_free (rowref);
}

gboolean
ocr_image_list_check_selected (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  GList               *selected_items;
  gboolean             retval;

  g_return_val_if_fail (OCR_IS_IMAGE_LIST (self), TRUE);

  priv = self->priv;

  selected_items = gtk_icon_view_get_selected_items (priv->iconview);

  retval = FALSE;
  if (selected_items == NULL)
    retval = TRUE;
  if (selected_items != NULL &&
      g_list_length (selected_items) < 2)
  {
    ocr_utils_message (_("Not enought files selected"),
		       _("Please, select two or more files to save."),
		       GTK_MESSAGE_ERROR);
    retval = TRUE;
  }

  if (selected_items != NULL)
    g_list_free (selected_items);

  return retval;
}

static void
ocr_image_list_check_state_for_selected_1 (gpointer data,
					   gpointer user_data)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  GtkTreePath         *path;
  GtkTreeIter          iter;
  OcrPanel            *panel;
  gchar               *file_display_basename;

  g_return_if_fail (OCR_IS_IMAGE_LIST (user_data));

  self = OCR_IMAGE_LIST (user_data);
  priv = self->priv;

  path = (GtkTreePath *) data;

  gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->liststore),
			   &iter,
			   path);

  gtk_tree_model_get (GTK_TREE_MODEL (priv->liststore),
		      &iter,
		      COLUMN_FILE_NAME, &file_display_basename,
		      COLUMN_PANEL_POINTER, &panel,
		      -1);
  
  if (!OCR_PANEL_OCR_PERFORMED (panel))
    priv->file_list = g_list_append (priv->file_list,
				     g_strdup (file_display_basename));

  gtk_tree_path_free (path);
}

gboolean
ocr_image_list_check_state_for_selected (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  GList               *selected_items;
  GList               *file_list_pointer;
  gchar               *file_string;
  gchar               *message;
  gchar               *temp;
  gboolean             retval;

  g_return_val_if_fail (OCR_IS_IMAGE_LIST (self), TRUE);

  priv = self->priv;

  selected_items = gtk_icon_view_get_selected_items (priv->iconview);

  if (selected_items == NULL)
    return FALSE;

  if (priv->file_list != NULL)
    ocr_image_list_free_file_list (self);

  g_list_foreach (selected_items,
		  ocr_image_list_check_state_for_selected_1,
		  self);

  g_list_free (selected_items);

  retval = FALSE;
  if (priv->file_list != NULL)
  {
    message = g_strdup (_("Text not yet extracted for the following images:"));

    file_list_pointer = g_list_first (priv->file_list);
    while (file_list_pointer)
    {
      file_string = g_strdup_printf ("«%s»", (gchar*)(file_list_pointer->data));
      temp = g_strjoin ("\n", message, file_string, NULL);
      g_free (file_string);
      g_free (message);
      message = g_strdup (temp);
      g_free (temp);

      file_list_pointer = file_list_pointer->next;
    }

    ocr_utils_message (_("No text extracted"),
		       message,
		       GTK_MESSAGE_ERROR);

    g_free (message);
    ocr_image_list_free_file_list (self);
    retval = TRUE;
  }

  return retval;
}

void
ocr_image_list_save_multiple_text (OcrImageList *self,
				   gchar        *uri)
{
  OcrImageListPrivate *priv;
  gchar               *filename;
  gchar               *separator;
  GList               *selected_items;
  GList               *selected_items_pointer;
  GtkTreeIter          iter;
  OcrPanel            *panel;
  GtkTextBuffer       *textbuffer;
  gchar               *content = NULL;
  gchar               *single_file_content;
  gchar               *temp = NULL;
  gboolean             retval;
  GError              *error;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));
  g_return_if_fail (uri != NULL);

  priv = self->priv;

  error = NULL;
  filename = g_filename_from_uri (uri, NULL, &error);

  if (error != NULL)
  {
    g_critical (error->message,
		NULL);
    g_clear_error (&error);
    return;
  }

  selected_items = gtk_icon_view_get_selected_items (priv->iconview);

  if (selected_items == NULL)
  {
    g_free (filename);
    g_list_free (selected_items);
    return;
  }

  separator = ocr_preferences_get_separator (OCR_PREFERENCES (ocr_window_get_preferences (priv->main_window)));

  selected_items_pointer = g_list_first (selected_items);
  while (selected_items_pointer)
  {
    gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->liststore),
			     &iter,
			     selected_items_pointer->data);

    gtk_tree_model_get (GTK_TREE_MODEL (priv->liststore),
			&iter,
			COLUMN_PANEL_POINTER, &panel,
			-1);
    
    textbuffer = GTK_TEXT_BUFFER (ocr_panel_get_textbuffer (panel));
    single_file_content = ocr_textbuffer_get_content (OCR_TEXTBUFFER (textbuffer)); 

    if (single_file_content != NULL)
    {
      single_file_content = ocr_modify_content (panel,
						single_file_content);

      if (content == NULL)
	content = g_strdup (single_file_content);
      else
      {
	temp = g_strdup (content);
	g_free (content);
	content = g_strjoin (separator, temp, single_file_content, NULL);
	g_free (temp);
      }
      g_free (single_file_content);
    }

    selected_items_pointer = selected_items_pointer->next;
  }

  g_list_free (selected_items);
  
  if (content != NULL)
  {
    retval = ocr_utils_write_file (filename,
				   content);

    if (retval)
      gtk_text_buffer_set_modified (textbuffer,
				    FALSE);

    g_free (content);
  }
}

static void
ocr_image_list_set_reference (gpointer data,
			      gpointer user_data)
{
  GtkTreePath         *path;
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  GtkTreeRowReference *rowref;

  g_return_if_fail (OCR_IS_IMAGE_LIST (user_data));

  self = OCR_IMAGE_LIST (user_data);
  priv = self->priv;

  path = (GtkTreePath *) data;

  rowref = gtk_tree_row_reference_new (GTK_TREE_MODEL (priv->liststore),
				       path);

  priv->reference = g_list_append (priv->reference,
				   rowref);

  gtk_tree_path_free (path);
}


static void
ocr_image_list_close_image (gpointer data,
			    gpointer user_data)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  GtkTreeRowReference *rowref;
  GtkTreePath         *path;
  GtkTreeIter          iter;
  GFile               *col_file;
  gchar               *col_path;
  gchar               *col_file_name;
  GdkPixbuf           *col_pixbuf;
  OcrPanel            *col_panel;
  gboolean             retval;

  g_return_if_fail (OCR_IS_IMAGE_LIST (user_data));

  self = OCR_IMAGE_LIST (user_data);
  priv = self->priv;

  rowref = (GtkTreeRowReference *) data;

  g_return_if_fail (gtk_tree_row_reference_valid (rowref));

  path = gtk_tree_row_reference_get_path (rowref);

  if (gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->liststore),
			       &iter,
			       path))
  {
    gtk_tree_model_get (GTK_TREE_MODEL (priv->liststore),
			&iter,
			COLUMN_FILE, &col_file,
			COLUMN_PATH, &col_path,
			COLUMN_FILE_NAME, &col_file_name,
			COLUMN_PIXBUF, &col_pixbuf,
			COLUMN_PANEL_POINTER, &col_panel,
			-1);

    if (OCR_PANEL_TEXT_BUFFER_MODIFIED (col_panel))
    {
      gchar *message = g_strdup_printf (_("Text has been modified for image «%s».\nDo you want to close the file without saving?"),
					col_file_name);

      retval = ocr_utils_question (_("Text modified"),
				   message);

      g_free (message);
      if (!retval)
	goto out;
    }

    ocr_image_list_activate_next_row (self, path);

    g_object_unref (col_file);
    g_free (col_path);
    g_free (col_file_name);
    g_object_unref (col_pixbuf);

    ocr_panel_destroy_selection_font_dialog (col_panel);
    ocr_panel_destroy_save_text_dialog (col_panel);

    gtk_widget_destroy (GTK_WIDGET (col_panel));

    gtk_list_store_remove (priv->liststore,
			   &iter);
  }

out:
  gtk_tree_row_reference_free (rowref);
  gtk_tree_path_free (path);
}

void
ocr_image_list_close_images (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  GList               *selected_items;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  selected_items = gtk_icon_view_get_selected_items (priv->iconview);

  if (selected_items == NULL)
    return;

  g_list_foreach (selected_items,
		  ocr_image_list_set_reference,
		  self);

  g_list_free (selected_items);

  if (priv->reference != NULL)
  {
    g_list_foreach (priv->reference,
		    ocr_image_list_close_image,
		    self);

    g_list_free (priv->reference);
    priv->reference = NULL;
  }
}

static void
ocr_image_list_row_deleted (GtkTreeModel *tree_model,
			    GtkTreePath  *path,
			    gpointer      user_data)
{
  OcrImageList *self;

  g_return_if_fail (OCR_IS_IMAGE_LIST (user_data));

  self = OCR_IMAGE_LIST (user_data);

  ocr_image_list_action_group_set_sensitivity (self);
}

/* TODO This function set an emblemed icon for the SaveMultipleText action
   NOTE - gtk_action_set_gicon is in GTK+ toolkit since 2.16 
static void
ocr_imagelist_set_SaveMultipleText_icon (OcrImageList *self)
{
  OcrImageListPrivate *priv;
  GtkUIManager        *ui_mgr;
  GtkIconTheme        *icon_theme;
  GtkIconInfo         *icon_info;
  GtkAction           *action;
  const gchar         *filename;
  GFile               *icon_file, *emblem_file;
  GIcon               *icon, *emblem_icon, *final_icon;
  GEmblem             *emblem;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;
  _________________________________________________________________  
  Comment this piece of code...

  icon_theme = gtk_icon_theme_get_default ();
  icon_info = gtk_icon_theme_lookup_icon (icon_theme,
					  GTK_STOCK_SAVE,
					  22,//gtk_toolbar_get_icon_size (GTK_IS_TOOLBAR (priv->toolbar_image)),
					  GTK_ICON_LOOKUP_USE_BUILTIN);

  filename = gtk_icon_info_get_filename (icon_info);
  if (filename == NULL)
    return;

  icon_file = g_file_new_for_path (filename);
  emblem_file = g_file_new_for_path ("./immagini_test/emblem-important.png");
  icon = g_file_icon_new (icon_file);
  emblem_icon = g_file_icon_new (emblem_file); 
  emblem = g_emblem_new (emblem_icon);
  final_icon = g_emblemed_icon_new (icon,
				    emblem);
  ui_mgr = GTK_UI_MANAGER (ocr_window_get_ui_manager (priv->main_window));
  action = gtk_ui_manager_get_action (ui_mgr,
				      "/ImageListToolBar/SaveMultipleText");
  gtk_action_set_gicon (action,
			icon);
  _________________________________________________________________  

  emblem_icon = g_themed_icon_new (GTK_STOCK_SAVE);
  emblem = g_emblem_new (emblem_icon);
  final_icon = g_emblemed_icon_new (icon,
				    emblem);
  ui_mgr = GTK_UI_MANAGER (ocr_window_get_ui_manager (priv->main_window));
  action = gtk_ui_manager_get_action (ui_mgr,
				      "/ImageListToolBar/SaveMultipleText");
  gtk_action_set_gicon (action,
			final_icon);
}
*/

static void
ocr_image_list_constructed (GObject *gobject)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;
  GtkUIManager        *ui_mgr;
  GtkWidget           *scrolledwindow;

  g_return_if_fail (OCR_IS_IMAGE_LIST (gobject));

  self = OCR_IMAGE_LIST (gobject);
  priv = self->priv;

  ui_mgr = GTK_UI_MANAGER (ocr_window_get_ui_manager (priv->main_window));
  priv->toolbar = gtk_ui_manager_get_widget (ui_mgr,
					     "/ImageListToolBar");
  g_assert (GTK_IS_TOOLBAR (priv->toolbar));

  //ocr_imagelist_set_SaveMultipleText_icon (self);

  priv->liststore = gtk_list_store_new (N_COLUMNS,
					G_TYPE_FILE,
					G_TYPE_STRING,
					G_TYPE_STRING,
					GDK_TYPE_PIXBUF,
					G_TYPE_INT,
					G_TYPE_POINTER);

  g_signal_connect (G_OBJECT (priv->liststore),
		    "row-deleted",
		    G_CALLBACK (ocr_image_list_row_deleted),
	    	    self);

  scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow),
				       GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow),
				  GTK_POLICY_NEVER,
				  GTK_POLICY_ALWAYS);
  gtk_widget_set_size_request (GTK_WIDGET (scrolledwindow),
			       ICON_VIEW_ICON_MAX_WIDTH + 35,
			       400);

  priv->iconview = GTK_ICON_VIEW (gtk_icon_view_new_with_model (GTK_TREE_MODEL (priv->liststore)));
  gtk_icon_view_set_markup_column (priv->iconview, COLUMN_FILE_NAME);
  gtk_icon_view_set_pixbuf_column (priv->iconview, COLUMN_PIXBUF);
  gtk_icon_view_set_selection_mode (priv->iconview, GTK_SELECTION_MULTIPLE);
  gtk_icon_view_set_tooltip_column (priv->iconview, COLUMN_PATH);
  gtk_icon_view_set_columns (priv->iconview, 1);
  gtk_icon_view_set_item_width (priv->iconview, ICON_VIEW_ICON_MAX_WIDTH);
  gtk_icon_view_set_margin (priv->iconview, 6);

  g_signal_connect (G_OBJECT (priv->iconview),
		    "item-activated",
		    G_CALLBACK (ocr_image_list_item_activated),
	    	    self);

  gtk_box_pack_start (GTK_BOX (self), GTK_WIDGET (priv->toolbar), FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (scrolledwindow), GTK_WIDGET (priv->iconview));
  gtk_box_pack_start (GTK_BOX (self), GTK_WIDGET (scrolledwindow), TRUE, TRUE, 0);

  ocr_image_list_action_group_set_sensitivity (self);
}

static void
ocr_image_list_set_property (GObject *gobject,
			     guint property_id,
			     const GValue *value,
			     GParamSpec *pspec)
{
  OcrImageList        *self;
  OcrImageListPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGE_LIST (gobject));

  self = OCR_IMAGE_LIST (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_MAIN_WINDOW:
      g_return_if_fail (OCR_IS_WINDOW (g_value_get_object (value)));
      priv->main_window = OCR_WINDOW (g_value_dup_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_image_list_class_init (OcrImageListClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrImageListPrivate));

  gobject_class->constructed = ocr_image_list_constructed;
  gobject_class->set_property = ocr_image_list_set_property;
  gobject_class->dispose = ocr_image_list_dispose;
  gobject_class->finalize = ocr_image_list_finalize;

  g_object_class_install_property (gobject_class, PROP_MAIN_WINDOW,
				   g_param_spec_object ("main-window", "main-window", "main-window",
							OCR_TYPE_WINDOW,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_image_list_init (OcrImageList *self)
{
  OcrImageListPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  self->priv = OCR_IMAGE_LIST_GET_PRIVATE (self);
  priv = self->priv;
  
  priv->activated_item = NULL;
}

static gboolean
ocr_image_list_set_panel_font (GtkTreeModel *treemodel,
			       GtkTreePath  *path,
			       GtkTreeIter  *iter,
			       gpointer      data)
{
  OcrPanel    *panel;
  const gchar *font_name;

  g_return_val_if_fail (GTK_TREE_MODEL (treemodel), TRUE);

  font_name = (const gchar *) data;

  gtk_tree_model_get (GTK_TREE_MODEL (treemodel),
		      iter,
		      COLUMN_PANEL_POINTER, &panel,
		      -1);

  ocr_panel_set_font (panel,
		      font_name);

  return FALSE;
}

void
ocr_image_list_set_font (OcrImageList *self,
			 const gchar  *font_name)
{
  OcrImageListPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  gtk_tree_model_foreach (GTK_TREE_MODEL (priv->liststore),
			  ocr_image_list_set_panel_font,
			  (gpointer) font_name);
}

static gboolean
ocr_image_list_change_panel_layout (GtkTreeModel *treemodel,
				    GtkTreePath  *path,
				    GtkTreeIter  *iter,
				    gpointer      data)
{
  OcrPanel             *panel;
  OcrPreferencesLayout  layout;

  g_return_val_if_fail (GTK_TREE_MODEL (treemodel), TRUE);

  layout = GPOINTER_TO_INT (data);

  gtk_tree_model_get (GTK_TREE_MODEL (treemodel),
		      iter,
		      COLUMN_PANEL_POINTER, &panel,
		      -1);

  ocr_panel_change_layout (panel,
			   layout);

  return FALSE;
}

void
ocr_image_list_change_layout (OcrImageList         *self,
			      OcrPreferencesLayout  layout)
{
  OcrImageListPrivate *priv;
  GtkWidget           *panel;

  g_return_if_fail (OCR_IS_IMAGE_LIST (self));

  priv = self->priv;

  panel = ocr_image_list_get_active_panel (self);
  if (panel != NULL)
  {
    gtk_tree_model_foreach (GTK_TREE_MODEL (priv->liststore),
			    ocr_image_list_change_panel_layout,
			    GINT_TO_POINTER (layout));

    ocr_panel_show_paned (OCR_PANEL (panel));
  }
}
