#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-imageview.h"
#include "ocr-define.h"
#include "ocr-utils.h"

#define OCR_IMAGEVIEW_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_IMAGEVIEW, OcrImageViewPrivate))

G_DEFINE_TYPE (OcrImageView, ocr_imageview, GTK_TYPE_TABLE);

typedef enum
{
  OCR_NO_BUTTON,
  OCR_LEFT_BUTTON,
  OCR_MIDDLE_BUTTON,
  OCR_RIGHT_BUTTON
} OcrMouseButton;

enum
{
  PROP_IMAGE_FILE = 1,
};

struct _OcrImageViewPrivate
{
  GFile         *image_file;
  GtkAdjustment *hadjustment;
  GtkAdjustment *vadjustment;
  GtkWidget     *scrolledwindow;
  GtkWidget     *viewport;
  GdkPixbuf     *pixbuf;
  GtkWidget     *image;
  gdouble        x, y;
  gdouble        x_max, y_max;
  gint           press_value_x, press_value_y;
  gboolean       hdrag, vdrag;
  OcrCursor      cursor;
  OcrMouseButton mousebutton;
};

static void
ocr_imageview_dispose (GObject *gobject)
{
  OcrImageView        *self;
  OcrImageViewPrivate *priv;
  static gboolean first_run = TRUE;

  g_return_if_fail (OCR_IS_IMAGEVIEW (gobject));

  self = OCR_IMAGEVIEW (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->pixbuf != NULL)
      g_object_unref (priv->pixbuf);

    if (priv->image_file != NULL)
      g_object_unref (priv->image_file);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_imageview_parent_class)->dispose (gobject);
}

static void
ocr_imageview_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_IMAGEVIEW (gobject));

  G_OBJECT_CLASS (ocr_imageview_parent_class)->finalize (gobject);
}

static void
ocr_imageview_set_cursor (GtkWidget    *widget,
			  OcrCursor     new_cursor,
			  OcrImageView *imageview)
{
  OcrImageViewPrivate *priv;
  GdkCursor           *cursor = NULL;

  g_return_if_fail (OCR_IS_IMAGEVIEW (imageview));

  priv = imageview->priv;

  if (priv->cursor == new_cursor)
    return;

  switch (new_cursor)
  {
    case OCR_CURSOR_NORMAL:
      gdk_window_set_cursor (widget->window,
			     NULL);
      priv->cursor = new_cursor;
      break;
    case OCR_CURSOR_DRAG:
      cursor = gdk_cursor_new (GDK_FLEUR);
      break;
    default:
      gdk_window_set_cursor (widget->window,
			     NULL);
      priv->cursor = OCR_CURSOR_NORMAL;
  }

  if (cursor) 
  {
    gdk_window_set_cursor (widget->window, cursor);
    gdk_cursor_unref (cursor);
    gdk_flush();
    priv->cursor = new_cursor;
  }
}

static gboolean
ocr_imageview_viewport_button_press_event (GtkWidget      *widget,
					   GdkEventButton *eventbutton,
					   gpointer        user_data)
{
  OcrImageView *self = OCR_IMAGEVIEW (user_data);
  OcrImageViewPrivate *priv;
  gint viewport_width, viewport_height;
  gint image_width, image_height;

  g_return_val_if_fail (OCR_IS_IMAGEVIEW (self), TRUE);

  priv = self->priv;

  if (eventbutton->button == 1 &&
      eventbutton->type == GDK_BUTTON_PRESS) /* this is the mouse left button */
  {
    priv->mousebutton = OCR_LEFT_BUTTON;

    ocr_imageview_set_cursor (widget,
			      OCR_CURSOR_DRAG,
			      self);

    viewport_width = GTK_WIDGET (priv->viewport)->allocation.width;
    viewport_height = GTK_WIDGET (priv->viewport)->allocation.height;
    image_width = gdk_pixbuf_get_width (priv->pixbuf);
    image_height = gdk_pixbuf_get_height (priv->pixbuf);

    GValue value = { 0, };
    g_value_init (&value, G_TYPE_DOUBLE);
    GValue upper = { 0, };
    g_value_init (&upper, G_TYPE_DOUBLE);
    
    if (viewport_width > image_width)
    {
      priv->hdrag = FALSE;
    }
    else
    {
      priv->hdrag = TRUE;
      priv->x = eventbutton->x;
      g_object_get_property (G_OBJECT (priv->hadjustment),
			     "value",
			     &value);
      priv->press_value_x = (gint) g_value_get_double (&value);
      g_object_get_property (G_OBJECT (priv->hadjustment),
			     "upper",
			     &upper);
      priv->x_max = g_value_get_double (&upper); 
      g_object_get_property (G_OBJECT (priv->hadjustment),
			     "page-size",
			     &value);
      priv->x_max -= g_value_get_double (&value);
    }

    if (viewport_height > image_height)
    {
      priv->vdrag = FALSE;
    }
    else
    {
      priv->vdrag = TRUE;
      priv->y = eventbutton->y;
      g_object_get_property (G_OBJECT (priv->vadjustment),
			     "value",
			     &value);
      priv->press_value_y = (gint) g_value_get_double (&value);
      g_object_get_property (G_OBJECT (priv->vadjustment),
			     "upper",
			     &upper);
      priv->y_max = g_value_get_double (&upper); 
      g_object_get_property (G_OBJECT (priv->vadjustment),
			     "page-size",
			     &value);
      priv->y_max -= g_value_get_double (&value);
    }

    g_value_unset (&value);
    g_value_unset (&upper);
  }

  return FALSE;
}

static gboolean
ocr_imageview_viewport_motion_notify_event (GtkWidget      *widget,
					    GdkEventMotion *eventmotion,
					    gpointer        user_data)
{
  OcrImageView *self = OCR_IMAGEVIEW (user_data);
  OcrImageViewPrivate *priv = self->priv;
  gint dx, dy;
  gdouble x, y;

  g_return_val_if_fail (OCR_IS_IMAGEVIEW (self), TRUE);

  if (priv->mousebutton == OCR_LEFT_BUTTON) 
  {
    dx = priv->x - eventmotion->x;
    dy = priv->y - eventmotion->y;
    
    if (priv->hdrag)
    {
      x = CLAMP (priv->press_value_x + dx, 0, priv->x_max);
      gtk_adjustment_set_value (priv->hadjustment, x);
    }
    if (priv->vdrag)
    {
      y = CLAMP (priv->press_value_y + dy, 0, priv->y_max);
      gtk_adjustment_set_value (priv->vadjustment, y);
    }
  }

  return FALSE;
}

static gboolean
ocr_imageview_viewport_button_release_event (GtkWidget      *widget,
					     GdkEventButton *eventbutton,
					     gpointer        user_data)
{
  OcrImageView *self = OCR_IMAGEVIEW (user_data);
  OcrImageViewPrivate *priv = self->priv;

  g_return_val_if_fail (OCR_IS_IMAGEVIEW (self), TRUE);
  
  if (eventbutton->button == 1 &&
      eventbutton->type == GDK_BUTTON_RELEASE) /* this is the mouse left button */
  {
    priv->mousebutton = OCR_NO_BUTTON;

    ocr_imageview_set_cursor (widget,
			      OCR_CURSOR_NORMAL,
			      self);
  }

  return FALSE;
}

static void
ocr_imageview_load_pixbuf (OcrImageView *self)
{
  OcrImageViewPrivate *priv;
  GFileInfo           *file_info;
  gchar               *file_display_name;
  gchar               *file_path;
  GError              *error = NULL;

  g_return_if_fail (OCR_IS_IMAGEVIEW (self));

  priv = self->priv;

  gtk_image_clear (GTK_IMAGE (priv->image));
  if (priv->pixbuf)
    g_object_unref (priv->pixbuf);

  file_path = g_file_get_path (priv->image_file);
  priv->pixbuf = gdk_pixbuf_new_from_file (file_path,
					   &error);

  if (error != NULL)
  {
    file_info = g_file_query_info (priv->image_file,
				   G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				   G_FILE_QUERY_INFO_NONE,
				   NULL,
				   NULL);

    file_display_name = g_strdup (g_file_info_get_display_name (file_info));

    gchar *title = g_strdup_printf (_("Error loading image «%s»"),
				    file_display_name);

    ocr_utils_message (title,
		       error->message,
		       GTK_MESSAGE_ERROR);

    g_free (file_path);
    g_free (title);
    g_clear_error (&error);
    return;
  }

  g_free (file_path);

  g_assert (GDK_IS_PIXBUF (priv->pixbuf));

  gtk_image_set_from_pixbuf (GTK_IMAGE (priv->image),
			     priv->pixbuf);
}

static void
ocr_imageview_constructed (GObject *gobject)
{
  OcrImageView        *self;
  OcrImageViewPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGEVIEW (gobject));

  self = OCR_IMAGEVIEW (gobject);
  priv = self->priv;

  priv->image = gtk_image_new ();
  if (priv->image_file)
    ocr_imageview_load_pixbuf (self);

  priv->scrolledwindow = gtk_scrolled_window_new (NULL, NULL);

  /* I don't use gtk_scrolled_window_with_viewport, but priv->viewport,
     to set event mask for viewport */
  priv->viewport = gtk_viewport_new (priv->hadjustment, priv->vadjustment);
  gtk_widget_add_events (priv->viewport, GDK_BUTTON_MOTION_MASK | GDK_BUTTON_RELEASE_MASK);

  gtk_container_add (GTK_CONTAINER (priv->viewport), priv->image);
  gtk_container_add (GTK_CONTAINER (priv->scrolledwindow), priv->viewport);

  priv->hadjustment = gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (priv->scrolledwindow));
  priv->vadjustment = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->scrolledwindow));
  
  g_signal_connect (GTK_WIDGET (priv->viewport), 
		    "button-press-event",
		    G_CALLBACK (ocr_imageview_viewport_button_press_event),
		    self);
  g_signal_connect (GTK_WIDGET (priv->viewport), 
		    "motion-notify-event",
		    G_CALLBACK (ocr_imageview_viewport_motion_notify_event),
		    self);
  g_signal_connect (GTK_WIDGET (priv->viewport), 
		    "button-release-event",
		    G_CALLBACK (ocr_imageview_viewport_button_release_event),
		    self);

  gtk_table_attach (GTK_TABLE (self),
		    priv->scrolledwindow,
		    0, 1, 0, 1,
		    GTK_EXPAND | GTK_FILL | GTK_SHRINK,
		    GTK_EXPAND | GTK_FILL | GTK_SHRINK,
		    0, 0);
}

static void
ocr_imageview_set_property (GObject *gobject,
			    guint property_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
  OcrImageView *self;
  OcrImageViewPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGEVIEW (gobject));

  self = OCR_IMAGEVIEW (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_IMAGE_FILE:
      g_assert (G_IS_FILE (g_value_get_object (value)));
      priv->image_file = g_value_dup_object (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_imageview_class_init (OcrImageViewClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrImageViewPrivate));

  gobject_class->constructed = ocr_imageview_constructed;
  gobject_class->set_property = ocr_imageview_set_property;
  gobject_class->dispose = ocr_imageview_dispose;
  gobject_class->finalize = ocr_imageview_finalize;
  
  g_object_class_install_property (gobject_class, PROP_IMAGE_FILE,
				   g_param_spec_object ("image-file", "image-file", "image-file",
							G_TYPE_FILE,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_imageview_init (OcrImageView *self)
{
  OcrImageViewPrivate *priv;

  g_return_if_fail (OCR_IS_IMAGEVIEW (self));

  self->priv = OCR_IMAGEVIEW_GET_PRIVATE (self);
  priv = self->priv;

  priv->image_file     = NULL;
  priv->hadjustment    = NULL;
  priv->vadjustment    = NULL;
  priv->scrolledwindow = NULL;
  priv->viewport       = NULL;
  priv->pixbuf         = NULL;
  priv->image          = NULL;
  priv->cursor         = OCR_CURSOR_NORMAL;
}

gdouble
ocr_imageview_get_fit_image_percentage (OcrImageView *self)
{
  OcrImageViewPrivate *priv;
  gint                 image_width, image_height;
  gdouble              h_page_size, v_page_size;
  gdouble              h_percentage, v_percentage;

  g_return_val_if_fail (OCR_IS_IMAGEVIEW (self), 0);

  priv = self->priv;

  image_width = gdk_pixbuf_get_width (priv->pixbuf);
  image_height = gdk_pixbuf_get_height (priv->pixbuf);
  h_page_size = priv->hadjustment->page_size; 
  v_page_size = priv->vadjustment->page_size; 

  h_percentage = (h_page_size * 100) / image_width;
  v_percentage = (v_page_size * 100) / image_height;

  return MIN (h_percentage, v_percentage);
}

void
ocr_imageview_zoom_image (OcrImageView *self,
			  gdouble       percentage)
{
  OcrImageViewPrivate *priv;
  GdkPixbuf           *pixbuf;
  gint                 image_width, image_height;

  g_return_if_fail (OCR_IS_IMAGEVIEW (self));
  g_return_if_fail (percentage > 0);

  priv = self->priv;

  image_width = gdk_pixbuf_get_width (priv->pixbuf);
  image_height = gdk_pixbuf_get_height (priv->pixbuf);

  image_width = (image_width * percentage) / 100;
  if (image_width == 0)
    image_width = 1;

  image_height = (image_height * percentage) / 100;
  if (image_height == 0)
    image_height = 1;

  pixbuf = gdk_pixbuf_scale_simple (priv->pixbuf,
				    image_width,
				    image_height,
				    GDK_INTERP_BILINEAR);

  gtk_image_clear (GTK_IMAGE (priv->image));
  gtk_image_set_from_pixbuf (GTK_IMAGE (priv->image), pixbuf);

  g_object_unref (pixbuf);
}

