#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-panel.h"
#include "ocr-define.h"
#include "ocr-window.h"
#include "ocr-ui-panel.h"
#include "ocr-imageview.h"
#include "ocr-textbuffer.h"
#include "ocr-textview.h"

#define OCR_PANEL_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_PANEL, OcrPanelPrivate))

G_DEFINE_TYPE (OcrPanel, ocr_panel, GTK_TYPE_VBOX);

enum
{
  PROP_MAIN_WINDOW = 1,
  PROP_IMAGE_FILE
};

struct _OcrPanelPrivate
{
  OcrWindow      *main_window;
  GtkUIManager   *ui_mgr;
  GtkActionGroup *actions_image;
  GtkActionGroup *actions_text;
  GtkWidget      *toolbar_image;
  GtkWidget      *toolbar_text;
  GtkWidget      *image_zoom;
  GtkWidget      *font_size;
  GtkWidget      *imageview;
  OcrTextBuffer  *textbuffer;
  GtkWidget      *textview;
  GtkWidget      *vbox_image;
  GtkWidget      *vbox_text;
  GtkWidget      *h_v_paned;
  GFile          *image_file;
};

GtkWidget *
ocr_panel_get_main_window (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return GTK_WIDGET (priv->main_window);
}

GtkWidget *
ocr_panel_get_imageview (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return priv->imageview;
}

GtkWidget *
ocr_panel_get_textview (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return priv->textview;
}

GObject *
ocr_panel_get_textbuffer (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->textbuffer);
}

GtkWidget *
ocr_panel_get_image_zoom (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return priv->image_zoom;
}

GObject *
ocr_panel_get_image_file (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->image_file);
}

gchar *
ocr_panel_get_image_file_display_name (OcrPanel *self)
{
  OcrPanelPrivate *priv;
  GFileInfo       *file_info;
  gchar           *file_display_name;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  file_info = g_file_query_info (priv->image_file,
				 G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
				 G_FILE_QUERY_INFO_NONE,
				 NULL,
				 NULL);

  file_display_name = g_strdup (g_file_info_get_display_name (file_info));

  g_object_unref (file_info);

  return file_display_name;
}

GtkWidget *
ocr_panel_get_font_size (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_val_if_fail (OCR_IS_PANEL (self), NULL);

  priv = self->priv;

  return priv->font_size;
}

static void
ocr_panel_dispose (GObject *gobject)
{
  OcrPanel        *self;
  OcrPanelPrivate *priv;
  static gboolean first_run = TRUE;

  g_return_if_fail (OCR_IS_PANEL (gobject));

  self = OCR_PANEL (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->main_window != NULL)
      g_object_unref (priv->main_window);

    if (priv->image_file != NULL)
      g_object_unref (priv->image_file);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_panel_parent_class)->dispose (gobject);
}

static void
ocr_panel_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_PANEL (gobject));

  G_OBJECT_CLASS (ocr_panel_parent_class)->finalize (gobject);
}

static void
ocr_panel_image_zoom_spinbutton_value_changed (GtkSpinButton *spinbutton,
					       gpointer       user_data)
{
  OcrPanel        *self;
  OcrPanelPrivate *priv;

  g_return_if_fail (GTK_IS_SPIN_BUTTON (spinbutton));
  g_return_if_fail (OCR_IS_PANEL (user_data));

  self = OCR_PANEL (user_data);
  priv = self->priv;

  ocr_imageview_zoom_image (OCR_IMAGEVIEW (priv->imageview),
			    gtk_spin_button_get_value (spinbutton));
}

void
ocr_panel_set_font (OcrPanel    *self,
		    const gchar *font_name)
{
  OcrPanelPrivate      *priv;
  PangoFontDescription *font_description;
  gdouble               font_size;

  g_return_if_fail (OCR_IS_PANEL (self));
  g_return_if_fail (font_name != NULL);

  priv = self->priv;

  font_description = pango_font_description_from_string (font_name);

  gtk_widget_modify_font (priv->textview,
			  font_description);

  font_size = (gdouble) pango_font_description_get_size (font_description);
  font_size /= PANGO_SCALE;
  font_size = CLAMP (font_size, FONT_SIZE_MIN, FONT_SIZE_MAX);

  gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->font_size),
			     font_size);

  pango_font_description_free (font_description);
}

static void
ocr_panel_font_size_spinbutton_value_changed (GtkSpinButton *spinbutton,
					      gpointer       user_data)
{
  OcrPanel             *self;
  OcrPanelPrivate      *priv;
  GtkStyle             *style;
  PangoFontDescription *font_description;
  gint                  font_size;

  g_return_if_fail (GTK_IS_SPIN_BUTTON (spinbutton));
  g_return_if_fail (OCR_IS_PANEL (user_data));

  self = OCR_PANEL (user_data);
  priv = self->priv;

  g_return_if_fail (OCR_IS_TEXTVIEW (priv->textview));
  
  /* Don't unref style! */
  style = gtk_widget_get_style (priv->textview);
  font_description = pango_font_description_copy (style->font_desc);
  font_size = (gint) gtk_spin_button_get_value (spinbutton);
  pango_font_description_set_size (font_description, font_size * PANGO_SCALE);

  gtk_widget_modify_font (priv->textview,
			  font_description);

  pango_font_description_free (font_description);
}

static void
ocr_panel_textbuffer_modified_changed (GtkTextBuffer *textbuffer,
				       gpointer       user_data)
{
  OcrPanel        *self;
  OcrPanelPrivate *priv;

  g_return_if_fail (GTK_IS_TEXT_BUFFER (textbuffer));
  g_return_if_fail (OCR_IS_PANEL (user_data));

  self = OCR_PANEL (user_data);
  priv = self->priv;

  if (gtk_text_buffer_get_modified (textbuffer))
    OCR_PANEL_SET_STATE (self,
			 OCR_PANEL_STATE_TEXT_BUFFER_MODIFIED);
  else
    OCR_PANEL_UNSET_STATE (self,
			   OCR_PANEL_STATE_TEXT_BUFFER_MODIFIED);
}

void
ocr_panel_change_layout (OcrPanel             *self,
			 OcrPreferencesLayout  layout)
{
  OcrPanelPrivate *priv;
  GtkWidget       *paned;

  g_return_if_fail (OCR_IS_PANEL (self));

  priv = self->priv;

  if (layout == OCR_PREFERENCES_LAYOUT_V)
    paned = gtk_hpaned_new ();
  else
    paned = gtk_vpaned_new ();

  if (GTK_IS_WIDGET (priv->h_v_paned))
  {
    gtk_widget_reparent (priv->vbox_image,
			 paned);
    gtk_widget_reparent (priv->vbox_text,
			 paned);
    gtk_widget_destroy (priv->h_v_paned);
  }

  priv->h_v_paned = paned;

  gtk_paned_pack1 (GTK_PANED (priv->h_v_paned),
		   priv->vbox_image,
		   FALSE,
		   TRUE);
  gtk_paned_pack2 (GTK_PANED (priv->h_v_paned),
		   priv->vbox_text,
		   FALSE,
		   TRUE);

  gtk_box_pack_start (GTK_BOX (self),
		      priv->h_v_paned,
		      TRUE, 
		      TRUE,
		      0);
}

void
ocr_panel_show_paned (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_if_fail (OCR_IS_PANEL (self));

  priv = self->priv;

  gtk_widget_show_all (priv->h_v_paned);
}

static void
ocr_panel_constructed (GObject *gobject)
{
  OcrPanel        *self;
  OcrPanelPrivate *priv;
  GtkToolItem     *toolitem;
  GtkWidget       *scrolled_window_text;
  OcrPreferences  *preferences;
  GError          *error = NULL;

  g_return_if_fail (OCR_IS_PANEL (gobject));

  self = OCR_PANEL (gobject);
  priv = self->priv;
 
  priv->ui_mgr = gtk_ui_manager_new ();
  
  priv->actions_image = gtk_action_group_new ("ActionsImage");
  gtk_action_group_add_actions (priv->actions_image,
				action_entries_image,
				G_N_ELEMENTS (action_entries_image),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_image, 0);

  priv->actions_text = gtk_action_group_new ("ActionsText");
  gtk_action_group_add_actions (priv->actions_text,
				action_entries_text,
				G_N_ELEMENTS (action_entries_text),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_text, 0);
  
  if (!gtk_ui_manager_add_ui_from_file (priv->ui_mgr,
					OCR_UI_DIR"/ocr-ui-panel.xml",
					&error))
  {
    g_error ("Building menus failed: %s",
	     error->message);
    g_clear_error (&error);
  }
  
  priv->toolbar_image = gtk_ui_manager_get_widget (priv->ui_mgr, "/ImageToolBar");
  g_assert (GTK_IS_TOOLBAR (priv->toolbar_image));

  priv->image_zoom = gtk_spin_button_new_with_range (IMAGE_ZOOM_MIN, IMAGE_ZOOM_MAX, 1);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->image_zoom), 100);
  gtk_widget_set_tooltip_text (priv->image_zoom, _("Zoom percentage"));
  toolitem = gtk_tool_item_new ();
  gtk_container_add (GTK_CONTAINER (toolitem), priv->image_zoom);
  gtk_toolbar_insert (GTK_TOOLBAR (priv->toolbar_image), toolitem, 4);

  g_signal_connect (G_OBJECT (priv->image_zoom), "value-changed",
		    G_CALLBACK (ocr_panel_image_zoom_spinbutton_value_changed),
		    self);

  priv->toolbar_text = gtk_ui_manager_get_widget (priv->ui_mgr, "/TextToolBar");
  g_assert (GTK_IS_TOOLBAR (priv->toolbar_text));

  priv->font_size = gtk_spin_button_new_with_range (FONT_SIZE_MIN, FONT_SIZE_MAX, 1);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->font_size), 12);
  gtk_widget_set_tooltip_text (priv->font_size, _("Font size"));
  toolitem = gtk_tool_item_new ();
  gtk_container_add (GTK_CONTAINER (toolitem), priv->font_size);
  gtk_toolbar_insert (GTK_TOOLBAR (priv->toolbar_text), toolitem, 5);

  g_signal_connect (G_OBJECT (priv->font_size), "value-changed",
		    G_CALLBACK (ocr_panel_font_size_spinbutton_value_changed),
		    self);

  priv->imageview = g_object_new (OCR_TYPE_IMAGEVIEW,
				  "can-focus", TRUE,
				  "n-columns", 1,
				  "n-rows", 1,
				  "homogeneous", FALSE,
				  "image-file", priv->image_file,
				  NULL);
  g_object_ref (priv->imageview);

  g_return_if_fail (OCR_IS_IMAGEVIEW (priv->imageview));

  priv->textbuffer = g_object_new (OCR_TYPE_TEXTBUFFER,
				   "tag-table", ocr_window_get_texttagtable (priv->main_window),
				   "spellcheck", ocr_window_get_spellcheck (priv->main_window),
				   NULL);

  g_signal_connect (G_OBJECT (priv->textbuffer), "modified-changed",
		    G_CALLBACK (ocr_panel_textbuffer_modified_changed),
		    self);

  g_return_if_fail (OCR_IS_TEXTBUFFER (priv->textbuffer));

  priv->textview = g_object_new (OCR_TYPE_TEXTVIEW,
				 "textbuffer", priv->textbuffer,
				 "spellcheck", ocr_window_get_spellcheck (priv->main_window),
				 NULL);
  g_object_ref (priv->textview);

  g_return_if_fail (OCR_IS_TEXTVIEW (priv->textview));

  priv->vbox_image = gtk_vbox_new (FALSE, 3);

  gtk_box_pack_start (GTK_BOX (priv->vbox_image),
		      priv->toolbar_image,
		      FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (priv->vbox_image),
		      priv->imageview,
		      TRUE, TRUE, 0);

  priv->vbox_text = gtk_vbox_new (FALSE, 3);
  
  scrolled_window_text = gtk_scrolled_window_new (NULL, NULL);
  g_object_ref (scrolled_window_text);
  gtk_container_add (GTK_CONTAINER (scrolled_window_text), priv->textview);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window_text),
				       GTK_SHADOW_IN);

  gtk_box_pack_start (GTK_BOX (priv->vbox_text),
		      priv->toolbar_text,
		      FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (priv->vbox_text),
		      scrolled_window_text,
		      TRUE, TRUE, 0);

  /* only for test
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (priv->textbuffer),
			    "PHP si è conquistato un seguito tra i designer Web non tecnici che hanno la necessità di inpserire aspetti interattivi nwi propi siti. Con una curva di apprendimento molto bassa, ...\nQuanto è bella l'uva fogarina, quanto è bello saperla vendemiar!\nLa cità di Treviso è cirpondata da nura!\n", -1);*/

  preferences = OCR_PREFERENCES (ocr_window_get_preferences (priv->main_window));
  ocr_panel_change_layout (self,
			   ocr_preferences_get_layout (preferences));
}

static void
ocr_panel_set_property (GObject *gobject,
			   guint property_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
  OcrPanel *self;
  OcrPanelPrivate *priv;

  g_return_if_fail (OCR_IS_PANEL (gobject));

  self = OCR_PANEL (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_MAIN_WINDOW:
      g_assert (OCR_IS_WINDOW (g_value_get_object (value)));
      priv->main_window = OCR_WINDOW (g_value_dup_object (value));
      break;
    case PROP_IMAGE_FILE:
      g_assert (G_IS_FILE (g_value_get_object (value)));
      priv->image_file = G_FILE (g_value_dup_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_panel_class_init (OcrPanelClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrPanelPrivate));

  gobject_class->constructed = ocr_panel_constructed;
  gobject_class->set_property = ocr_panel_set_property;
  gobject_class->dispose = ocr_panel_dispose;
  gobject_class->finalize = ocr_panel_finalize;

  g_object_class_install_property (gobject_class, PROP_MAIN_WINDOW,
				   g_param_spec_object ("main-window", "main-window", "main-window",
							GTK_TYPE_WINDOW,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_IMAGE_FILE,
				   g_param_spec_object ("image-file", "image-file", "image-file",
							G_TYPE_FILE,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_panel_init (OcrPanel *self)
{
  OcrPanelPrivate *priv;

  g_return_if_fail (OCR_IS_PANEL (self));

  self->priv = OCR_PANEL_GET_PRIVATE (self);
  priv = self->priv;

  priv->imageview = NULL;
  priv->textview = NULL;
}

void
ocr_panel_destroy_selection_font_dialog (OcrPanel *self)
{
  gpointer data;

  g_return_if_fail (OCR_IS_PANEL (self));

  data = g_object_get_data (G_OBJECT (self),
			    OCR_PANEL_FONT_SELECTION_DIALOG_KEY);

  if (data != NULL)
  {
    g_return_if_fail (GTK_IS_FONT_SELECTION_DIALOG (data));
    
    gtk_widget_destroy (GTK_WIDGET (data));
  }
}

void
ocr_panel_destroy_save_text_dialog (OcrPanel *self)
{
  gpointer data;

  g_return_if_fail (OCR_IS_PANEL (self));

  data = g_object_get_data (G_OBJECT (self),
			    OCR_PANEL_SAVE_TEXT_DIALOG_KEY);

  if (data != NULL)
  {
    g_return_if_fail (GTK_IS_FILE_CHOOSER_DIALOG (data));
    
    gtk_widget_destroy (GTK_WIDGET (data));
  }
}
