#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>
#include <gdk/gdkkeysyms.h>

#include "ocr-window.h"
#include "ocr-preferences.h"
#include "ocr-imagelist.h"
#include "ocr-spellcheck.h"
#include "ocr-utils.h"

#define OCR_PREFERENCES_BUILDER_FILE "ocr-preferences.ui"
#define OCR_PREFERENCES_OCR_FOLDER "/."PACKAGE_NAME
#define OCR_PREFERENCES_INI_FILE_NAME "/preferences"
#define OCR_PREFERENCES_TEMP_FOLDER "/temp"

#define OCR_PREFERENCES_TESSDATA_PREFIX_DEFAULT_1 "/usr/share/tessdata"
#define OCR_PREFERENCES_TESSDATA_PREFIX_DEFAULT_2 "/usr/share/tesseract-ocr/tessdata"

#define OCR_PREFERENCES_DIALOG_KEY "ocr-preferences-dialog-key"
#define OCR_PREFERENCES_DIALOG_BUILDER_KEY "ocr-preferences-dialog-builder-key"

#define OCR_PREFERENCES_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_PREFERENCES, OcrPreferencesPrivate))

G_DEFINE_TYPE (OcrPreferences, ocr_preferences, G_TYPE_OBJECT);

static void
ocr_preferences_create_tesseract_language_list (OcrPreferences *self,
						GtkBuilder     *builder);
static void
ocr_preferences_destroy_tesseract_language_list (OcrPreferences *self);

enum
{
  PROP_MAIN_WINDOW = 1
};

struct _OcrPreferencesPrivate
{
  OcrWindow *main_window;
  GKeyFile  *ini_file;
  GList     *hunspell_dictionary_list;
  GList     *tesseract_language_list;
};

static GtkBuilder *
ocr_preferences_get_dialog_builder (OcrPreferences *self)
{
  gpointer builder;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  builder = g_object_get_data (G_OBJECT (self),
			       OCR_PREFERENCES_DIALOG_BUILDER_KEY);
  if (GTK_IS_BUILDER (builder))
    return GTK_BUILDER (builder);
  else
    return NULL;
}

static void
ocr_preferences_dispose (GObject *gobject)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GObject               *obj;
  GList                 *it;
  static gboolean        first_run = TRUE;

  g_return_if_fail (OCR_IS_PREFERENCES (gobject));

  self = OCR_PREFERENCES (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->main_window != NULL)
      g_object_unref (priv->main_window);

    if (priv->ini_file != NULL)
      g_key_file_free (priv->ini_file);

    if (priv->hunspell_dictionary_list != NULL)
      g_list_free (priv->hunspell_dictionary_list);

    ocr_preferences_destroy_tesseract_language_list (self);

    obj = NULL;
    obj = G_OBJECT (ocr_preferences_get_dialog_builder (OCR_PREFERENCES (gobject)));
    if (obj != NULL)
      g_object_unref (obj);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_preferences_parent_class)->dispose (gobject);
}

static void
ocr_preferences_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_PREFERENCES (gobject));

  G_OBJECT_CLASS (ocr_preferences_parent_class)->finalize (gobject);
}

static void
ocr_preferences_load_preferences (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *file_name, *dir;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;
  
  dir = g_strconcat (ocr_utils_get_home_dir (),
		     OCR_PREFERENCES_OCR_FOLDER,
		     NULL);
  file_name = g_strconcat (dir,
			   OCR_PREFERENCES_INI_FILE_NAME,
			   NULL);
  priv->ini_file = g_key_file_new ();
  
  if (!g_key_file_load_from_file (priv->ini_file,
				  file_name,
				  G_KEY_FILE_NONE,
				  NULL))
    g_mkdir (dir, 0777);
  
  g_free (dir);
  g_free (file_name);
}

static void
ocr_preferences_write_preferences (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *content;
  gchar                 *dir;
  gchar                 *file_name;
  GError                *error = NULL;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  content = g_key_file_to_data (priv->ini_file,
				NULL,
				&error);

  if (error != NULL)
  {
    ocr_utils_message (_("Error writing preferences file"),
		       error->message,
		       GTK_MESSAGE_ERROR);

    g_clear_error (&error);
    return;
  }

  dir = g_strconcat (ocr_utils_get_home_dir (),
		     OCR_PREFERENCES_OCR_FOLDER,
		     NULL);
  file_name = g_strconcat (dir,
			   OCR_PREFERENCES_INI_FILE_NAME,
			   NULL); 

  ocr_utils_write_file (file_name,
			content);

  g_free (content);
  g_free (dir);
  g_free (file_name);
}

static void
ocr_preferences_constructed (GObject *gobject)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (gobject));

  self = OCR_PREFERENCES (gobject);
  priv = self->priv;

  ocr_preferences_load_preferences (self);
}

static void
ocr_preferences_set_property (GObject      *gobject,
			      guint         property_id,
			      const GValue *value,
			      GParamSpec   *pspec)
{
  OcrPreferences *self;
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (gobject));

  self = OCR_PREFERENCES (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_MAIN_WINDOW:
      g_assert (OCR_IS_WINDOW (g_value_get_object (value)));
      priv->main_window = OCR_WINDOW (g_value_dup_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_preferences_class_init (OcrPreferencesClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrPreferencesPrivate));

  gobject_class->constructed = ocr_preferences_constructed;
  gobject_class->set_property = ocr_preferences_set_property;
  gobject_class->dispose = ocr_preferences_dispose;
  gobject_class->finalize = ocr_preferences_finalize;

  g_object_class_install_property (gobject_class, PROP_MAIN_WINDOW,
				   g_param_spec_object ("main-window", "main-window", "main-window",
							OCR_TYPE_WINDOW,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_preferences_init (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;

  self->priv = OCR_PREFERENCES_GET_PRIVATE (self);
  priv = self->priv;

  priv->main_window = NULL;
  priv->hunspell_dictionary_list = NULL;
  priv->tesseract_language_list = NULL;
}

gchar * 
ocr_preferences_get_text_font (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;
  
  str = g_key_file_get_string (priv->ini_file,
			       "General",
			       "TextFont",
			       NULL);

  if (str == NULL)
    return g_strdup ("Sans 12");

  return str;
}

gint 
ocr_preferences_get_layout (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gint                   integer;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  integer = g_key_file_get_integer (priv->ini_file,
				    "General",
				    "Layout",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return OCR_PREFERENCES_LAYOUT_V;
  }

  return integer;
}

gchar * 
ocr_preferences_get_save_default_folder (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;
  
  str = g_key_file_get_string (priv->ini_file,
			       "General",
			       "SaveDefaultFolder",
			       NULL);

  if (str == NULL)
    return g_strdup (ocr_utils_get_home_dir ());

  return str;
}

gboolean
ocr_preferences_get_remove_nl (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "General",
				    "RemoveNewLineCharacter",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return FALSE;
  }

  return boolean;
}

gboolean
ocr_preferences_get_substitute_with_space (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "General",
				    "SubstituteWithSpace",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return FALSE;
  }

  return boolean;
}

gdouble
ocr_preferences_get_number_blank_rows (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gdouble                ndouble;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  ndouble = g_key_file_get_double (priv->ini_file,
				   "General",
				   "NumberBlankRows",
				   &error);

  if (error != NULL)
  {
    g_error_free (error);
    return 1;
  }

  return ndouble;
}

gchar * 
ocr_preferences_get_temp_folder (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;
  gchar                 *dir;
  gchar                 *temp_dir;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;
  
  str = g_key_file_get_string (priv->ini_file,
			       "General",
			       "TempFolder",
			       NULL);

  if (str == NULL)
  {
    dir = g_strconcat (ocr_utils_get_home_dir (),
		       OCR_PREFERENCES_OCR_FOLDER,
		       NULL);

    temp_dir = g_strconcat (dir,
			    OCR_PREFERENCES_TEMP_FOLDER,
			    NULL);

    if (!g_file_test (temp_dir,
		      G_FILE_TEST_IS_DIR))
      g_mkdir (temp_dir, 0777);

    g_free (dir);

    return temp_dir;
  }

  return str;
}

gint 
ocr_preferences_get_preferred_ocr_engine (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gint                   integer;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  integer = g_key_file_get_integer (priv->ini_file,
				    "General",
				    "PreferredOcrEngine",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return OCR_PREFERENCES_PREFERRED_OCR_ENGINE_TESSERACT;
  }

  return integer;
}

gchar *
ocr_preferences_get_tesseract_language_code (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;

  str = g_key_file_get_string (priv->ini_file,
			       "Tesseract",
			       "LanguageCode",
			       NULL);

  return str;
}

gchar *
ocr_preferences_get_tesseract_dictionary_path (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;

  str = g_key_file_get_string (priv->ini_file,
			       "Tesseract",
			       "DictionaryPath",
			       NULL);

  /* function gtk_entry_get_text, used in function ocr_preferences_set_tesseract_dictionary_path, returns an empty string (but not NULL), if field Tesseract Dictionary Path has no characters. In this case, is necessary to check the string and set a NULL string */
  if (g_strcmp0 (str, "") == 0)
  {
    g_free (str);
    str = NULL;
  }

  return str;
}

gdouble
ocr_preferences_get_gocr_gray_level (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gdouble                ndouble;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  ndouble = g_key_file_get_double (priv->ini_file,
				   "Gocr",
				   "GreyLevel",
				   &error);

  if (error != NULL)
  {
    g_error_free (error);
    return 0;  /* 0 for autodetect */
  }

  return ndouble;
}

gboolean
ocr_preferences_get_gocr_gray_level_autodetect (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "Gocr",
				    "GreyLevelAutodetect",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return TRUE;
  }

  return boolean;
}

gdouble
ocr_preferences_get_gocr_dust_size (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gdouble                ndouble;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  ndouble = g_key_file_get_double (priv->ini_file,
				   "Gocr",
				   "DustSize",
				   &error);

  if (error != NULL)
  {
    g_error_free (error);
    return -1;  /* -1 for autodetect */
  }

  return ndouble;
}

gboolean
ocr_preferences_get_gocr_dust_size_autodetect (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "Gocr",
				    "DustSizeAutodetect",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return TRUE;
  }

  return boolean;
}

gdouble
ocr_preferences_get_gocr_spacewidth_between_words (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gdouble                ndouble;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  ndouble = g_key_file_get_double (priv->ini_file,
				   "Gocr",
				   "SpaceWidthBetweenWords",
				   &error);

  if (error != NULL)
  {
    g_error_free (error);
    return 0;  /* 0 for autodetect */
  }

  return ndouble;
}

gboolean
ocr_preferences_get_gocr_spacewidth_between_words_autodetect (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "Gocr",
				    "SpaceWidthBetweenWordsAutodetect",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return TRUE;
  }

  return boolean;
}

gdouble
ocr_preferences_get_gocr_value_for_certainty_of_recognition (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gdouble                ndouble;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), 0);

  priv = self->priv;

  ndouble = g_key_file_get_double (priv->ini_file,
				   "Gocr",
				   "ValueForCertaintyOfRecognition",
				   &error);

  if (error != NULL)
  {
    g_error_free (error);
    return 95;  /* default 95 */
  }

  return ndouble;
}

gboolean
ocr_preferences_get_gocr_value_for_certainty_of_recognition_default (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gboolean               boolean;
  GError                *error = NULL;

  g_return_val_if_fail (OCR_IS_PREFERENCES (self), FALSE);

  priv = self->priv;

  boolean = g_key_file_get_boolean (priv->ini_file,
				    "Gocr",
				    "ValueForCertaintyOfRecognitionDefault",
				    &error);

  if (error != NULL)
  {
    g_error_free (error);
    return TRUE;
  }

  return boolean;
}

gchar *
ocr_preferences_get_hunspell_dictionary_path_1 (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;

  str = g_key_file_get_string (priv->ini_file,
			       "Hunspell",
			       "DictionaryPath1",
			       NULL);

  return str;
}

gchar *
ocr_preferences_get_hunspell_dictionary_path_2 (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gchar                 *str;
  GError                *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;

  str = g_key_file_get_string (priv->ini_file,
			       "Hunspell",
			       "DictionaryPath2",
			       NULL);

  /* this piece of code return the name of system language type
     e.g. "it_IT" or "es_US"
  if (str == NULL)
  {
    gchar **langs; 
    gchar   lang[20];
    gint    i, m;

    langs = (gchar **) g_get_language_names();
    for (i = 0, m = 0; langs[0][i]; i++)
    {
      if (langs[0][i] == '.')
        break;
      lang[m] = langs[0][i];
      m++;
    }
    lang[m] = 0;

    if (lang[0] != 0)
      str = g_strdup_printf ("%s",
			     lang);
  }*/

  return str;
}

static void 
ocr_preferences_set_text_font (OcrPreferences *self, 
			       const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "General",
			 "TextFont",
			 value);
}

static void
ocr_preferences_set_layout (OcrPreferences *self,
			    const gint      value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_integer (priv->ini_file,
			 "General",
			 "Layout",
			 value);
}

static void 
ocr_preferences_set_save_default_folder (OcrPreferences *self, 
					 const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "General",
			 "SaveDefaultFolder",
			 value);
}

static void 
ocr_preferences_set_remove_nl (OcrPreferences *self, 
			       const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "General",
			  "RemoveNewLineCharacter",
			  value);
}

static void 
ocr_preferences_set_substitute_with_space (OcrPreferences *self, 
					   const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "General",
			  "SubstituteWithSpace",
			  value);
}

static void
ocr_preferences_set_number_blank_rows (OcrPreferences *self,
				       const gdouble   value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_double (priv->ini_file,
			 "General",
			 "NumberBlankRows",
			 value);
}

static void 
ocr_preferences_set_temp_folder (OcrPreferences *self, 
				 const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "General",
			 "TempFolder",
			 value);
}

static void
ocr_preferences_set_preferred_ocr_engine (OcrPreferences *self,
					  const gint      value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_integer (priv->ini_file,
			 "General",
			 "PreferredOcrEngine",
			 value);
}

static void
ocr_preferences_set_tesseract_language_code (OcrPreferences *self,
					     const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "Tesseract",
			 "LanguageCode",
			 value);
}

static gchar *
ocr_preferences_tesseract_language_list_read_selected (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GList                 *it;
  GtkRadioButton        *radio_button;
  gchar                 *str;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  str = NULL;
  if (priv->tesseract_language_list != NULL)
  {
    for (it = priv->tesseract_language_list; it != NULL; it = it->next)
    {
      radio_button = (GtkRadioButton*) it->data;
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio_button)))
	str = g_strdup (gtk_button_get_label (GTK_BUTTON (radio_button)));
    }
  }
 
  return str;
}

static void
ocr_preferences_tesseract_language_list_set_selected (OcrPreferences *self,
						      gchar          *value)
{
  OcrPreferencesPrivate *priv;
  GList                 *it;
  GtkRadioButton        *radio_button;

  g_return_if_fail (OCR_IS_PREFERENCES (self));
  g_return_if_fail (value != NULL);

  priv = self->priv;

  if (priv->tesseract_language_list != NULL)
  {
    for (it = priv->tesseract_language_list; it != NULL; it = it->next)
    {
      radio_button = (GtkRadioButton*) it->data;
      if (!g_strcmp0 (gtk_button_get_label (GTK_BUTTON (radio_button)),
		      value))
      {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_button),
				      TRUE);
      
	break;
      }
    }
  }
}

static void 
ocr_preferences_set_tesseract_dictionary_path (OcrPreferences *self,
                                               const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "Tesseract",
			 "DictionaryPath",
			 value);
}

static void
ocr_preferences_set_gocr_gray_level (OcrPreferences *self,
				     const gdouble   value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_double (priv->ini_file,
			 "Gocr",
			 "GreyLevel",
			 value);
}

static void 
ocr_preferences_set_gocr_gray_level_autodetect (OcrPreferences *self, 
						const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "Gocr",
			  "GreyLevelAutodetect",
			  value);
}

static void
ocr_preferences_set_gocr_dust_size (OcrPreferences *self,
				    const gdouble   value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_double (priv->ini_file,
			 "Gocr",
			 "DustSize",
			 value);
}

static void 
ocr_preferences_set_gocr_dust_size_autodetect (OcrPreferences *self, 
					       const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "Gocr",
			  "DustSizeAutodetect",
			  value);
}

static void
ocr_preferences_set_gocr_spacewidth_between_words (OcrPreferences *self,
						   const gdouble   value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_double (priv->ini_file,
			 "Gocr",
			 "SpaceWidthBetweenWords",
			 value);
}

static void 
ocr_preferences_set_gocr_spacewidth_between_words_autodetect (OcrPreferences *self, 
							      const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "Gocr",
			  "SpaceWidthBetweenWordsAutodetect",
			  value);
}

static void
ocr_preferences_set_gocr_value_for_certainty_of_recognition (OcrPreferences *self,
							     const gdouble   value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_double (priv->ini_file,
			 "Gocr",
			 "ValueForCertaintyOfRecognition",
			 value);
}

static void 
ocr_preferences_set_gocr_value_for_certainty_of_recognition_default (OcrPreferences *self, 
								     const gboolean  value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_boolean (priv->ini_file,
			  "Gocr",
			  "ValueForCertaintyOfRecognitionDefault",
			  value);
}

static void 
ocr_preferences_set_hunspell_dictionary_path_1 (OcrPreferences *self, 
						const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "Hunspell",
			 "DictionaryPath1",
			 value);
}

static void 
ocr_preferences_set_hunspell_dictionary_path_2 (OcrPreferences *self, 
						const gchar    *value)
{
  OcrPreferencesPrivate *priv;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  g_key_file_set_string (priv->ini_file,
			 "Hunspell",
			 "DictionaryPath2",
			 value);
}

static gchar *
ocr_preferences_hunspell_dictionary_list_read_selected (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GList                 *it;
  GtkRadioButton        *radio_button;
  gchar                 *str;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  str = NULL;
  if (priv->hunspell_dictionary_list != NULL)
  {
    for (it = priv->hunspell_dictionary_list; it != NULL; it = it->next)
    {
      radio_button = (GtkRadioButton*) it->data;
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio_button)))
	str = g_strdup (gtk_button_get_label (GTK_BUTTON (radio_button)));
    }
  }
 
  return str;
}

static void
ocr_preferences_hunspell_dictionary_list_set_selected (OcrPreferences *self,
						       gchar          *value)
{
  OcrPreferencesPrivate *priv;
  GList                 *it;
  GtkRadioButton        *radio_button;

  g_return_if_fail (OCR_IS_PREFERENCES (self));
  g_return_if_fail (value != NULL);

  priv = self->priv;

  if (priv->hunspell_dictionary_list != NULL)
  {
    for (it = priv->hunspell_dictionary_list; it != NULL; it = it->next)
    {
      radio_button = (GtkRadioButton*) it->data;
      if (!g_strcmp0 (gtk_button_get_label (GTK_BUTTON (radio_button)),
		      value))
      {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_button),
				      TRUE);
      
	break;
      }
    }
  }
}

gchar *
ocr_preferences_get_hunspell_dictionary_path (OcrPreferences *self)
{
  gchar *str = NULL;
  gchar *title;
  gchar *message;

  str = ocr_preferences_get_hunspell_dictionary_path_2 (self);
  if (!g_strcmp0 (str,""))
  {
    g_free (str);
    str = NULL;
  }

  if (str == NULL)
    str = ocr_preferences_get_hunspell_dictionary_path_1 (self);

  if (!g_strcmp0 (str,""))
  {
    g_free (str);
    str = NULL;
  }

  return str;
}

static void
ocr_preferences_get_keys_from_dialog (OcrPreferences *self,
				      const gchar *property_id,
				      ...)
{
  GtkBuilder  *builder;
  GtkWidget   *wptr;
  va_list      varargs;
  const gchar *cstr;
  gchar       *str;
  gint         integer;
  gdouble      ndouble;
  gboolean     boolean;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  builder = ocr_preferences_get_dialog_builder (self);

  g_assert (GTK_IS_BUILDER (builder));

  va_start (varargs, property_id);

  while (property_id != NULL)
  {
    wptr = GTK_WIDGET (gtk_builder_get_object (builder,
					       property_id));

    if (g_strcmp0 (property_id, "text_font") == 0)
    {
      cstr = gtk_font_button_get_font_name (GTK_FONT_BUTTON (wptr));
      ocr_preferences_set_text_font (self,
				     cstr);
    }

    if (g_strcmp0 (property_id, "layout_v") == 0)
    {
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr)))
	ocr_preferences_set_layout (self,
				    OCR_PREFERENCES_LAYOUT_V);
    }

    if (g_strcmp0 (property_id, "layout_h") == 0)
    {
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr)))
	ocr_preferences_set_layout (self,
				    OCR_PREFERENCES_LAYOUT_H);
    }

    if (g_strcmp0 (property_id, "save_default_folder") == 0)
    {
      str = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (wptr));
      ocr_preferences_set_save_default_folder (self,
					       str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "remove_nl") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_remove_nl (self,
				     boolean);
    }

    if (g_strcmp0 (property_id, "substitute_with_space") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_substitute_with_space (self,
						 boolean);
    }

    if (g_strcmp0 (property_id, "number_blank_rows") == 0)
    {
      ndouble = gtk_spin_button_get_value (GTK_SPIN_BUTTON (wptr));
      ocr_preferences_set_number_blank_rows (self,
					     ndouble);
    }

    if (g_strcmp0 (property_id, "temp_folder") == 0)
    {
      str = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (wptr));
      ocr_preferences_set_temp_folder (self,
				       str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "preferred_tesseract") == 0)
    {
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr)))
	ocr_preferences_set_preferred_ocr_engine (self,
						  OCR_PREFERENCES_PREFERRED_OCR_ENGINE_TESSERACT);
    }

    if (g_strcmp0 (property_id, "preferred_gocr") == 0)
    {
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr)))
	ocr_preferences_set_preferred_ocr_engine (self,
						  OCR_PREFERENCES_PREFERRED_OCR_ENGINE_GOCR);
    }

    if (g_strcmp0 (property_id, "tesseract_dictionary_path") == 0)
    {
      str = g_strdup ((gchar *) gtk_entry_get_text (GTK_ENTRY (wptr)));
      ocr_preferences_set_tesseract_dictionary_path (self,
						     str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "gocr_l") == 0)
    {
      ndouble = gtk_spin_button_get_value (GTK_SPIN_BUTTON (wptr));
      ocr_preferences_set_gocr_gray_level (self,
					   ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_l_autodetect") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_gocr_gray_level_autodetect (self,
						      boolean);
    }

    if (g_strcmp0 (property_id, "gocr_d") == 0)
    {
      ndouble = gtk_spin_button_get_value (GTK_SPIN_BUTTON (wptr));
      ocr_preferences_set_gocr_dust_size (self,
					  ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_d_autodetect") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_gocr_dust_size_autodetect (self,
						     boolean);
    }

    if (g_strcmp0 (property_id, "gocr_s") == 0)
    {
      ndouble = gtk_spin_button_get_value (GTK_SPIN_BUTTON (wptr));
      ocr_preferences_set_gocr_spacewidth_between_words (self,
							 ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_s_autodetect") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_gocr_spacewidth_between_words_autodetect (self,
								    boolean);
    }

    if (g_strcmp0 (property_id, "gocr_a") == 0)
    {
      ndouble = gtk_spin_button_get_value (GTK_SPIN_BUTTON (wptr));
      ocr_preferences_set_gocr_value_for_certainty_of_recognition (self,
								   ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_a_default") == 0)
    {
      boolean = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wptr));
      ocr_preferences_set_gocr_value_for_certainty_of_recognition_default (self,
									   boolean);
    }

    if (g_strcmp0 (property_id, "hunspell_dictionary_path") == 0)
    {
      str = g_strdup ((gchar *) gtk_entry_get_text (GTK_ENTRY (wptr)));
      ocr_preferences_set_hunspell_dictionary_path_2 (self,
						      str);
      g_free (str);
    }

    property_id = va_arg (varargs, const gchar *);
  }

  str = ocr_preferences_tesseract_language_list_read_selected (self);
  if (str == NULL)
    str = g_strdup ("");
  ocr_preferences_set_tesseract_language_code (self,
					       str);
  g_free (str);

  str = ocr_preferences_hunspell_dictionary_list_read_selected (self);
  if (str == NULL)
    str = g_strdup ("");
  ocr_preferences_set_hunspell_dictionary_path_1 (self,
						  str);
  g_free (str);

  va_end (varargs);
}

static void
ocr_preferences_set_keys_on_dialog (OcrPreferences *self,
				    const gchar *property_id,
				    ...)
{
  GtkBuilder *builder;
  va_list     varargs;
  GtkWidget  *wptr;
  gchar      *str;
  gint        integer;
  gdouble     ndouble;
  gboolean    boolean;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  builder = ocr_preferences_get_dialog_builder (self);

  g_assert (GTK_IS_BUILDER (builder));

  va_start (varargs, property_id);

  while (property_id != NULL)
  {
    wptr = GTK_WIDGET (gtk_builder_get_object (builder,
					       property_id));

    if (g_strcmp0 (property_id, "text_font") == 0)
    {
      str = ocr_preferences_get_text_font (self);
      gtk_font_button_set_font_name (GTK_FONT_BUTTON (wptr),
				     str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "layout_v") == 0)
    {
      integer = ocr_preferences_get_layout (self);
      if (integer == OCR_PREFERENCES_LAYOUT_V)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      TRUE);
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      FALSE);
    }

    if (g_strcmp0 (property_id, "layout_h") == 0)
    {
      integer = ocr_preferences_get_layout (self);
      if (integer == OCR_PREFERENCES_LAYOUT_H)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      TRUE);
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      FALSE);
    }

    if (g_strcmp0 (property_id, "save_default_folder") == 0)
    {
      str = ocr_preferences_get_save_default_folder (self);
      gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (wptr),
					   str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "remove_nl") == 0)
    {
      boolean = ocr_preferences_get_remove_nl (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "substitute_with_space") == 0)
    {
      boolean = ocr_preferences_get_substitute_with_space (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "number_blank_rows") == 0)
    {
      ndouble = ocr_preferences_get_number_blank_rows (self);
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wptr),
				 ndouble);
    }

    if (g_strcmp0 (property_id, "temp_folder") == 0)
    {
      str = ocr_preferences_get_temp_folder (self);
      gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (wptr),
					   str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "preferred_tesseract") == 0)
    {
      integer = ocr_preferences_get_preferred_ocr_engine (self);
      if (integer == OCR_PREFERENCES_PREFERRED_OCR_ENGINE_TESSERACT)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      TRUE);
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      FALSE);
    }

    if (g_strcmp0 (property_id, "preferred_gocr") == 0)
    {
      integer = ocr_preferences_get_preferred_ocr_engine (self);
      if (integer == OCR_PREFERENCES_PREFERRED_OCR_ENGINE_GOCR)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      TRUE);
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				      FALSE);
    }

    if (g_strcmp0 (property_id, "tesseract_dictionary_path") == 0)
    {
      str = ocr_preferences_get_tesseract_dictionary_path (self);
      if (str != NULL)
	gtk_entry_set_text (GTK_ENTRY (wptr),
			    str);
      g_free (str);
    }

    if (g_strcmp0 (property_id, "gocr_l") == 0)
    {
      ndouble = ocr_preferences_get_gocr_gray_level (self);
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wptr),
				 ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_l_autodetect") == 0)
    {
      boolean = ocr_preferences_get_gocr_gray_level_autodetect (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "gocr_d") == 0)
    {
      ndouble = ocr_preferences_get_gocr_dust_size (self);
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wptr),
				 ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_d_autodetect") == 0)
    {
      boolean = ocr_preferences_get_gocr_dust_size_autodetect (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "gocr_s") == 0)
    {
      ndouble = ocr_preferences_get_gocr_spacewidth_between_words (self);
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wptr),
				 ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_s_autodetect") == 0)
    {
      boolean = ocr_preferences_get_gocr_spacewidth_between_words_autodetect (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "gocr_a") == 0)
    {
      ndouble = ocr_preferences_get_gocr_value_for_certainty_of_recognition (self);
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wptr),
				 ndouble);
    }

    if (g_strcmp0 (property_id, "gocr_a_default") == 0)
    {
      boolean = ocr_preferences_get_gocr_value_for_certainty_of_recognition_default (self);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wptr),
				    boolean);
    }

    if (g_strcmp0 (property_id, "hunspell_dictionary_path") == 0)
    {
      str = ocr_preferences_get_hunspell_dictionary_path_2 (self);
      if (str != NULL)
	gtk_entry_set_text (GTK_ENTRY (wptr),
			    str);
      g_free (str);
    }

    property_id = va_arg (varargs, const gchar *);
  }

  str = ocr_preferences_get_tesseract_language_code (self);
  if (str != NULL)
    ocr_preferences_tesseract_language_list_set_selected (self,
							  str);
  g_free (str);

  str = ocr_preferences_get_hunspell_dictionary_path_1 (self);
  if (str != NULL)
    ocr_preferences_hunspell_dictionary_list_set_selected (self,
							   str);
  g_free (str);

  va_end (varargs);
}

static void
ocr_preferences_text_font_font_set (GtkFontButton *text_font,
				    gpointer       user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  OcrImageList          *imagelist;
  const gchar           *font_name;

  g_return_if_fail (GTK_IS_FONT_BUTTON (text_font));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  font_name = gtk_font_button_get_font_name (text_font);

  imagelist = OCR_IMAGE_LIST (ocr_window_get_imagelist (priv->main_window));
  ocr_image_list_set_font (imagelist,
			   font_name);
}

static void
ocr_preferences_layout_v_toggled (GtkToggleButton *togglebutton,
				  gpointer         user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  OcrImageList          *imagelist;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (togglebutton));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  imagelist = OCR_IMAGE_LIST (ocr_window_get_imagelist (priv->main_window));
  if (gtk_toggle_button_get_active (togglebutton))
    ocr_image_list_change_layout (imagelist,
				  OCR_PREFERENCES_LAYOUT_V);
  else
    ocr_image_list_change_layout (imagelist,
				  OCR_PREFERENCES_LAYOUT_H);
}

static void
ocr_preferences_remove_nl_toggled (GtkToggleButton *remove_nl,
				   gpointer         user_data) 
{
  GtkBuilder *builder;
  GtkWidget  *widget;
  gboolean    value;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (remove_nl));
  g_return_if_fail (GTK_IS_BUILDER (user_data));

  builder = (GtkBuilder *) user_data;

  value = gtk_toggle_button_get_active (remove_nl);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "substitute_with_space"));
  
  gtk_widget_set_sensitive (widget,
			    value);
}

static void
ocr_preferences_tesseract_dictionary_path_reset_list (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkWidget             *widget;
  gchar                 *str;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  builder = ocr_preferences_get_dialog_builder (self);

  g_assert (GTK_IS_BUILDER (builder));

  priv = self->priv;

  /* 1: setting of new dictionary path in preferences' file */
  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_dictionary_path"));
  str = g_strdup ((gchar *) gtk_entry_get_text (GTK_ENTRY (widget)));
  ocr_preferences_set_tesseract_dictionary_path (self,
						 str);
  g_free (str);

  /* 2: looking for dictionaries in new path */
  ocr_preferences_create_tesseract_language_list (self,
						  builder);
}

static gboolean
ocr_preferences_tesseract_path_pressed_key (GtkWidget   *entry,
					    GdkEventKey *event,
					    gpointer     user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;

  g_return_if_fail (GTK_IS_WIDGET (entry));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  if (event->type == GDK_KEY_PRESS &&
      event->keyval == GDK_Return)  /* return key pressed when in tesseract_dictionary_path field */
    ocr_preferences_tesseract_dictionary_path_reset_list (self);

  return FALSE;
}

static void
ocr_preferences_tesseract_directory_path_find_pressed (GtkButton *button,
						       gpointer   user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  ocr_preferences_tesseract_dictionary_path_reset_list (self);
}

static void
ocr_preferences_tesseract_directory_path_help_pressed (GtkButton *button,
						       gpointer   user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  gtk_show_uri (gtk_widget_get_screen (GTK_WIDGET (priv->main_window)),
		"file://"OCR_DOC_DIR"/ocrgui_manual.html#tesseract_dictionary_path_help",
		GDK_CURRENT_TIME,
		&error);

  if (error != NULL)
    g_clear_error (&error);
}

static void
ocr_preferences_tesseract_site_button_pressed (GtkButton *button,
					       gpointer   user_data)
{
  GdkAppLaunchContext *context;

  g_return_if_fail (GTK_IS_BUTTON (button));

  context = gdk_app_launch_context_new ();

  /* automatically launch the browers opening link button's uri */
  g_app_info_launch_default_for_uri (gtk_link_button_get_uri (GTK_LINK_BUTTON (button)),
				     NULL,
				     NULL);

  g_object_unref (context);
}

static void
ocr_preferences_gocr_l_autodetect_toggled (GtkToggleButton *togglebutton,
					   gpointer         user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkSpinButton         *spinbutton;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (togglebutton));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  builder = ocr_preferences_get_dialog_builder (self);

  spinbutton = GTK_SPIN_BUTTON (gtk_builder_get_object (builder,
							"gocr_l"));

  if (gtk_toggle_button_get_active (togglebutton))
  {
    gtk_spin_button_set_value (spinbutton,
			       -1);
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      FALSE);
  }
  else
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      TRUE);
}

static void
ocr_preferences_gocr_d_autodetect_toggled (GtkToggleButton *togglebutton,
					   gpointer         user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkSpinButton         *spinbutton;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (togglebutton));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  builder = ocr_preferences_get_dialog_builder (self);

  spinbutton = GTK_SPIN_BUTTON (gtk_builder_get_object (builder,
							"gocr_d"));

  if (gtk_toggle_button_get_active (togglebutton))
  {
    gtk_spin_button_set_value (spinbutton,
			       -1);
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      FALSE);
  }
  else
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      TRUE);
}

static void
ocr_preferences_gocr_s_autodetect_toggled (GtkToggleButton *togglebutton,
					   gpointer         user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkSpinButton         *spinbutton;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (togglebutton));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  builder = ocr_preferences_get_dialog_builder (self);

  spinbutton = GTK_SPIN_BUTTON (gtk_builder_get_object (builder,
							"gocr_s"));

  if (gtk_toggle_button_get_active (togglebutton))
  {
    gtk_spin_button_set_value (spinbutton,
			       0);
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      FALSE);
  }
  else
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      TRUE);
}

static void
ocr_preferences_gocr_a_default_toggled (GtkToggleButton *togglebutton,
					gpointer         user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkSpinButton         *spinbutton;

  g_return_if_fail (GTK_IS_TOGGLE_BUTTON (togglebutton));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  builder = ocr_preferences_get_dialog_builder (self);

  spinbutton = GTK_SPIN_BUTTON (gtk_builder_get_object (builder,
							"gocr_a"));

  if (gtk_toggle_button_get_active (togglebutton))
  {
    gtk_spin_button_set_value (spinbutton,
			       95);
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      FALSE);
  }
  else
    gtk_widget_set_sensitive (GTK_WIDGET (spinbutton),
			      TRUE);
}

static void
ocr_preferences_gocr_site_button_pressed (GtkButton *button,
					  gpointer   user_data)
{
  GdkAppLaunchContext *context;

  g_return_if_fail (GTK_IS_BUTTON (button));

  context = gdk_app_launch_context_new ();

  /* automatically launch the browers opening link button's uri */
  g_app_info_launch_default_for_uri (gtk_link_button_get_uri (GTK_LINK_BUTTON (button)),
				     NULL,
				     NULL);

  g_object_unref (context);
}

static void
ocr_preferences_hunspell_dictionary_path_help_pressed (GtkButton *button,
						       gpointer   user_data)
{
  OcrPreferences        *self;
  OcrPreferencesPrivate *priv;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (OCR_IS_PREFERENCES (user_data));

  self = OCR_PREFERENCES (user_data);
  priv = self->priv;

  gtk_show_uri (gtk_widget_get_screen (GTK_WIDGET (priv->main_window)),
		"file://"OCR_DOC_DIR"/ocrgui_manual.html#hunspell_dictionary_path_help",
		GDK_CURRENT_TIME,
		&error);

  if (error != NULL)
    g_clear_error (&error);
}

gboolean
ocr_preferences_check_program (OcrPreferences *self,
			       const gchar    *program_name)
{
  gchar  **command;
  gint     exit_status;
  GError  *error = NULL;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  command = g_new (gchar*, 3);
  command[0] = g_strdup ("which");
  command[1] = g_strdup (program_name);
  command[2] = NULL;
  
  g_spawn_sync (ocr_utils_get_home_dir (),
		command,
		NULL,
		G_SPAWN_SEARCH_PATH,
		NULL,
		NULL,
		NULL,
		NULL,
		&exit_status,
		&error);

  g_strfreev (command);

  if (error != NULL)
  {
    gchar *message = g_strdup_printf (_("Can't execute command «which %s»:\n%s"),
				      program_name,
				      error->message);

    ocr_utils_message (_("Error checking for missing programs"),
		       message,
		       GTK_MESSAGE_ERROR);

    g_free (message);
    g_clear_error (&error);
    return FALSE;
  }

  if (exit_status)
    return FALSE;
  else
    return TRUE;
}

static void
ocr_preferences_check_missing_programs (OcrPreferences *self)
{
  GtkBuilder *builder;
  GtkWidget  *widget;
  GtkWidget  *widget_1;
  gboolean    retval[4];
  gchar      *installed;
  gchar      *not_installed;
  gchar      *installed_markup;
  gchar      *not_installed_markup;

  retval[0] = ocr_preferences_check_program (self,
					     "tesseract");
  retval[1] = ocr_preferences_check_program (self,
					     "gocr");
  retval[2] = ocr_preferences_check_program (self,
					     "convert");
  retval[3] = ocr_preferences_check_program (self,
					     "hunspell");

  builder = ocr_preferences_get_dialog_builder (self);

  g_assert (GTK_IS_BUILDER (builder));

  installed            = g_strdup (_("Installed"));
  not_installed        = g_strdup (_("Not installed"));
  installed_markup     = g_strdup_printf ("<span color='#00FF00'><b>%s</b></span>",
				      installed);
  not_installed_markup = g_strdup_printf ("<span color='#FF0000'><b>%s</b></span>",
					  not_installed);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					      "tesseract_status"));
  if (retval[0])
    gtk_label_set_markup (GTK_LABEL (widget),
			  installed_markup);
  else
    gtk_label_set_markup (GTK_LABEL (widget),
			  not_installed_markup);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					      "gocr_status"));
  if (retval[1])
    gtk_label_set_markup (GTK_LABEL (widget),
			  installed_markup);
  else
    gtk_label_set_markup (GTK_LABEL (widget),
			  not_installed_markup);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					      "convert_status"));
  if (retval[2])
    gtk_label_set_markup (GTK_LABEL (widget),
			  installed_markup);
  else
    gtk_label_set_markup (GTK_LABEL (widget),
			  not_installed_markup);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					      "hunspell_status"));
  if (retval[3])
    gtk_label_set_markup (GTK_LABEL (widget),
			  installed_markup);
  else
    gtk_label_set_markup (GTK_LABEL (widget),
			  not_installed_markup);

  if (!(retval[0] &&
        retval[1] &&
        retval[2] &&
	retval[3]))
  {
    widget_1 = GTK_WIDGET (gtk_builder_get_object (builder,
						   "various_widget"));
    widget = GTK_WIDGET (gtk_builder_get_object (builder,
						 "warning_hbox"));
    gtk_container_remove (GTK_CONTAINER (widget_1),
			  widget);

    widget_1 = GTK_WIDGET (gtk_builder_get_object (builder,
						   "programs_tab"));
    widget = GTK_WIDGET (gtk_builder_get_object (builder,
						 "warning_hbox"));
    g_assert (GTK_BOX (widget));
    gtk_box_pack_start (GTK_BOX (widget_1),
		      widget,
		      FALSE,
		      FALSE,
		      0);
  }

  g_free (installed);
  g_free (not_installed);
  g_free (installed_markup);
  g_free (not_installed_markup);
}

gint
ocr_preferences_find_tesseract_version (OcrPreferences *self)
{
  gint     version;
  gchar   *standard_error;
  gchar  **tokens;
  gchar  **subtokens;
  GError  *error = NULL;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  version = 1; /* error code */

  g_spawn_command_line_sync ("tesseract -v",
			     NULL,
			     &(standard_error),
			     NULL,
			     &error);
  if (error != NULL)
  {
    g_clear_error (&error);
  }

  tokens = g_strsplit_set (standard_error,
			   "\n",
			   -1);
  if (tokens[0] != NULL)
  {
    subtokens = g_strsplit_set (tokens[0],
				" ",
				-1);
    if (!g_strcmp0 (subtokens[0], "tesseract:Error:Usage:tesseract"))
    {
      version = 2;  /* Tesseract version is 2.x */
    }
    else
    {
      version = 3;  /* Tesseract version is 3.x 
		       In this case the command 'tesseract -v' returns the string 'tesseract 3.00' */
    }
  }

  if (standard_error != NULL)
  {
    g_free (standard_error);
    g_strfreev (tokens);
    g_strfreev (subtokens);
  }

  return version;
}

gchar **
ocr_preferences_find_tesseract_installed_languages (OcrPreferences *self)
{
  gchar        *tessdata_prefix_value = NULL;
  GDir         *dir;
  const gchar  *file_name;
  gchar       **tokens;
  gchar        *token;
  gchar       **installed_languages;
  GSList       *languages_list = NULL;
  GSList       *it = NULL;
  guint         n_lang;
  gint          version;
  gchar        *dic_file_name;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  version = ocr_preferences_find_tesseract_version (self);
  if (version == 1 ||  /* error */
      version == 2)    /* Tesseract version is 2.x 
			  In this case there are files named <language_code>.unicharset */
    dic_file_name = g_strdup ("unicharset");
  if (version == 3)  /* Tesseract version is 3.x
			In this case dictionary files are named <language_code>.traineddata */
    dic_file_name = g_strdup ("traineddata");

  n_lang = 0;

  /* Look for installed dictionaries first in custom folder
     then in TESSDATA_PREFIX, as suggestes by Tesseract documentation,
     and then in TESSERACT_PREFIX_DEFAULT_1 and TESSERACT_PREFIX_DEFAULT_2 */
  tessdata_prefix_value = ocr_preferences_get_tesseract_dictionary_path (self);

  if (tessdata_prefix_value == NULL)
  {
    if (g_getenv ("TESSDATA_PREFIX") != NULL)
    {
      tessdata_prefix_value = g_strdup (g_getenv ("TESSDATA_PREFIX"));
    }
  }
  
  if (tessdata_prefix_value == NULL)
  {
    tessdata_prefix_value = g_strdup (OCR_PREFERENCES_TESSDATA_PREFIX_DEFAULT_1);
    if (tessdata_prefix_value != NULL)
    {
      dir = g_dir_open (tessdata_prefix_value,
			0,
			NULL);
      if (dir == NULL)
      {
	g_free (tessdata_prefix_value);
	tessdata_prefix_value = g_strdup (OCR_PREFERENCES_TESSDATA_PREFIX_DEFAULT_2);
      }
      else
	g_dir_close (dir);
    }
  }

  dir = g_dir_open (tessdata_prefix_value,
		    0,
		    NULL);

  if (dir != NULL)
  {
    do
    {
      file_name = g_dir_read_name (dir);

      if (file_name != NULL)
      {
	tokens = g_strsplit_set (file_name,
				 ".",
				 -1);

	if (tokens[0] != NULL)
	{
	  if (tokens[1] != NULL &&
	      !g_strcmp0 (tokens[1], dic_file_name))
	  {
	    token = g_strdup (tokens[0]);
	    languages_list = g_slist_prepend (languages_list, token);
	    ++n_lang;
	  }
	}

	g_strfreev (tokens);
      }
    }
    while (file_name != NULL);
  }

  if (n_lang == 0)
  {
    installed_languages = g_new (gchar *, 1);
    installed_languages[0] = NULL;
  }
  else
  {
    installed_languages = g_new (gchar *, n_lang + 1);
    installed_languages[n_lang] = NULL;
    for (it = languages_list; it != NULL; it = it->next)
      installed_languages[--n_lang] = it->data;
  }

  if (languages_list != NULL)
    g_slist_free (languages_list);

  if (tessdata_prefix_value != NULL)
    g_free (tessdata_prefix_value);
  if (dir != NULL)
    g_dir_close(dir);

  return installed_languages;
}

gchar *
ocr_preferences_get_tesseract_valid_language_code (OcrPreferences *self)
{
  OcrPreferencesPrivate  *priv;
  gchar                  *str;
  gchar                 **installed_languages;
  guint                   i;
  gboolean                found;
  GError                 *error = NULL;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  str = ocr_preferences_get_tesseract_language_code (self);
  if (!g_strcmp0 (str, "") ||
      !g_strcmp0 (str, _("NONE")))
  {
    gchar *message = g_strdup_printf (_("Set a language in preferences window for better results with Tesseract."));

    ocr_utils_message (_("No language set for Tesseract"),
		       message,
		       GTK_MESSAGE_WARNING);

    g_free (message);
    g_free (str);
    str = g_strdup ("NOT_SET");
  }
  else
  {
    installed_languages = ocr_preferences_find_tesseract_installed_languages (self);
    found = FALSE;
    for (i = 0; installed_languages[i] != NULL; i++)
    {
      if (!g_strcmp0 (installed_languages[i], str))
      {
	found = TRUE;
	break;
      }
    }

    if (!found)
    {
      gchar *message = g_strdup_printf (_("«%s» language files for Tesseract not installed.\nPlease, choose another language in preferences window."),
					str);

      ocr_utils_message (_("Tesseract language files not available"),
			 message,
			 GTK_MESSAGE_ERROR);

      g_free (message);
      g_free (str);
      str = NULL;
    }
  }

  return str;
}

static void
ocr_preferences_create_tesseract_language_list (OcrPreferences *self,
						GtkBuilder     *builder)
{
  OcrPreferencesPrivate  *priv;
  GtkWidget              *vbox;
  GtkWidget              *widget;
  GtkWidget              *radio1;
  GtkWidget              *radio;
  GtkWidget              *child;
  gchar                 **installed_languages;
  guint                   i;

  g_return_if_fail (OCR_IS_PREFERENCES (self));
  g_return_if_fail (GTK_IS_BUILDER (builder));

  priv = self->priv;

  ocr_preferences_destroy_tesseract_language_list (self);

  vbox = gtk_vbox_new (FALSE, 3);

  installed_languages = ocr_preferences_find_tesseract_installed_languages (self);

  if (installed_languages[0] != NULL)
  {
    if (radio1 == NULL)
      radio = radio1 = gtk_radio_button_new_with_label (NULL,
							_("NONE"));
    else
      radio = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (radio1),
							   _("NONE"));

    priv->tesseract_language_list = g_list_append (priv->tesseract_language_list,
						   GTK_RADIO_BUTTON (radio));
    gtk_box_pack_start (GTK_BOX (vbox),
			radio,
			FALSE,
			FALSE,
			0);

    for (i = 0; installed_languages[i] != NULL; i++)
    {
      if (radio1 == NULL)
	radio = radio1 = gtk_radio_button_new_with_label (NULL,
							  installed_languages[i]);
      else
	radio = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (radio1),
							     installed_languages[i]);

      priv->tesseract_language_list = g_list_append (priv->tesseract_language_list,
						     GTK_RADIO_BUTTON (radio));
      gtk_box_pack_start (GTK_BOX (vbox),
			  radio,
			  FALSE,
			  FALSE,
			  0);
    }
  }
  else
  {
    widget = gtk_label_new (_("No language files found."));
    gtk_label_set_justify (GTK_LABEL (widget),
			   GTK_JUSTIFY_LEFT);

    gtk_box_pack_start (GTK_BOX (vbox),
			widget,
			FALSE,
			FALSE,
			0);

  }

  g_strfreev (installed_languages);
  
  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_languages_list"));
  child = gtk_bin_get_child (GTK_BIN (widget));
  if (child != NULL)
    gtk_widget_destroy (child);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (widget),
					 vbox);
  gtk_widget_show_all (widget);  /* needed for tesseract_dictionary_path_find button */
}

static void
ocr_preferences_destroy_tesseract_language_list (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GList                 *it;
  GtkWidget             *radio_button;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  if (priv->tesseract_language_list != NULL)
  {
    for (it = priv->tesseract_language_list; it != NULL; it = it->next)
    {
      if (it->data != NULL)
      {
	radio_button = GTK_WIDGET (it->data);
	gtk_widget_destroy (radio_button);
	it->data = NULL;
      }
    }
  }

  if (priv->tesseract_language_list != NULL)
  {
    g_list_free (priv->tesseract_language_list);
    priv->tesseract_language_list = NULL;
  }
}

static void
ocr_preferences_create_hunspell_dictionary_list (OcrPreferences *self,
						 GtkBuilder     *builder)
{
  OcrPreferencesPrivate  *priv;
  gchar                  *standard_error;
  gchar                 **tokens;
  gint                    i;
  gboolean                flag = FALSE;
  GtkWidget              *vbox;
  gchar                  *message;
  gchar                  *message_markup;
  GtkWidget              *widget;
  GtkWidget              *radio1;
  GtkWidget              *radio;
  GError                 *error = NULL;

  g_return_if_fail (OCR_IS_PREFERENCES (self));
  g_return_if_fail (GTK_IS_BUILDER (builder));

  priv = self->priv;

  vbox = gtk_vbox_new (FALSE, 3);

  g_spawn_command_line_sync ("hunspell -D",
			     NULL,
			     &(standard_error),
			     NULL,
			     &error);
  if (error != NULL)
  {
    message = g_strdup_printf (_("Error looking for available dictionaries: %s"),
			       error->message);
    message_markup = g_strdup_printf ("<span color='#FF0000'><b>%s</b></span>",
				      message);

    widget = gtk_label_new (NULL);
    gtk_label_set_markup (GTK_LABEL (widget),
			  message_markup);
    gtk_label_set_justify (GTK_LABEL (widget),
			   GTK_JUSTIFY_LEFT);

    gtk_box_pack_start (GTK_BOX (vbox),
			widget,
			FALSE,
			FALSE,
			0);

    g_clear_error (&error);
    g_free (message);
  }
  else
  {
    tokens = g_strsplit_set (standard_error,
			     "\n",
			     -1);

    flag = FALSE;
    radio1 = NULL;
    for (i = 0; tokens[i] != NULL; i++)
    {
      if (flag)
      {
	if (tokens[i][0] != '/')  /* path list is finished */
	  break;
      }

      if (tokens[i][0] == '/')  /* it is a path */
      {
	flag = TRUE;

	if (radio1 == NULL)
	  radio = radio1 = gtk_radio_button_new_with_label (NULL,
							    tokens[i]);

	radio = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (radio1),
							     tokens[i]);

	priv->hunspell_dictionary_list = g_list_append (priv->hunspell_dictionary_list,
							GTK_RADIO_BUTTON (radio));
	gtk_box_pack_start (GTK_BOX (vbox),
			    radio,
			    FALSE,
			    FALSE,
			    0);
      }
    }

    g_strfreev (tokens);

 
    if (radio1 == NULL)
    {
      widget = gtk_label_new (_("No available dictionary found."));
      gtk_label_set_justify (GTK_LABEL (widget),
			     GTK_JUSTIFY_LEFT);

      gtk_box_pack_start (GTK_BOX (vbox),
			  widget,
			  FALSE,
			  FALSE,
			  0);
    }
  }

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "hunspell_dictionary_list"));
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (widget),
					 vbox);
}

static void
ocr_preferences_dialog_response (GtkWidget      *dialog,
				 gint            response_id,
				 OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  gchar                 *dictionary_path;

  g_return_if_fail (GTK_IS_DIALOG (dialog));
  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  ocr_preferences_get_keys_from_dialog (self,
					"text_font",
					"layout_v",
					"layout_h",
					"save_default_folder",
					"remove_nl",
					"substitute_with_space",
					"number_blank_rows",
					"temp_folder",
					"preferred_tesseract",
					"preferred_gocr",
					"tesseract_language",
					"tesseract_dictionary_path",
					"gocr_l",
					"gocr_l_autodetect",
					"gocr_d",
					"gocr_d_autodetect",
					"gocr_s",
					"gocr_s_autodetect",
					"gocr_a",
					"gocr_a_default",
					"hunspell_dictionary_path",
					NULL);

  ocr_preferences_write_preferences (self);

  dictionary_path = ocr_preferences_get_hunspell_dictionary_path (self);
  ocr_spellcheck_change_dictionary (OCR_SPELLCHECK (ocr_window_get_spellcheck (priv->main_window)),
				    dictionary_path);

  if (dictionary_path != NULL)
    g_free (dictionary_path);

  builder = ocr_preferences_get_dialog_builder (self);

  g_assert (GTK_IS_BUILDER (builder));

  g_object_unref (builder);

  g_list_free (priv->hunspell_dictionary_list);
  priv->hunspell_dictionary_list = NULL;

  g_list_free (priv->tesseract_language_list);
  priv->tesseract_language_list = NULL;

  g_object_set_data (G_OBJECT (self),
		     OCR_PREFERENCES_DIALOG_KEY,
		     NULL);

  gtk_widget_destroy (dialog);
}

void
ocr_preferences_create_dialog (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  GtkBuilder            *builder;
  GtkWidget             *dialog;
  GtkWidget             *widget;
  GtkWidget             *widget_1;
  gboolean               value;
  gpointer               data;
  GError                *error;

  g_return_if_fail (OCR_IS_PREFERENCES (self));

  priv = self->priv;

  data = g_object_get_data (G_OBJECT (self),
			    OCR_PREFERENCES_DIALOG_KEY);

  if (data != NULL)
  {
    g_return_if_fail (GTK_IS_WINDOW (data));

    gtk_window_present (GTK_WINDOW (data));

    return;
  }

  builder = gtk_builder_new ();

  g_assert (GTK_IS_BUILDER (builder));

  error = NULL;
  gtk_builder_add_from_file (builder,
			     OCR_UI_DIR"/"OCR_PREFERENCES_BUILDER_FILE,
			     &error);

  if (error != NULL)
  {
    g_error (error->message,
	     NULL);
    g_clear_error (&error);
    return;
  }

  dialog = GTK_WIDGET (gtk_builder_get_object (builder,
					       "preferences_dialog"));

  g_assert (GTK_IS_DIALOG (dialog));

  g_signal_connect (dialog,
		    "response", 
		    G_CALLBACK (ocr_preferences_dialog_response),
		    (gpointer) self);

  g_object_set_data (G_OBJECT (self),
		     OCR_PREFERENCES_DIALOG_BUILDER_KEY,
		     (gpointer) builder);
  g_object_set_data (G_OBJECT (self),
                     OCR_PREFERENCES_DIALOG_KEY,
                     (gpointer) dialog);

  ocr_preferences_check_missing_programs (self);

  ocr_preferences_create_tesseract_language_list (self,
						  builder);

  ocr_preferences_create_hunspell_dictionary_list (self,
						   builder);

  ocr_preferences_set_keys_on_dialog (self,
				      "text_font",
				      "layout_v",
				      "layout_h",
				      "save_default_folder",
				      "remove_nl",
				      "substitute_with_space",
				      "number_blank_rows",
				      "temp_folder",
				      "preferred_tesseract",
				      "preferred_gocr",
				      "tesseract_dictionary_path",
				      "gocr_l",
				      "gocr_l_autodetect",
				      "gocr_d",
				      "gocr_d_autodetect",
				      "gocr_s",
				      "gocr_s_autodetect",
				      "gocr_a",
				      "gocr_a_default",
				      "hunspell_dictionary_path",
				      NULL);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "text_font"));
  g_signal_connect (widget, "font-set",
		    G_CALLBACK (ocr_preferences_text_font_font_set),
		    self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "layout_v"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_layout_v_toggled),
		    self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "remove_nl"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_remove_nl_toggled),
		    builder);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "substitute_with_space"));
  value = ocr_preferences_get_remove_nl (self);
  gtk_widget_set_sensitive (widget,
			    value);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_dictionary_path"));
  g_signal_connect (widget, "key-press-event",
		    G_CALLBACK (ocr_preferences_tesseract_path_pressed_key),
		    self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_dictionary_path_find"));
  g_signal_connect (widget, "pressed",
		    G_CALLBACK (ocr_preferences_tesseract_directory_path_find_pressed),
		    self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_dictionary_path_help"));
  g_signal_connect (widget, "pressed",
		    G_CALLBACK (ocr_preferences_tesseract_directory_path_help_pressed),
		    self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "tesseract_site_button"));
  g_signal_connect (widget, "pressed",
		    G_CALLBACK (ocr_preferences_tesseract_site_button_pressed),
		    NULL);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_l_autodetect"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_gocr_l_autodetect_toggled),
		    self);
  value = ocr_preferences_get_gocr_gray_level_autodetect (self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_l"));
  gtk_widget_set_sensitive (widget,
			    !value);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_d_autodetect"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_gocr_d_autodetect_toggled),
		    self);
  value = ocr_preferences_get_gocr_dust_size_autodetect (self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_d"));
  gtk_widget_set_sensitive (widget,
			    !value);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_s_autodetect"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_gocr_s_autodetect_toggled),
		    self);
  value = ocr_preferences_get_gocr_spacewidth_between_words_autodetect (self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_s"));
  gtk_widget_set_sensitive (widget,
			    !value);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_a_default"));
  g_signal_connect (widget, "toggled",
		    G_CALLBACK (ocr_preferences_gocr_a_default_toggled),
		    self);
  value = ocr_preferences_get_gocr_value_for_certainty_of_recognition_default (self);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_a"));
  gtk_widget_set_sensitive (widget,
			    !value);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "gocr_site_button"));
  g_signal_connect (widget, "pressed",
		    G_CALLBACK (ocr_preferences_gocr_site_button_pressed),
		    NULL);

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "hunspell_dictionary_path_help"));
  g_signal_connect (widget, "pressed",
		    G_CALLBACK (ocr_preferences_hunspell_dictionary_path_help_pressed),
		    self);

  gtk_widget_show_all (dialog);
}

gchar *
ocr_preferences_get_separator (OcrPreferences *self)
{
  OcrPreferencesPrivate *priv;
  gint                   number_of_rows;
  gchar                 *separator;
  gchar                 *temp;
  gint                   i;
  
  g_return_val_if_fail (OCR_IS_PREFERENCES (self), NULL);

  priv = self->priv;

  number_of_rows = (gint) ocr_preferences_get_number_blank_rows (self);

  separator = NULL;
  if (number_of_rows > 0)
  {
    separator = g_strdup ("\n\n");
    for (i = 1; i < number_of_rows; i++)
    {
      temp = g_strdup (separator);
      g_free (separator);
      separator = g_strconcat (temp, "\n", NULL);
      g_free (temp);
    }
  }

  return separator;
}

