#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <hunspell/hunspell.h>

#include "ocr-spellcheck.h"
#include "ocr-define.h"

#define OCR_SPELLCHECK_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_SPELLCHECK, OcrSpellcheckPrivate))

G_DEFINE_TYPE (OcrSpellcheck, ocr_spellcheck, G_TYPE_OBJECT);

typedef enum
{
  FROM_UTF8_TO_DIC_ENCODING = 1,
  FROM_DIC_ENCODING_TO_UTF8
} ConversionType;

struct _OcrSpellcheckPrivate
{
  Hunhandle *pHunspell;
};

static void
ocr_spellcheck_dispose (GObject *gobject)
{
  OcrSpellcheck        *self;
  OcrSpellcheckPrivate *priv;
  static gboolean       first_run = TRUE;

  g_return_if_fail (OCR_IS_SPELLCHECK (gobject));

  self = OCR_SPELLCHECK (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->pHunspell != NULL)
    {
      Hunspell_destroy (priv->pHunspell);
      priv->pHunspell = NULL;
    }

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_spellcheck_parent_class)->dispose (gobject);
}

static void
ocr_spellcheck_finalize (GObject *gobject)
{
  OcrSpellcheck        *self;
  OcrSpellcheckPrivate *priv;

  g_return_if_fail (OCR_IS_SPELLCHECK (gobject));

  self = OCR_SPELLCHECK (gobject);
  priv = self->priv;

  if (self->dictionary_path != NULL)
    g_free (self->dictionary_path);

  G_OBJECT_CLASS (ocr_spellcheck_parent_class)->finalize (gobject);
}

void
ocr_spellcheck_create (OcrSpellcheck *self)
{
  OcrSpellcheckPrivate *priv;
  gchar                *affixes, *dictionary;
  
  g_return_if_fail (OCR_IS_SPELLCHECK (self));

  priv = self->priv;

  if (priv->pHunspell != NULL)
  {
    Hunspell_destroy (priv->pHunspell);
    priv->pHunspell = NULL;
  }

  if (self->dictionary_path != NULL)
  {
    affixes = g_strdup_printf ("%s.aff",
			       self->dictionary_path);
    dictionary = g_strdup_printf ("%s.dic",
				  self->dictionary_path);
    priv->pHunspell = Hunspell_create(affixes,
				      dictionary);
    g_free (affixes);
    g_free (dictionary);

    g_assert (priv->pHunspell != NULL);
  }
}

void
ocr_spellcheck_change_dictionary (OcrSpellcheck *self,
				  const gchar   *dictionary_path)
{
  OcrSpellcheckPrivate *priv;

  g_return_if_fail (OCR_IS_SPELLCHECK (self));

  priv = self->priv;

  if (self->dictionary_path == NULL ||
      g_strcmp0 (dictionary_path,
		 self->dictionary_path))
  {
    if (self->dictionary_path != NULL)
      g_free (self->dictionary_path);
    self->dictionary_path = g_strdup (dictionary_path);

    ocr_spellcheck_create (self);
  }
}

static void
ocr_spellcheck_class_init (OcrSpellcheckClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrSpellcheckPrivate));

  gobject_class->dispose = ocr_spellcheck_dispose;
  gobject_class->finalize = ocr_spellcheck_finalize;
}

static void
ocr_spellcheck_init (OcrSpellcheck *self)
{
  OcrSpellcheckPrivate *priv;

  g_return_if_fail (OCR_IS_SPELLCHECK (self));

  self->priv = OCR_SPELLCHECK_GET_PRIVATE (self);
  priv = self->priv;

  self->dictionary_path = NULL;

  priv->pHunspell       = NULL;
}

gboolean
ocr_spellcheck_is_ok (OcrSpellcheck *self)
{
  OcrSpellcheckPrivate *priv;
  gboolean retval = TRUE;

  g_return_val_if_fail (OCR_IS_SPELLCHECK (self), FALSE);

  priv = self->priv;

  if (self->dictionary_path == NULL)
  {
    ocr_utils_message (_("No dictionary found"),
		       _("Install a dictionary to use spell check.\nLook at OcrGui preferences."),
		       GTK_MESSAGE_ERROR);
    retval = FALSE;
  }
  else 
  {
    if (priv->pHunspell == NULL)
      retval = FALSE;
  }

  return retval;
}

static const gchar *
ocr_spellcheck_get_dic_encoding (OcrSpellcheck *self)
{
  OcrSpellcheckPrivate *priv;

  g_return_val_if_fail (OCR_IS_SPELLCHECK (self), NULL);

  priv = self->priv;

  return Hunspell_get_dic_encoding (priv->pHunspell);
}

static gchar *
ocr_spellcheck_convert_encoding (OcrSpellcheck  *self,
				 ConversionType  type,
				 const gchar    *word)
{
  gchar  *word_conv = NULL;
  GError *error = NULL;

  g_return_val_if_fail (word != NULL, NULL);

  switch (type)
  {
    case FROM_UTF8_TO_DIC_ENCODING:
      word_conv = g_convert (word, -1,
			     ocr_spellcheck_get_dic_encoding (self),
			     "UTF-8",
			     NULL,
			     NULL,
			     &error);

      if (error != NULL)
      {
	if (word_conv != NULL)
	  g_free (word_conv);
	word_conv = g_strdup (word);
      }

      g_clear_error (&error);
      break;
    case FROM_DIC_ENCODING_TO_UTF8:
      word_conv = g_convert (word, -1,
			     "UTF-8",
			     ocr_spellcheck_get_dic_encoding (self),
			     NULL,
			     NULL,
			     &error);

      if (error != NULL)
      {
	if (word_conv != NULL)
	  g_free (word_conv);
	word_conv = g_strdup (word);
      }

      g_clear_error (&error);
      break;
    default:
      g_assert_not_reached();
      break;
  }

  return word_conv;
}

gboolean
ocr_spellcheck_check_word (OcrSpellcheck *self,
			   const gchar   *word)
{
  gchar    *word_conv;
  gboolean  ret_val;

  g_return_val_if_fail (OCR_IS_SPELLCHECK (self), FALSE);

  word_conv = ocr_spellcheck_convert_encoding (self,
					       FROM_UTF8_TO_DIC_ENCODING,
					       word);

  ret_val = (gboolean) Hunspell_spell (self->priv->pHunspell,
				       word_conv);

  g_free (word_conv);

  return ret_val; 
}

GList *
ocr_spellcheck_suggest (OcrSpellcheck *self,
			guint         *n_suggestions,
			const gchar   *word)
{
  OcrSpellcheckPrivate  *priv;
  GList                 *list = NULL;
  gchar                 *word_conv;
  guint                  suggestions_number, i;
  char                 **suggestions;

  g_return_val_if_fail (OCR_IS_SPELLCHECK (self), FALSE);

  priv = self->priv;

  word_conv = ocr_spellcheck_convert_encoding (self,
					       FROM_UTF8_TO_DIC_ENCODING,
					       word);
  suggestions_number = Hunspell_suggest (priv->pHunspell,
					 &suggestions,
					 word_conv);
  g_free (word_conv);

  for (i = 0; i < suggestions_number; i++)
  {
    word_conv = ocr_spellcheck_convert_encoding (self,
						 FROM_DIC_ENCODING_TO_UTF8,
						 suggestions[i]);
    list = g_list_prepend (list, g_strdup (word_conv));
    g_free (word_conv);
  }
  /* TODO - This function is only in new version of Hunspell API 
  Hunspell_free_list (priv->pHunspell, &suggestions, suggestions_number);*/

  if (n_suggestions)
    *n_suggestions = suggestions_number; 

  return g_list_reverse (list);
}
