#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-textbuffer.h"
#include "ocr-spellcheck.h"
#include "ocr-utils.h"

#define OCR_TEXTBUFFER_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_TEXTBUFFER, OcrTextBufferPrivate))

G_DEFINE_TYPE (OcrTextBuffer, ocr_textbuffer, GTK_TYPE_TEXT_BUFFER);

enum
{
  PROP_SPELLCHECK = 1,
};

struct _OcrTextBufferPrivate
{
  OcrSpellcheck *spellcheck;
  GtkTextMark   *click_mark;
  GtkTextMark   *spellcheck_mark;
  GtkTextMark   *word_start_mark, *word_end_mark;
};

static void
ocr_textbuffer_dispose (GObject *gobject)
{
  OcrTextBuffer        *self;
  OcrTextBufferPrivate *priv;
  static gboolean       first_run = TRUE;

  g_return_if_fail (OCR_IS_TEXTBUFFER (gobject));

  self = OCR_TEXTBUFFER (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->spellcheck != NULL)
      g_object_unref (priv->spellcheck);

    if (priv->click_mark != NULL &&
	GTK_IS_TEXT_MARK (priv->click_mark))
      gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (self),
				   priv->click_mark);

    if (priv->spellcheck_mark != NULL &&
	GTK_IS_TEXT_MARK (priv->spellcheck_mark))
      gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (self),
				   priv->spellcheck_mark);

    if (priv->word_start_mark != NULL &&
	GTK_IS_TEXT_MARK (priv->word_start_mark))
      gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (self),
				   priv->word_start_mark);

    if (priv->word_end_mark != NULL &&
	GTK_IS_TEXT_MARK (priv->word_end_mark))
      gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (self),
				   priv->word_end_mark);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_textbuffer_parent_class)->dispose (gobject);
}

static void
ocr_textbuffer_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_TEXTBUFFER (gobject));

  G_OBJECT_CLASS (ocr_textbuffer_parent_class)->finalize (gobject);
}

static void
ocr_textbuffer_constructed (GObject *gobject)
{
  OcrTextBuffer        *self;
  OcrTextBufferPrivate *priv;
  GtkTextIter           iter;

  g_return_if_fail (OCR_IS_TEXTBUFFER (gobject));

  self = OCR_TEXTBUFFER (gobject);
  priv = self->priv;

  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				  &iter);
  priv->click_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
						  "click_mark",
						  &iter,
						  TRUE);
}

static void
ocr_textbuffer_set_property (GObject      *gobject,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  OcrTextBuffer        *self;
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (gobject));

  self = OCR_TEXTBUFFER (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_SPELLCHECK:
      if (value != NULL)
      {
	g_return_if_fail (OCR_IS_SPELLCHECK (g_value_get_object (value)));
	priv->spellcheck = OCR_SPELLCHECK (g_value_dup_object (value));
      }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_textbuffer_class_init (OcrTextBufferClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrTextBufferPrivate));

  gobject_class->constructed = ocr_textbuffer_constructed;
  gobject_class->set_property = ocr_textbuffer_set_property;
  gobject_class->dispose = ocr_textbuffer_dispose;
  gobject_class->finalize = ocr_textbuffer_finalize;

  g_object_class_install_property (gobject_class, PROP_SPELLCHECK,
				   g_param_spec_object ("spellcheck", "spellcheck", "spellcheck",
							OCR_TYPE_SPELLCHECK,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

static void
ocr_textbuffer_init (OcrTextBuffer *self)
{
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  self->priv = OCR_TEXTBUFFER_GET_PRIVATE (self);
  priv = self->priv;

  priv->spellcheck      = NULL;
  priv->click_mark      = NULL;
  priv->spellcheck_mark = NULL;
  priv->word_start_mark = NULL;
  priv->word_end_mark   = NULL;
}

void
ocr_textbuffer_get_iter_at_click_mark (OcrTextBuffer *self,
				       GtkTextIter   *iter)
{
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  priv = self->priv;
  
  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    iter,
				    priv->click_mark);
}

void
ocr_textbuffer_set_click_mark (OcrTextBuffer *self,
			       GtkTextIter   *iter) 
{
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  priv = self->priv;

  gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
			     priv->click_mark,
			     iter);
}

void
ocr_textbuffer_set_word_start_end_mark (OcrTextBuffer *self,
					GtkTextIter   *word_start,
					GtkTextIter   *word_end)
{
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  priv = self->priv;

  if (priv->word_start_mark == NULL)
    priv->word_start_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
							 "word_start_mark",
							 word_start,
							 TRUE);
  else
    gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
			       priv->word_start_mark,
			       word_start);

  if (priv->word_end_mark == NULL)
    priv->word_end_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
						       "word_end_mark",
						       word_end,
						       TRUE);
  else
    gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
			       priv->word_end_mark,
			       word_end);
}

gboolean
ocr_textbuffer_get_bad_word (OcrTextBuffer  *self,
			     gchar         **word)
{
  OcrTextBufferPrivate *priv;
  GtkTextIter           click_iter;
  GtkTextIter           word_start, word_end;

  g_return_val_if_fail (OCR_IS_TEXTBUFFER (self), FALSE);

  priv = self->priv;

  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &click_iter,
				    priv->click_mark);

  if (gtk_text_iter_has_tag (&click_iter,
			     gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
							"mispelled-word-underline")))
  {
    word_start = word_end = click_iter;
    if (gtk_text_iter_starts_word (&click_iter))
      gtk_text_iter_forward_word_end (&word_end);
    else if (gtk_text_iter_ends_word (&click_iter))
      gtk_text_iter_backward_word_start (&word_start);
    else if (gtk_text_iter_inside_word (&click_iter))
    {
      gtk_text_iter_backward_word_start (&word_start);
      gtk_text_iter_forward_word_end (&word_end);
    }

    ocr_textbuffer_set_word_start_end_mark (self,
					    &word_start,
					    &word_end);

    *word = gtk_text_buffer_get_text (GTK_TEXT_BUFFER (self),
				     &word_start,
				     &word_end,
				     TRUE);
    return TRUE;
  }
  else
    return FALSE;
}

gboolean
ocr_textbuffer_spellcheck_put_underline_tag (OcrTextBuffer *self)
{
  OcrTextBufferPrivate *priv;
  GtkTextTag           *texttag_underline;
  GtkTextIter           start, end;
  GtkTextIter           word_start, word_end, word_end_previous;
  gchar                *word;
  gboolean              good_word;
  gboolean              retval;

  g_return_val_if_fail (OCR_IS_TEXTBUFFER (self), FALSE);

  priv = self->priv;

  texttag_underline = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
						 "mispelled-word-underline");
  gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (self), &start, &end);

  retval = FALSE;

  word_end = start;
  word_end_previous = start;
  while (gtk_text_iter_compare (&word_end, &end) < 0)
  {
    gtk_text_iter_forward_word_end (&word_end);
    word_start = word_end;
    gtk_text_iter_backward_word_start (&word_start);

    /* this control is needed for symbols (points, commas, question marks, etc.) and spaces after last word */
    if (gtk_text_iter_compare (&word_end_previous, &word_start) > 0)
      break;

    /* remove tags if a previous spellcheck applied them
     * (for example when you change hunspell dictionary) */
    ocr_textbuffer_set_word_start_end_mark (self,
					    &word_start,
					    &word_end);

    ocr_textbuffer_remove_tags (self,
				TRUE,
				FALSE);

    /* apply tags if a word is not correct */
    word_end_previous = word_end;

    word = gtk_text_buffer_get_text (GTK_TEXT_BUFFER (self),
				     &word_start,
				     &word_end,
				     TRUE);

    good_word = ocr_spellcheck_check_word (priv->spellcheck,
					   word);
    if (!good_word)
    {
      gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER (self),
				 texttag_underline,
				 &word_start,
				 &word_end);

      retval = TRUE;
    }

    g_free (word);
  }

  return retval;
}

void
ocr_textbuffer_clear_spellcheck_mark (OcrTextBuffer *self)
{
  OcrTextBufferPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  priv = self->priv;

  if (priv->spellcheck_mark != NULL &&
      GTK_IS_TEXT_MARK (priv->spellcheck_mark))
    gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (self),
				 priv->spellcheck_mark);

  priv->spellcheck_mark = NULL;
}

gboolean
ocr_textbuffer_get_next_bad_word (OcrTextBuffer  *self,
				  gchar         **word,
				  GList         **suggestions,
				  guint          *suggestions_number)
{
  OcrTextBufferPrivate *priv;
  GtkTextTag           *texttag_underline;
  GtkTextTag           *texttag_background;
  GtkTextIter           start, end;
  GtkTextIter           word_start, word_end;
  gboolean              retval;

  g_return_val_if_fail (OCR_IS_TEXTBUFFER (self), FALSE);

  priv = self->priv;

  texttag_underline = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
						 "mispelled-word-underline");
  texttag_background = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
						  "mispelled-word-during-check");

  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (self), &end);

  if (priv->spellcheck_mark == NULL)
  {
    gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				    &start);
    priv->spellcheck_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
							 "spellcheck_mark",
							 &start,
							 TRUE);
  }
  else
  {
    gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				      &start,
				      priv->spellcheck_mark);
  }

  retval = FALSE;

  word_end = start;
  while (gtk_text_iter_compare (&word_end, &end) < 0)
  {
    gtk_text_iter_forward_word_end (&word_end);
    word_start = word_end;
    gtk_text_iter_backward_word_start (&word_start);
    if (gtk_text_iter_has_tag (&word_start,
			       texttag_underline))
    {
      *word = gtk_text_buffer_get_text (GTK_TEXT_BUFFER (self),
				       &word_start,
				       &word_end,
				       TRUE);

      gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER (self),
				 texttag_background, 
				 &word_start,
				 &word_end);

      *suggestions = ocr_spellcheck_suggest (priv->spellcheck,
					     suggestions_number,
					     *word);

      gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
				 priv->spellcheck_mark,
				 &word_end);

      ocr_textbuffer_set_word_start_end_mark (self,
					      &word_start,
					      &word_end);

      retval = TRUE;
      break;
    }
  }

  return retval;
}

void
ocr_textbuffer_remove_tags (OcrTextBuffer *self,
			    gboolean       remove_underline,
			    gboolean       remove_during_check)
{
  OcrTextBufferPrivate *priv;
  GtkTextIter           word_start, word_end;
  GtkTextTag           *texttag_underline;
  GtkTextTag           *texttag_background;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  priv = self->priv;

  texttag_underline = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
						 "mispelled-word-underline");
  texttag_background = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (self)),
						  "mispelled-word-during-check");

  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &word_start,
				    priv->word_start_mark);
  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &word_end,
				    priv->word_end_mark);

  if (remove_underline == TRUE &&
      gtk_text_iter_has_tag (&word_start,
			     texttag_underline))
  {
    gtk_text_buffer_remove_tag (GTK_TEXT_BUFFER (self),
				texttag_underline,
				&word_start,
				&word_end);
  }

  if (remove_during_check == TRUE &&
      gtk_text_iter_has_tag (&word_start,
			     texttag_background))
  {
    gtk_text_buffer_remove_tag (GTK_TEXT_BUFFER (self),
				texttag_background,
				&word_start,
				&word_end);
  }
}

void
ocr_textbuffer_substitute_word	(OcrTextBuffer *self,
				 const gchar   *new_word)
{
  OcrTextBufferPrivate *priv;
  GtkTextIter           word_start, word_end;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));
  g_return_if_fail (new_word != NULL);

  priv = self->priv;

  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &word_start,
				    priv->word_start_mark);
  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &word_end,
				    priv->word_end_mark);

  gtk_text_buffer_delete (GTK_TEXT_BUFFER (self),
			  &word_start,
			  &word_end); /* the start iter and end iter will be re-initialized
					 to point to the location where text was deleted */
  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self),
				    &word_start,
				    priv->word_start_mark);
  gtk_text_buffer_insert (GTK_TEXT_BUFFER (self),
			  &word_start,
			  new_word,
			  -1);
}

gchar *
ocr_textbuffer_get_content (OcrTextBuffer *self)
{
  GtkTextIter  start, end;
  gchar       *content = NULL;

  g_return_val_if_fail (OCR_IS_TEXTBUFFER (self), NULL);

  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				  &start);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (self),
				&end);
  content = gtk_text_buffer_get_text (GTK_TEXT_BUFFER (self),
				      &start,
				      &end,
				      FALSE);

  return content;
}

void
ocr_textbuffer_load_content (OcrTextBuffer *self,
			     GFile         *file)
{
  GtkTextIter  start, end;
  gchar       *content = NULL;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));
  g_return_if_fail (G_IS_FILE (file));

  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				  &start);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (self),
				&end);
  gtk_text_buffer_delete (GTK_TEXT_BUFFER (self),
			  &start,
			  &end);

  ocr_utils_load_file (file,
		       &content);

  if (content != NULL)
  {
    gtk_text_buffer_set_text (GTK_TEXT_BUFFER (self),
			      content,
			      -1);
  
  }

  g_free (content);
}

GtkTextChildAnchor *
ocr_textbuffer_load_message (OcrTextBuffer *self,
			     const gchar   *message)
{
  GtkTextIter  start, end;
  GtkTextChildAnchor *image_anchor;

  g_return_if_fail (OCR_IS_TEXTBUFFER (self));

  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				  &start);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (self),
				&end);
  gtk_text_buffer_delete (GTK_TEXT_BUFFER (self),
			  &start,
			  &end);

  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self),
				  &start);

  image_anchor = gtk_text_buffer_create_child_anchor (GTK_TEXT_BUFFER (self),
						      &start);

  if (message != NULL)
  {
    gtk_text_buffer_insert (GTK_TEXT_BUFFER (self),
			    &start,
			    message,
			    -1);
  }

  return image_anchor;
}

