#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include "ocr-textview.h"
#include "ocr-textbuffer.h"

#define SPELLCHECK_DIALOG_BUILDER_FILE "ocr-spellcheck.ui"

#define OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY "ocr-textview-spellcheck-dialog-key"
#define OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY "ocr-textview-spellcheck-dialog-builder-key"

#define OCR_TEXTVIEW_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_TEXTVIEW, OcrTextViewPrivate))

G_DEFINE_TYPE (OcrTextView, ocr_textview, GTK_TYPE_TEXT_VIEW);

enum
{
  PROP_TEXTBUFFER = 1,
  PROP_SPELLCHECK
};

struct _OcrTextViewPrivate
{
  OcrTextBuffer *textbuffer;
  OcrSpellcheck *spellcheck;
};

static void
ocr_textview_dispose (GObject *gobject)
{
  OcrTextView        *self;
  OcrTextViewPrivate *priv;
  static gboolean     first_run = TRUE;

  g_return_if_fail (OCR_IS_TEXTVIEW (gobject));

  self = OCR_TEXTVIEW (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->textbuffer != NULL)
      g_object_unref (priv->textbuffer);

    if (priv->spellcheck != NULL)
      g_object_unref (priv->spellcheck);

    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_textview_parent_class)->dispose (gobject);
}

static void
ocr_textview_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_TEXTVIEW (gobject));

  G_OBJECT_CLASS (ocr_textview_parent_class)->finalize (gobject);
}

static void
ocr_textview_constructed (GObject *gobject)
{
  OcrTextView        *self;
  OcrTextViewPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTVIEW (gobject));

  self = OCR_TEXTVIEW (gobject);
  priv = self->priv;

  g_return_if_fail (GTK_IS_TEXT_BUFFER (priv->textbuffer));
  g_return_if_fail (OCR_IS_SPELLCHECK (priv->spellcheck));

  gtk_text_view_set_buffer (GTK_TEXT_VIEW (self),
			    GTK_TEXT_BUFFER (priv->textbuffer));
}

static void
ocr_textview_set_property (GObject      *gobject,
			   guint         property_id,
			   const GValue *value,
			   GParamSpec   *pspec)
{
  OcrTextView        *self;
  OcrTextViewPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTVIEW (gobject));

  self = OCR_TEXTVIEW (gobject);
  priv = self->priv;

  switch (property_id)
  {
    case PROP_TEXTBUFFER:
      g_return_if_fail (OCR_IS_TEXTBUFFER (g_value_get_object (value)));
      priv->textbuffer = OCR_TEXTBUFFER (g_value_dup_object (value));
      break;
    case PROP_SPELLCHECK:
      g_return_if_fail (OCR_IS_SPELLCHECK (g_value_get_object (value)));
      priv->spellcheck = OCR_SPELLCHECK (g_value_dup_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, property_id, pspec);
      break;
  }
}

static void
ocr_textview_class_init (OcrTextViewClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrTextViewPrivate));

  gobject_class->constructed = ocr_textview_constructed;
  gobject_class->set_property = ocr_textview_set_property;
  gobject_class->dispose = ocr_textview_dispose;
  gobject_class->finalize = ocr_textview_finalize;

  g_object_class_install_property (gobject_class, PROP_TEXTBUFFER,
				   g_param_spec_object ("textbuffer", "textbuffer", "textbuffer",
							GTK_TYPE_TEXT_BUFFER,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_SPELLCHECK,
				   g_param_spec_object ("spellcheck", "spellcheck", "spellcheck",
							OCR_TYPE_SPELLCHECK,
							G_PARAM_WRITABLE | G_PARAM_CONSTRUCT));
}

gboolean
ocr_textview_button_press_event (OcrTextView    *self,
	 			 GdkEventButton *eventbutton)
{
  if (eventbutton->button == 3 &&
      eventbutton->type == GDK_BUTTON_PRESS) /* this is the mouse right button */
  {
    OcrTextViewPrivate *priv;
    gint                x, y;
    GtkTextIter         iter;
  
    g_return_val_if_fail (OCR_IS_TEXTVIEW (self), TRUE);

    priv = self->priv;

    gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (self),
                                           GTK_TEXT_WINDOW_TEXT,
                                           eventbutton->x, eventbutton->y,
                                           &x, &y);

    gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (self), &iter, x, y);
    
    ocr_textbuffer_set_click_mark (priv->textbuffer,
				   &iter);
  }

  return FALSE; /* TRUE to stop other handlers from being invoked for the event
                   FALSE to propagate the event further */
}

static void
ocr_textview_ignore_word (GtkMenuItem *menuitem,
			  OcrTextView *self)
{
  OcrTextViewPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  priv = self->priv;

  ocr_textbuffer_remove_tags (priv->textbuffer,
			      TRUE,
			      TRUE);
}

static void
ocr_textview_substitute_word (GtkMenuItem *menuitem,
			      OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  const gchar        *new_word = NULL;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  priv = self->priv;

  new_word = gtk_label_get_text (GTK_LABEL (GTK_BIN (menuitem)->child));

  ocr_textbuffer_substitute_word (priv->textbuffer,
				  new_word);
}

static void
ocr_textview_create_suggestion_menu (OcrTextView *self,
				     GtkMenu     *menu,
				     const gchar *word)
{
  OcrTextViewPrivate *priv;
  GtkWidget          *_menu;
  GtkWidget          *menuitem;
  GtkWidget          *label;
  GList              *suggestions;
  guint               suggestions_number, i;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));
  g_return_if_fail (GTK_IS_MENU (menu));
  g_return_if_fail (word != NULL);

  priv = self->priv;

  if (!ocr_spellcheck_is_ok (priv->spellcheck))
    return;

  _menu = GTK_WIDGET (menu);

  suggestions = ocr_spellcheck_suggest (priv->spellcheck,
					&suggestions_number,
					word);
  
  if (suggestions == NULL || suggestions_number == 0)
  {
    gchar *label_str = g_strdup (_("No suggestions"));
    gchar *label_str_markup = g_strdup_printf ("<i>%s</i>",
					       label_str);
    label = gtk_label_new("");
    gtk_label_set_markup (GTK_LABEL(label),
			  label_str_markup);

    menuitem = gtk_separator_menu_item_new ();
    gtk_container_add (GTK_CONTAINER (menuitem), label);
    gtk_widget_show_all (menuitem);
    gtk_menu_shell_prepend (GTK_MENU_SHELL (_menu), menuitem);

    g_free (label_str);
    g_free (label_str_markup);
  }
  else
  {
    for (i = 0; i < suggestions_number; i++)
    {
      if ((i != 0) && (i % 10 == 0))
      {
	menuitem = gtk_separator_menu_item_new();
	gtk_menu_shell_append (GTK_MENU_SHELL (_menu), menuitem);

	menuitem = gtk_menu_item_new_with_label (_("More..."));
	gtk_menu_shell_append (GTK_MENU_SHELL (_menu), menuitem);

	_menu = gtk_menu_new();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), _menu);
      }

      menuitem = gtk_menu_item_new_with_label ((gchar *) g_list_nth_data (suggestions, i));
      gtk_menu_shell_append (GTK_MENU_SHELL (_menu), menuitem);

      g_signal_connect (G_OBJECT (menuitem),
			"activate",
			G_CALLBACK (ocr_textview_substitute_word),
			self);
    }
  }

  g_list_free (suggestions);
}

GtkWidget *
ocr_textview_create_spelling_menu (OcrTextView *self,
				   const gchar *word)
{
  OcrTextViewPrivate *priv;
  GtkWidget          *spelling_menu;
  GtkWidget          *menuitem;

  g_return_val_if_fail (OCR_IS_TEXTVIEW (self), NULL);
  g_return_val_if_fail (word != NULL, NULL);

  priv = self->priv;

  spelling_menu = gtk_menu_new();

  ocr_textview_create_suggestion_menu (self,
				       GTK_MENU (spelling_menu),
				       word);

  /* Separator */
  menuitem = gtk_separator_menu_item_new ();
  gtk_menu_shell_append (GTK_MENU_SHELL (spelling_menu), menuitem);

  /* + Add to Dictionary */
  /* TODO - Implement Add to Dictionary call
  label = g_strdup_printf("Add \"%s\" to Dictionary", word);
  menuitem = gtk_image_menu_item_new_with_label (label);
  g_free(label);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
				 gtk_image_new_from_stock (GTK_STOCK_ADD,
							   GTK_ICON_SIZE_MENU));
  gtk_menu_shell_append (GTK_MENU_SHELL (spelling_menu), menuitem);
  */

  /* - Ignore All */
  menuitem = gtk_image_menu_item_new_with_label(_("Ignore"));
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
				 gtk_image_new_from_stock (GTK_STOCK_REMOVE,
							   GTK_ICON_SIZE_MENU));
  g_signal_connect (G_OBJECT (menuitem),
		    "activate",
		    G_CALLBACK (ocr_textview_ignore_word),
		    self);

  gtk_menu_shell_append (GTK_MENU_SHELL (spelling_menu), menuitem);

  return spelling_menu;
}

void
ocr_textview_populate_popup (OcrTextView *self,
                             GtkMenu     *menu)
{
  OcrTextViewPrivate *priv;
  gchar              *word = NULL;
  gboolean            bad_word;
  GtkWidget          *menuitem, *icon;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));
  g_return_if_fail (GTK_IS_MENU (menu));

  priv = self->priv;

  bad_word = ocr_textbuffer_get_bad_word (priv->textbuffer,
					  &word);
  if (!bad_word)
    return;

  g_assert (word != NULL);

  /* Separator */
  menuitem = gtk_separator_menu_item_new();
  gtk_widget_show (menuitem);
  gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), menuitem);

  /* Above the separator, show the suggestions menu */
  icon = gtk_image_new_from_stock (GTK_STOCK_SPELL_CHECK, GTK_ICON_SIZE_MENU);
  menuitem = gtk_image_menu_item_new_with_label (_("Spelling suggestions"));
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem), icon);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem),
			     ocr_textview_create_spelling_menu (self, word));

  gtk_widget_show_all (menuitem);
  gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menuitem);

  g_free (word);
}

static void
ocr_textview_init (OcrTextView *self)
{
  OcrTextViewPrivate *priv;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  self->priv = OCR_TEXTVIEW_GET_PRIVATE (self);
  priv = self->priv;

  priv->textbuffer = NULL;
  priv->spellcheck = NULL;

  g_signal_connect (self,
		    "button-press-event",
		    G_CALLBACK (ocr_textview_button_press_event),
		    NULL);
  
  g_signal_connect (self,
		    "populate-popup",
		    G_CALLBACK (ocr_textview_populate_popup),
		    NULL);
}

static void
ocr_textview_spellcheck_dialog_put_data (OcrTextView *self,
					 gchar       *word,
					 GList       *suggestions,
					 guint        suggestions_number)
{
  GtkDialog    *dialog;
  GtkBuilder   *builder;
  GtkLabel     *label_misspelled_word;
  GtkListStore *suggestions_list;
  GtkTreeIter   iter;
  gchar        *markup;
  gint          i;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  dialog = g_object_get_data (G_OBJECT (self),
			      OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY);
  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);

  g_assert (GTK_IS_BUILDER (builder));

  label_misspelled_word = GTK_LABEL (gtk_builder_get_object (builder,
							     "label_misspelled_word"));
  suggestions_list = GTK_LIST_STORE (gtk_builder_get_object (builder,
							     "suggestions_list"));

  markup = g_markup_printf_escaped ("<span weight = \"bold\" foreground=\"red\">%s</span>", word);
  gtk_label_set_markup (label_misspelled_word,
			markup);
  g_free (markup);

  for (i = 0; i < suggestions_number; i++)
  {
    gtk_list_store_append (suggestions_list,
			   &iter);
    gtk_list_store_set (suggestions_list,
			&iter,
			0, g_strdup (g_list_nth_data (suggestions,
						      i)),
			-1);
  }
}

static void
ocr_utils_free_suggestions_model (OcrTextView *self)
{
  GtkDialog    *dialog;
  GtkBuilder   *builder;
  GtkTreeModel *suggestions_list;
  GtkTreeIter   iter;
  gboolean      valid;
  gchar        *str;
  
  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  dialog = g_object_get_data (G_OBJECT (self),
			      OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY);
  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);

  g_assert (GTK_IS_BUILDER (builder));

  suggestions_list = GTK_TREE_MODEL (gtk_builder_get_object (builder,
							     "suggestions_list"));
  g_assert (GTK_IS_LIST_STORE (suggestions_list));

  valid = gtk_tree_model_get_iter_first (suggestions_list,
					 &iter);
  while (valid)
  {
    gtk_tree_model_get (suggestions_list,
			&iter, 
			0, &str,
			-1);

    g_free (str);

    valid = gtk_tree_model_iter_next (suggestions_list,
				      &iter);
  }

  gtk_list_store_clear (GTK_LIST_STORE (suggestions_list));
}

static void
ocr_textview_spellcheck_dialog_response (GtkWidget   *dialog,
					 gint         response_id,
					 OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkBuilder         *builder;

  g_return_if_fail (GTK_IS_DIALOG (dialog));
  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  priv = self->priv;
  
  ocr_textbuffer_remove_tags (priv->textbuffer,
			      FALSE,
			      TRUE);

  ocr_textbuffer_clear_spellcheck_mark (priv->textbuffer);

  ocr_utils_free_suggestions_model (self);

  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);
  if (builder != NULL)
    g_object_unref (builder);

  gtk_widget_destroy (dialog);
}

static void
ocr_textview_spellcheck_finished (OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkDialog          *dialog;
  GtkBuilder         *builder;
  GtkWidget          *widget;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  dialog = g_object_get_data (G_OBJECT (self),
			      OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY);
  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);

  g_assert (GTK_IS_BUILDER (builder));

  priv = self->priv;

  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "substitute_button"));
  gtk_widget_set_sensitive (widget,
			    FALSE);
  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "ignore_button"));
  gtk_widget_set_sensitive (widget,
			    FALSE);
  widget = GTK_WIDGET (gtk_builder_get_object (builder,
					       "label_misspelled_word"));
  gtk_label_set_text (GTK_LABEL (widget),
		      "");

  ocr_utils_message (_("Spell check finished"),
		     NULL,
		     GTK_MESSAGE_INFO);
}

static void
ocr_textview_spellcheck_dialog_substitute (GtkButton   *button,
					   OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkDialog          *dialog;
  GtkBuilder         *builder;
  GtkEntry           *new_word_entry;
  GtkTreeView        *suggestions_view;
  GtkTreeModel       *suggestions_model;
  GtkTreeSelection   *suggestions_selection;
  GtkTreeIter         iter;
  gchar              *word = NULL;
  guint               suggestions_number;
  GList              *suggestions;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  dialog = g_object_get_data (G_OBJECT (self),
			      OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY);
  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);

  g_assert (GTK_IS_BUILDER (builder));

  priv = self->priv;

  suggestions_view = GTK_TREE_VIEW (gtk_builder_get_object (builder,
							    "suggestions_view"));
  new_word_entry = GTK_ENTRY (gtk_builder_get_object (builder,
						      "new_word_entry"));
  word = g_strdup (gtk_entry_get_text (new_word_entry));
  gtk_entry_set_text (new_word_entry,
		      "");

  if (word == NULL || !g_strcmp0 (word, ""))
  {
    suggestions_model = gtk_tree_view_get_model (suggestions_view);
    suggestions_selection = gtk_tree_view_get_selection (suggestions_view);
    if (gtk_tree_selection_get_selected (suggestions_selection,
					 &suggestions_model,
					 &iter))
    {
      gtk_tree_model_get (suggestions_model,
			  &iter,
			  0, &word,
			  -1);
    }
  }

  ocr_textbuffer_substitute_word (priv->textbuffer,
				  word);

  if (word != NULL)
    g_free (word);

  ocr_utils_free_suggestions_model (self);

  if (ocr_textbuffer_get_next_bad_word (priv->textbuffer,
					&word,
					&suggestions,
					&suggestions_number))
  {
    ocr_textview_spellcheck_dialog_put_data (self,
					     word,
					     suggestions,
					     suggestions_number);

    if (!GTK_WIDGET_HAS_FOCUS (GTK_WIDGET (suggestions_view)))
      gtk_widget_grab_focus (GTK_WIDGET (suggestions_view));

    g_free (word);
  }
  else
    ocr_textview_spellcheck_finished (self);
}

static void
ocr_textview_spellcheck_dialog_ignore (GtkButton   *button,
				       OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkDialog          *dialog;
  GtkBuilder         *builder;
  GtkEntry           *new_word_entry;
  GtkTreeView        *suggestions_view;
  gchar              *word = NULL;
  GList              *suggestions;
  guint               suggestions_number;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  dialog = g_object_get_data (G_OBJECT (self),
			      OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY);
  builder = g_object_get_data (G_OBJECT (dialog),
			       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY);

  g_assert (GTK_IS_BUILDER (builder));

  priv = self->priv;

  ocr_textbuffer_remove_tags (priv->textbuffer,
			      TRUE,
			      TRUE);

  ocr_utils_free_suggestions_model (self);

  new_word_entry = GTK_ENTRY (gtk_builder_get_object (builder,
						      "new_word_entry"));
  gtk_entry_set_text (new_word_entry,
		      "");

  suggestions_view = GTK_TREE_VIEW (gtk_builder_get_object (builder,
							    "suggestions_view"));
  if (ocr_textbuffer_get_next_bad_word (priv->textbuffer,
					&word,
					&suggestions,
					&suggestions_number))
  {
    ocr_textview_spellcheck_dialog_put_data (self,
					     word,
					     suggestions,
					     suggestions_number);

    if (!GTK_WIDGET_HAS_FOCUS (GTK_WIDGET (suggestions_view)))
      gtk_widget_grab_focus (GTK_WIDGET (suggestions_view));

    g_free (word);
  }
  else
    ocr_textview_spellcheck_finished (self);
}

void
ocr_textview_spellcheck (OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkBuilder         *builder;
  GtkWidget          *dialog;
  GtkTreeView        *suggestions_view;
  GtkTreeSelection   *suggestions_list_selection;
  GtkButton          *button;
  GtkWidget          *widget;
  gchar              *word;
  GList              *suggestions;
  guint               suggestions_number;
  GError             *error;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  priv = self->priv;

  if (!ocr_spellcheck_is_ok (priv->spellcheck))
    return;

  if (!ocr_textbuffer_spellcheck_put_underline_tag (priv->textbuffer))
  {
    ocr_utils_message (_("Spell check finished"),
		       NULL,
		       GTK_MESSAGE_INFO);
  }
  else
  {
    builder = gtk_builder_new ();

    g_assert (GTK_IS_BUILDER (builder));

    error = NULL;
    gtk_builder_add_from_file (builder,
			       OCR_UI_DIR"/"SPELLCHECK_DIALOG_BUILDER_FILE,
			       &error);

    if (error != NULL)
    {
      g_error (error->message,
	       NULL);
      g_clear_error (&error);
      return;
    }

    dialog = GTK_WIDGET (gtk_builder_get_object (builder,
						 "spellcheck_dialog"));

    g_assert (GTK_IS_DIALOG (dialog));

    g_object_set_data (G_OBJECT (self),
		       OCR_TEXTVIEW_SPELLCHECK_DIALOG_KEY, 
		       (gpointer) dialog);

    g_object_set_data (G_OBJECT (dialog),
		       OCR_TEXTVIEW_SPELLCHECK_DIALOG_BUILDER_KEY, 
		       (gpointer) builder);

    g_signal_connect (dialog,
		      "response", 
		      G_CALLBACK (ocr_textview_spellcheck_dialog_response),
		      (gpointer) self);


    widget = GTK_WIDGET (gtk_builder_get_object (builder,
						 "dictionary"));
    gtk_label_set_text (GTK_LABEL (widget),
			(priv->spellcheck)->dictionary_path);

    suggestions_view = GTK_TREE_VIEW (gtk_builder_get_object (builder,
							      "suggestions_view"));
    suggestions_list_selection = gtk_tree_view_get_selection (suggestions_view);
    gtk_tree_selection_set_mode (suggestions_list_selection,
				 GTK_SELECTION_BROWSE);

    button = GTK_BUTTON (gtk_builder_get_object (builder,
						 "substitute_button"));
    g_signal_connect (button,
		      "pressed", 
		      G_CALLBACK (ocr_textview_spellcheck_dialog_substitute),
		      (gpointer) self);

    button = GTK_BUTTON (gtk_builder_get_object (builder,
						 "ignore_button"));
    g_signal_connect (button,
		      "pressed", 
		      G_CALLBACK (ocr_textview_spellcheck_dialog_ignore),
		      (gpointer) self);

    if (ocr_textbuffer_get_next_bad_word (priv->textbuffer,
					  &word,
					  &suggestions,
					  &suggestions_number))
    {
      ocr_textview_spellcheck_dialog_put_data (self,
					       word,
					       suggestions,
					       suggestions_number);

      g_free (word);
    }

    gtk_widget_show_all (dialog);
  }
}

/* This is a try with Unicode library 
void
ocr_textview_spellcheck (OcrTextView *self)
{
  OcrTextViewPrivate *priv;
  GtkTextBuffer      *textbuffer;
  GtkTextIter         start, end;
  OcrSpellcheck      *spellcheck;
  UText              *utext;
  int32_t             len;
  UBreakIterator     *biter;
  int32_t             biter_start;
  int32_t             biter_end;
  GString            *string;
  UChar              *word_utf16;
  gchar              *word_utf8;
  gchar              *word_conv;
  gboolean            good_word;
  UErrorCode          uerror;
  GError             *error;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));

  priv = OCR_TEXTVIEW_GET_PRIVATE (self);

  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (self));
  spellcheck = priv->spellcheck;

  g_return_if_fail (GTK_IS_TEXT_BUFFER (textbuffer));
  g_return_if_fail (OCR_IS_SPELLCHECK (spellcheck));

  gtk_text_buffer_get_bounds (textbuffer, &start, &end);

  string = g_string_new (gtk_text_buffer_get_text(textbuffer,
                                                  &start,
                                                  &end,
                                                  FALSE));
  if (!(string->str))
  {
    g_object_unref (string);
    return;
  }

  if (g_utf8_validate(string->str, -1, NULL))
  {
    //"The text you are trying to check contains "
    //"an invalid byte sequence."
  }

  utext = NULL;
  biter = NULL;
  uerror = U_ZERO_ERROR;
  error = NULL;

  utext = utext_openUTF8 (utext, string->str, string->len, &uerror);
  biter = ubrk_open (UBRK_WORD, "it_IT", NULL, 0, &uerror);

  ubrk_setUText (biter, utext, &uerror);
  word_utf16 = g_new0(UChar, 20);
  biter_start = 0;
  biter_end = 0;
  while (ubrk_next (biter) != UBRK_DONE)
  {
    if (ubrk_getRuleStatus (biter) != UBRK_WORD_NONE)
    {
      biter_end = ubrk_current (biter);
      uerror = U_ZERO_ERROR;
      len = utext_extract (utext, biter_start, biter_end, NULL, 0, &uerror);


      if (len+1 > 20)
      {
        g_free (word_utf16);
        word_utf16 = g_new0(UChar, len+1);
      }


      uerror = U_ZERO_ERROR;
      len = utext_extract (utext, biter_start, biter_end, word_utf16, len+1, &uerror);
      word_utf8 = g_utf16_to_utf8 ((gunichar2*) word_utf16, (gsize) len, NULL, NULL, &error);
      if (!g_utf8_validate (word_utf8, -1, NULL)) break;
      //g_print("dic: %s", ocr_spellcheck_get_dic_encoding (spellcheck));
      word_conv = g_convert (word_utf8, -1, ocr_spellcheck_get_dic_encoding (spellcheck), "UTF-8", NULL, NULL, &error);
      good_word = ocr_spellcheck_check_word (spellcheck,
                                             word_conv);
      g_print ("word_utf8: %s\t good_word: %d\n", word_conv, good_word);
      g_free (word_utf8);
      g_free (word_conv);
    }
    biter_start = ubrk_current (biter);
  }

  utext_close (utext);
  ubrk_close (biter);
  g_string_free (string, TRUE);
  g_free (word_utf16);
  //g_assert (U_SUCCESS (uerror));

  priv->spellchecked_text = TRUE;
}
*/

void
ocr_textview_load_error_message (OcrTextView *self,
				 const gchar *stock_id,
				 const gchar *file_name,
				 const gchar *title,
				 const gchar *message)
{
  OcrTextViewPrivate *priv;
  GtkTextChildAnchor *image_anchor;
  GtkWidget          *hbox;
  GtkWidget          *image;
  GtkWidget          *label;
  gchar              *label_string;

  g_return_if_fail (OCR_IS_TEXTVIEW (self));
  g_return_if_fail (!(stock_id == NULL && file_name == NULL));
  g_return_if_fail (title != NULL);

  priv = self->priv;

  image_anchor = ocr_textbuffer_load_message (priv->textbuffer,
					      message);
  g_return_if_fail (GTK_IS_TEXT_CHILD_ANCHOR (image_anchor));

  if (stock_id != NULL)
    image = gtk_image_new_from_stock (stock_id,
				      GTK_ICON_SIZE_DIALOG);
  else
    image = gtk_image_new_from_file (file_name);

  label = gtk_label_new("");
  label_string = g_strdup (title);

  gtk_label_set_markup (GTK_LABEL (label),
			label_string);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (hbox), image, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
  gtk_widget_show_all (hbox);

  gtk_text_view_add_child_at_anchor (GTK_TEXT_VIEW (self),
				     hbox,
				     image_anchor);

  g_free (label_string);
}

