#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <glib/gprintf.h>

#include "ocr-window.h"
#include "ocr-panel.h"
#include "ocr-preferences.h"
#include "ocr-ui.h"
#include "ocr-commands.h"
#include "ocr-spellcheck.h"
#include "ocr-imagelist.h"

#define OCR_WINDOW_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), OCR_TYPE_WINDOW, OcrWindowPrivate))

G_DEFINE_TYPE (OcrWindow, ocr_window, GTK_TYPE_WINDOW);

struct _OcrWindowPrivate
{
  GtkUIManager    *ui_mgr;
  GtkActionGroup  *actions_window;
  GtkActionGroup  *actions_image;
  GtkActionGroup  *actions_text;
  GtkActionGroup  *actions_common;
  GtkWidget       *menubar;

  GtkWidget       *spinbutton;

  GtkTextTagTable *texttagtable;
  OcrSpellcheck   *spellcheck;
  OcrPreferences  *preferences;
  OcrImageList    *imagelist;
  GtkWidget       *hbox_panels;
  GtkWidget       *logo;

  OcrCursor        cursor;
};

static void
ocr_window_dispose (GObject *gobject)
{
  OcrWindow        *self;
  OcrWindowPrivate *priv;
  static gboolean   first_run = TRUE;

  g_return_if_fail (OCR_IS_WINDOW (gobject));

  self = OCR_WINDOW (gobject);
  priv = self->priv;

  if (first_run)
  {
    if (priv->spellcheck != NULL)
      g_object_unref (priv->spellcheck);

    if (priv->preferences != NULL)
      g_object_unref (priv->preferences);

    /* priv->imagelist is automatically destroyed */
    first_run = FALSE;
  }

  G_OBJECT_CLASS (ocr_window_parent_class)->dispose (gobject);
}

static void
ocr_window_finalize (GObject *gobject)
{
  g_return_if_fail (OCR_IS_WINDOW (gobject));

  G_OBJECT_CLASS (ocr_window_parent_class)->finalize (gobject);
}

GObject *
ocr_window_get_action_group (OcrWindow *self,
			     guint      action_group_number)
{
  OcrWindowPrivate *priv;
  GtkActionGroup *action_group;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  switch (action_group_number)
  {
    case 1:
      action_group = priv->actions_image;
      break;
    case 2:
      action_group = priv->actions_text;
      break;
    case 3:
      action_group = priv->actions_window;
      break; 
    default:
      action_group = NULL;
      break;
  }

  g_assert (GTK_IS_ACTION_GROUP (action_group));

  return G_OBJECT (action_group);
}

GObject *
ocr_window_get_ui_manager (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->ui_mgr);
}

GObject *
ocr_window_get_texttagtable (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->texttagtable);
}

GObject *
ocr_window_get_spellcheck (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->spellcheck);
}

GObject *
ocr_window_get_preferences (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return G_OBJECT (priv->preferences);
}

GtkWidget *
ocr_window_get_imagelist (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return GTK_WIDGET (priv->imagelist);
}

GtkWidget *
ocr_window_get_logo (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  priv = self->priv;

  return GTK_WIDGET (priv->logo);
}

void
ocr_window_set_cursor (OcrWindow *self,
		       OcrCursor  new_cursor)
{
  OcrWindowPrivate *priv;
  GdkCursor        *cursor = NULL;

  g_return_if_fail (OCR_IS_WINDOW (self));

  priv = self->priv;

  if (priv->cursor == new_cursor)
    return;

  switch (new_cursor)
  {
    case OCR_CURSOR_NORMAL:
      gdk_window_set_cursor ((GTK_WIDGET (self))->window,
			     NULL);
      priv->cursor = new_cursor;
      break;
    case OCR_CURSOR_OCCUPIED:
      cursor = gdk_cursor_new (GDK_WATCH);
      break;
    default:
      gdk_window_set_cursor ((GTK_WIDGET (self))->window,
			     NULL);
      priv->cursor = OCR_CURSOR_NORMAL;
  }

  if (cursor) 
  {
    gdk_window_set_cursor ((GTK_WIDGET (self))->window,
			   cursor);
    gdk_cursor_unref (cursor);
    gdk_flush();
    priv->cursor = new_cursor;
  }
}

static void
ocr_activate_recent_item (GtkRecentChooser *chooser,
			  OcrWindow        *window)
{
  gchar *uri;

  g_return_if_fail (GTK_IS_RECENT_CHOOSER (chooser));
  g_return_if_fail (OCR_IS_WINDOW (window));

  uri = gtk_recent_chooser_get_current_uri (chooser);

  g_assert (uri != NULL);

  ocr_load_image (window, uri);

  g_free (uri);
}

static GtkTextTagTable *
ocr_window_create_texttagtable (OcrWindow *self)
{
  GtkTextTagTable *texttagtable;
  GtkTextTag      *texttag;

  g_return_val_if_fail (OCR_IS_WINDOW (self), NULL);

  texttagtable = gtk_text_tag_table_new ();

  texttag = gtk_text_tag_new ("mispelled-word-underline");
  g_object_set (G_OBJECT (texttag),
		"underline", PANGO_UNDERLINE_ERROR,
		NULL);

  gtk_text_tag_table_add (texttagtable,
			  texttag);

  texttag = gtk_text_tag_new ("mispelled-word-during-check");
  g_object_set (G_OBJECT (texttag),
		"background", "yellow",
		NULL);

  gtk_text_tag_table_add (texttagtable,
			  texttag);

  return texttagtable;
}

void 
ocr_window_create_menu_and_toolbar (OcrWindow *self)
{
  OcrWindowPrivate *priv;
  GtkRecentManager *manager;
  GtkWidget        *menu;
  GtkWidget        *menuitem;
  GtkRecentFilter  *filter;
  GError           *error = NULL;

  g_return_if_fail (OCR_IS_WINDOW (self));

  priv = self->priv;

  priv->ui_mgr = gtk_ui_manager_new ();

  priv->actions_window = gtk_action_group_new ("ActionsWindow");
  gtk_action_group_set_translation_domain (priv->actions_window,
					   GETTEXT_PACKAGE);
  gtk_action_group_add_actions (priv->actions_window,
				action_entries_window,
				G_N_ELEMENTS (action_entries_window),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_window, 0);

  priv->actions_image = gtk_action_group_new ("ActionsImage");
  gtk_action_group_set_translation_domain (priv->actions_image,
					   GETTEXT_PACKAGE);
  gtk_action_group_add_actions (priv->actions_image,
				action_entries_image,
				G_N_ELEMENTS (action_entries_image),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_image, 0);

  priv->actions_text = gtk_action_group_new ("ActionsText");
  gtk_action_group_set_translation_domain (priv->actions_text,
					   GETTEXT_PACKAGE);
  gtk_action_group_add_actions (priv->actions_text,
				action_entries_text,
				G_N_ELEMENTS (action_entries_text),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_text, 0);

  gtk_window_add_accel_group (GTK_WINDOW (self),
			      gtk_ui_manager_get_accel_group (priv->ui_mgr));

  /* TODO - In ActionCommon there is a Lock button to lock scrollbars of image view and text view
  priv->actions_common = gtk_action_group_new ("ActionsCommon");
  gtk_action_group_add_actions (priv->actions_common,
				action_entries_common,
				G_N_ELEMENTS (action_entries_common),
				self);
  gtk_ui_manager_insert_action_group (priv->ui_mgr, priv->actions_common, 0);
  */

  if (!gtk_ui_manager_add_ui_from_file (priv->ui_mgr,
					OCR_UI_DIR"/ocr-ui.xml",
					&error))
  {
    g_error ("Building menus failed: %s", error->message);
    g_clear_error (&error);
    return;
  }

  manager = gtk_recent_manager_get_default ();
  menu = gtk_recent_chooser_menu_new_for_manager (manager);
  gtk_recent_chooser_set_show_private (GTK_RECENT_CHOOSER (menu),
				       TRUE);
  gtk_recent_chooser_set_show_not_found (GTK_RECENT_CHOOSER (menu),
					 FALSE);
  gtk_recent_chooser_set_show_tips (GTK_RECENT_CHOOSER (menu),
				    TRUE);
  gtk_recent_chooser_set_sort_type (GTK_RECENT_CHOOSER (menu),
				    GTK_RECENT_SORT_MRU);
  filter = gtk_recent_filter_new ();
  gtk_recent_filter_add_application (filter,
				     g_get_application_name());
  gtk_recent_chooser_add_filter (GTK_RECENT_CHOOSER (menu),
				 filter);
  
  menuitem = gtk_ui_manager_get_widget (priv->ui_mgr,
					"/MenuBar/FileMenu/RecentFiles");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem),
			     menu);
  
  g_signal_connect (G_OBJECT (menu),
		    "item-activated",
                    G_CALLBACK (ocr_activate_recent_item),
		    self);

  priv->menubar = gtk_ui_manager_get_widget (priv->ui_mgr, "/MenuBar");
  g_assert (GTK_IS_MENU_BAR (priv->menubar));
}

static void
ocr_window_create_ui (OcrWindow *self)
{
  OcrWindowPrivate *priv;
  GtkWidget        *vbox_main;
  GtkWidget        *hbox_main;
  GtkWidget        *hbox_panels;
  GList            *icon_list = NULL;
  GdkPixbuf        *pixbuf;
  GtkWidget        *logo;

  g_return_if_fail (OCR_IS_WINDOW (self));

  priv = self->priv;

  gtk_window_set_title (GTK_WINDOW (self), _("OcrGui"));
  gtk_container_set_border_width (GTK_CONTAINER (self), 2);

  /*
  GtkIconTheme  *icon_theme;
  gchar        **paths;
  gint           n, i;
  icon_theme = gtk_icon_theme_get_default ();
  gtk_icon_theme_get_search_path (icon_theme,
				  &paths,
				  &n);
  for (i=0; i<n; i++)
    g_print ("path %d: %s\n",
	     i+1,
	     paths[i]);*/

  gtk_window_set_default_icon_name ("ocrgui");

  vbox_main = gtk_vbox_new (FALSE, 3);
  gtk_container_add (GTK_CONTAINER (self), vbox_main);

  gtk_box_pack_start (GTK_BOX (vbox_main), priv->menubar, FALSE, TRUE, 0);

  hbox_main = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (vbox_main), hbox_main, TRUE, TRUE, 0);

  gtk_box_pack_start (GTK_BOX (hbox_main), GTK_WIDGET (priv->imagelist), FALSE, FALSE, 0);

  hbox_panels = gtk_hbox_new (FALSE, 5);
  priv->hbox_panels = hbox_panels;
  gtk_box_pack_start (GTK_BOX (hbox_main), hbox_panels, TRUE, TRUE, 0);

  priv->logo = gtk_viewport_new (NULL,
				 NULL);
  logo = gtk_image_new_from_file (OCR_UI_DIR"/logo.png");
  gtk_container_add (GTK_CONTAINER (priv->logo),
		     logo);
  gtk_box_pack_start (GTK_BOX (hbox_panels), GTK_WIDGET (priv->logo), TRUE, TRUE, 0);

  self->statusbar = GTK_STATUSBAR (gtk_statusbar_new ());
  gtk_box_pack_start (GTK_BOX (vbox_main), GTK_WIDGET (self->statusbar), FALSE, FALSE, 0);
  self->context_id = gtk_statusbar_get_context_id (self->statusbar,
						   "main");
}

void
ocr_window_destroy (OcrWindow *self)
{
  gtk_main_quit ();
}

static void
ocr_window_constructed (GObject *gobject)
{
  OcrWindow        *self;
  OcrWindowPrivate *priv;
  gchar            *dictionary_path;

  g_return_if_fail (OCR_IS_WINDOW (gobject));

  self = OCR_WINDOW (gobject);
  priv = self->priv;

  ocr_window_create_menu_and_toolbar (self);

  priv->preferences = g_object_new (OCR_TYPE_PREFERENCES,
				    "main-window", (gpointer) self,
				    NULL); 

  priv->spellcheck = g_object_new (OCR_TYPE_SPELLCHECK,
				   NULL);
  dictionary_path = ocr_preferences_get_hunspell_dictionary_path (priv->preferences);
  if (dictionary_path != NULL)
  {
    (priv->spellcheck)->dictionary_path = g_strdup (dictionary_path);
    ocr_spellcheck_create (priv->spellcheck);

    g_free (dictionary_path);
  }


  priv->texttagtable = ocr_window_create_texttagtable (self);

  priv->imagelist = g_object_new (OCR_TYPE_IMAGE_LIST,
				  "homogeneous", FALSE,
				  "spacing", 3,
				  "main-window", self,
				  NULL);

  ocr_window_create_ui (self);
}

void
ocr_window_add_panel (OcrWindow *self,
		      GtkWidget *panel)
{
  OcrWindowPrivate *priv;

  g_return_if_fail (OCR_IS_WINDOW (self));
  g_return_if_fail (GTK_IS_WIDGET (panel));

  priv = self->priv;

  g_return_if_fail (GTK_IS_HBOX (priv->hbox_panels));

  if (GTK_WIDGET_VISIBLE (priv->logo))
    gtk_widget_hide (priv->logo);

  gtk_box_pack_start (GTK_BOX (priv->hbox_panels), panel, TRUE, TRUE, 0);
}

static void
ocr_window_class_init (OcrWindowClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (OcrWindowPrivate));

  gobject_class->constructed = ocr_window_constructed;
  gobject_class->dispose = ocr_window_dispose;
  gobject_class->finalize = ocr_window_finalize;
}

static void
ocr_window_init (OcrWindow *self)
{
  OcrWindowPrivate *priv;

  g_return_if_fail (OCR_IS_WINDOW (self));

  self->priv = OCR_WINDOW_GET_PRIVATE (self);
  priv = self->priv;

  priv->ui_mgr = NULL;

  g_signal_connect (self, "delete-event",
		    G_CALLBACK (ocr_window_destroy),
		    NULL);
}

